import datetime
import struct


# Порядок следования байт в пакете события
BYTE_EVENT_LENGTH = 2
BYTE_EVENT_INDEX = 3
BYTE_EVENT_DESC_LENGTH = 5
BYTE_EVENT_CODE = 6
BYTE_EVENT_DESC = 7

# Порядок следования байт в описании типов полей
FIELD_BYTE_CODE = 0
FIELD_BYTE_LENGTH = 1
FIELD_BYTE_DATA = 2

# Коды типов полей
FIELD_CODE_USER = 1
FIELD_CODE_PARTITION = 2
FIELD_CODE_AREA = 3
FIELD_CODE_RELAY_INDEX = 5
FIELD_CODE_RELAY_STATE = 7
FIELD_CODE_TIME = 11
FIELD_CODE_PARTITION_ID = 24


class Event:
    """ Описание события, содержащее дополнительные поля, идентифицирующие данное событи.
        Байты 0, 1, 2 являются необходимыми атрибутами пакета Modbus;
        байты 3 и 4 – номер события;
        байт 5 – длина описания события;
        байт 6 – собственно код события.
        Далее идут поля атрибутов события. Каждое поле имеет следующую структуру:
            - код типа поля: No зоны Modbus / No раздела Modbus / идентификатор раздела
                Modbus/No пользователя / Время и Дата/No реле Modbus / состояние реле Modbus;
            - количество байт данных поля. Только поле «Дата и Время» содержит 6 байт данных,
                остальные поля содержат 2 байта данных;
            - данные поля. Все поля, кроме поля «Дата и Время», содержат номер (код) того или иного
                атрибута.
    """

    user_id = None
    partition = None
    area = None
    relay_index = None
    relay_state = None
    time = None
    partitionID = None

    def __init__(self, data: bytes):
        self.index = struct.unpack_from('>H', data, BYTE_EVENT_INDEX)[0]
        self.code = data[BYTE_EVENT_CODE]

        # длина
        i = BYTE_EVENT_DESC + FIELD_BYTE_CODE
        while i < data[BYTE_EVENT_LENGTH] + BYTE_EVENT_LENGTH:
            code_type = data[i]
            if code_type == FIELD_CODE_TIME:
                try:
                    self.time = datetime.datetime(2000 + data[i+7], data[i+6], data[i+5], *data[i+2:i+5])
                except Exception:
                    self.time = '20%s-%s-%s %s:%s:%s' % (data[i+7], data[i+6], data[i+5], data[i+2], data[i+3], data[i+4])
            elif code_type > 0:
                value = struct.unpack_from('>H', data, i+FIELD_BYTE_DATA)[0]
                if code_type == FIELD_CODE_USER:
                    self.user_id = value
                elif code_type == FIELD_CODE_PARTITION:
                    self.partition = value
                elif code_type == FIELD_CODE_AREA:
                    self.area = value
                elif code_type == FIELD_CODE_RELAY_INDEX:
                    self.relay_index = value
                elif code_type == FIELD_CODE_RELAY_STATE:
                    self.relay_state = value
                elif code_type == FIELD_CODE_PARTITION_ID:
                    self.partitionID = value
            else:
                break

            i += data[i+FIELD_BYTE_LENGTH] + FIELD_BYTE_DATA

    def __str__(self):
        res = []
        if self.user_id is not None:
            res.append('User: %s' % self.user_id)
        if self.partition is not None:
            res.append('Раздел: %s' % self.partition)
        if self.area is not None:
            res.append('Зона: %s' % self.area)
        if self.relay_index is not None:
            res.append('Реле №%s' % self.relay_index)
        if self.relay_state is not None:
            res.append('Состояние реле: %s' % ("On" if self.relay_state else "Off"))
        if self.time is not None:
            res.append('Время: %s' % self.time)
        if self.partitionID is not None:
            res.append('ID раздела: %s' % self.partitionID)

        return f'{CODES[self.code] or "Event(%s)" % self.code} {"; ".join(res)}'


CODES = [
    "",
    "Восстановление сети 220 В",
    "Авария сети 220 В",
    "Тревога проникновения",
    "Помеха", "",
    "Помеха устранена", "", "",
    "Активация УДП",
    "Восстановление УДП", "", "", "", "", "", "",
    "Неудачное взятие",
    "Предъявлен код принуждения",
    "Тест",
    "Вход в режим тестирования",
    "Выход из режима тестирования",
    "Восстановление контроля",
    "Задержка взятия",
    "Взятие входа на охрану", "", "", "", "", "", "", "", "", "",
    "Идентификация 1,",
    "Восстановление технологического входа",
    "Нарушение технологического входа",
    "Пожар",
    "Нарушение 2-го технологического входа",
    "Восстановление нормы оборудования",
    "Пожар 2",
    "Неисправность оборудования",
    "Неизвестное устройство", "",
    "«Внимание!»",
    "Обрыв входа",
    "Обрыв ДПЛС", "", "", "", "", "", "", "", "", "", "", "",
    "Тихая тревога", "", "", "", "", "", "", "", "", "", "", "", "",
    "Понижение уровня",
    "Норма уровня", "",
    "Повышение уровня",
    "Аварийное повышение уровня",
    "Повышение температуры",
    "Аварийное понижение уровня",
    "Температура в норме",
    "Тревога затопления",
    "Восстановление датчика затопления", "",
    "Неисправность термометра",
    "Восстановление термометра",
    "Начало локального программирования", "", "", "", "", "",
    "Неисправность канала связи",
    "Восстановление канала связи", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
    "Снятие входа с охраны",
    "Сброс тревоги", "", "", "", "", "", "",
    "Восстановление снятого входа",
    "Тревога входа",
    "Нарушение снятого входа", "",
    "Обрыв выхода",
    "КЗ выхода",
    "Восстановление выхода", "", "",
    "Потеря связи с выходом",
    "Восстановление связи с выходом",
    "Изменение состояния выхода", "",
    "Включение насоса",
    "Выключение насоса", "", "", "",
    "Ошибка при автоматическом тестировании", "",
    "Пуск", "",
    "Неудачный пуск пожаротушения",
    "Запуск внутреннего теста",
    "Задержка пуска АУП",
    "Автоматика АУП выключена",
    "Отмена пуска АУП",
    "Тушение",
    "Аварийный пуск АУП",
    "Пуск АУП",
    "Блокировка пуска АУП",
    "Автоматика АУП включена",
    "Взлом корпуса прибора",
    "Пуск речевого оповещения",
    "Отмена пуска РО",
    "Восстановление корпуса прибора",
    "ИУ в рабочем состоянии",
    "ИУ в исходном состоянии",
    "Отказ ИУ",
    "Ошибка ИУ", "",
    "Восстановление внутренней зоны",
    "Задержка пуска РО", "",
    "Останов задержки пуска АУП", "", "", "",
    "Ошибка параметров входа", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
    "Восстановление связи со входом",
    "Потеря связи по ДПЛС1",
    "Потеря связи по ДПЛС2",
    "Восстановление связи по ДПЛС1",
    "Отключение выходного напряжения",
    "Подключение выходного напряжения",
    "Перегрузка источника питания",
    "Перегрузка источника питания устранена",
    "Неисправность зарядного устройства",
    "Восстановление зарядного устройства",
    "Неисправность источника питания",
    "Восстановление источника питания",
    "Восстановление батареи",
    "Восстановление связи по ДПЛС2",
    "Неисправность батареи",
    "Перезапуск прибора",
    "Требуется обслуживание",
    "Ошибка теста АКБ",
    "Понижение температуры", "", "", "", "",
    "Батарея разряжена",
    "Разряд резервной батареи",
    "Восстановление резервной батареи",
    "Короткое замыкание входа",
    "Короткое замыкание ДПЛС",
    "Сработка датчика",
    "Отключение ветви RS-485",
    "Восстановление ветви RS-485", "",
    "Срабатывание СДУ",
    "Отказ СДУ",
    "Повышение напряжения ДПЛС",
    "Отметка наряда", "", "", "", "", "", "", "", "", "", "", "", "", "",
    "Раздел снят по принуждению", "", "", "",
    "Раздел взят",
    "Раздел снят", "", "", "", "", "", "", "",
    "Потеряна связь с прибором",
    "Восстановлена связь с прибором", "",
    "Включение пульта С2000М",
    ""
]
