from spread_core.bam.entities import Funit
from spread_core.bam.providers import Provider
from spread_core.bam.animeo import const as C
from spread_core.bam import const as BAM


class AnimeoRemote(Provider):
    def __init__(self, id: int, binding: str, **kwargs):
        super(AnimeoRemote, self).__init__(id, **kwargs)
        self._binding = binding

    @staticmethod
    def parse_args(attributes: dict, **kwargs) -> dict:
        return dict(
            binding=attributes['oid']
        )

    @property
    def binding(self):
        return self._binding

    def request(self, cmd, *args, **kwargs):
        return self.manager.send_interface.send(cmd, **kwargs)

    def on_update(self, funit_type, value, **kwargs):
        if funit_type == PositionLevel:
            self.on_update(Motion, Motion.from_level(self.get_value(PositionLevel), value), **kwargs)

        super(AnimeoRemote, self).on_update(funit_type, value, **kwargs)

    def query_status(self, sig=None):
        try:
            response = self.request(C.CMD_REMOTES_STATUS, id=self.binding)
            level = int(response['remotes']['remote']['position'])
        except:
            self.set_valid(self.INVALID)
        else:
            self.set_valid(self.VALID)
            self.on_update(PositionLevel, PositionLevel.from_raw(level), sig=sig)

    def set_level(self, target: int, sig=None):
        if not 0 <= target <= 100:
            raise ValueError('position expected as int[0..100] but %s given' % target)

        self.request(C.CMD_REMOTE, id=self.binding, action=C.ACTION_ABSOLUTE, position=target)

    def set_motion(self, direction: str, sig=None):
        self.request(C.CMD_REMOTE, id=self.binding, action=direction)

    def check_valid(self):
        self.query_status()

    def get_info(self):
        pass


class _AnimeoFunit(Funit):
    entity_type = AnimeoRemote

    @property
    def entity(self) -> entity_type:
        return super(_AnimeoFunit, self).entity

    def get(self):
        self.entity.query_status(sig=self.signature)


class Motion(_AnimeoFunit):
    variants = (BAM.MOTOR_OPEN, BAM.MOTOR_CLOSE, BAM.MOTOR_STOP)
    mapped_value = (C.ACTION_UP, C.ACTION_DOWN, C.ACTION_STOP)
    state_variants = (BAM.MOTOR_OPENING, BAM.MOTOR_CLOSING, BAM.MOTOR_STOPPED)

    @staticmethod
    def from_level(has_level, level):
        if level is None:
            return None
        elif has_level is None:
            return BAM.MOTOR_STOPPED
        else:
            different = level - has_level
            if different > 1:
                return BAM.MOTOR_CLOSING
            elif different < -1:
                return BAM.MOTOR_OPENING
            else:
                return BAM.MOTOR_STOPPED

    def set(self):
        self.entity.set_motion(self.set_value, sig=self.signature)


class PositionLevel(_AnimeoFunit):
    value_range = range(0, 10_001)

    @staticmethod
    def from_raw(raw_level):
        if raw_level is None:
            return None

        try:
            return int(raw_level * 100)
        except:
            return None

    @property
    def set_value(self):
        level = super(PositionLevel, self).set_value
        if level == 0:
            return 0

        return int(level / 100)

    def set(self):
        self.entity.set_level(self.set_value, sig=self.signature)
