import logging
import socket
import struct
import threading

from spread_bolid import commands, errors
from spread_core.tools.base_interface import AbstractInterfaceClass
from spread_core.tools.utils import bytes2str


class _Interface(AbstractInterfaceClass):
    def __init__(self):
        super(_Interface, self).__init__(None)


class ModbusTcp(_Interface):
    PROTOCOL_ID = 0
    WAIT_ON_BUSY = 0.5

    def __init__(self):
        super(ModbusTcp, self).__init__()
        self._socket = None
        self.counter = 0
        self._bus = None
        self._address = None

    def start(self):
        self.connect()

    def set_bus_params(self, bus: list):
        self._address = ('10.10.1.173', 502)
        # self._address = tuple(bus)

    def connect(self):
        if self._socket:
            self._socket.close()
            self._socket = None

        self.counter = 0
        _socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        _socket.settimeout(0.7)
        logging.info('TcpModbus connect...')
        while not self._stopped:
            try:
                _socket.connect(self._address)
            except ConnectionRefusedError:
                threading.Event().wait(3)
            else:
                logging.info('TcpModbus connected')
                self._socket = _socket
                self._bus = self._send()
                self._bus.send(None)
                break

    @property
    def next_count(self):
        self.counter += 1
        if self.counter > 0xffff:
            self.counter = 0

        return self.counter

    def send(self, cmd: commands.Command, repeat=False):
        header = struct.pack('>HHH', self.next_count, ModbusTcp.PROTOCOL_ID, len(cmd.frame))

        logging.debug(f'-> {cmd} {bytes2str(cmd.frame)}')
        response = self._bus.send(header + cmd.frame)
        try:
            result = cmd.of(response)
        except errors.ModbusError as ex:
            logging.debug(f'   <- {ex}')
            if not repeat:
                threading.Event().wait(self.WAIT_ON_BUSY)
                return self.send(cmd, repeat=True)
        except BaseException as ex:
            logging.debug(f'   ERR {bytes2str(response)}, {ex}')
        else:
            logging.debug(f'   <- {result if response else "<empty>"}')
            return result

    def _send(self):
        out = None
        while self._socket:
            frame = yield out
            try:
                self._socket.send(frame)
                header = self._socket.recv(6)
                out = self._socket.recv(struct.unpack_from('>H', header, 4)[0])
            except socket.timeout:
                if out:
                    logging.debug('bad data: %s' % bytes2str(out))


# def run():
#     i = ModbusTcp()
#     i.start()
#     event_desc_length = i.send(commands.EventDescLength(slave_id=5))
#     # _datetime = i.send(commands.ReadDatetime(slave_id=5))
#     # version = i.send(commands.Version(slave_id=5))
#     max_relay_count = 8 # i.send(commands.MaxRelayCount(slave_id=5))
#
#     # relay_status = i.send(commands.ReadGroupRelay(0, 8, slave_id=5))
#     # i.send(commands.SetRelayGroupState(0, [False]*8, slave_id=5))
#     # time.sleep(2)
#     # relay_status = i.send(commands.ReadGroupRelay(0, 8, slave_id=5))
#     # for index in range(8):
#     #     time.sleep(1)
#     #     i.send(commands.SetRelayState(index+1, True, slave_id=5))
#
#     # relay_status = i.send(commands.ReadGroupRelay(7, 1, slave_id=5))
#     # i.send(commands.SetRelayState(7, False, slave_id=5))
#     # time.sleep(1)
#     # relay_status = i.send(commands.ReadGroupRelay(7, 1, slave_id=5))
#     # i.send(commands.SetRelayGroupState(1, [not v for v in relay_status], slave_id=5))
#
#     # i.send(commands.SetAreaState(5, 24, slave_id=5))
#     # i.send(commands.SetAreaState(6, 24, slave_id=5))
#     # i.send(commands.SetAreaState(7, 24, slave_id=5))
#
#     while True:
#         for _ in range(i.send(commands.NewEventCount(slave_id=5), silent=True)):
#             event = i.send(commands.FirstEvent(event_desc_length, slave_id=5))
#             i.send(commands.MarkEvent(event.index, slave_id=5), silent=True)
#
#         time.sleep(1)
#
#
# # if __name__ == '__main__':
# #     run()
