import spread_dali as dali
from spread_core.bam.dali.managers import Dali, DALI1, DALI2


class ExternalDispatcher:
    def __init__(self, manager: Dali):
        self.manager = manager

    def dispatch(self, command, response, *args):
        if self.is_for_source(command):
            self.handle_by_source(command)
        else:
            self.handle_by_targets(command, response, *args)

    @staticmethod
    def is_for_source(command) -> bool:
        return isinstance(command, (dali.gear.general._SpecialCommand, dali.device.general._SpecialDeviceCommand))

    def handle_by_source(self, command):
        if isinstance(command, dali.gear.general.DTR0):
            self.on_dtr(DALI1, 0, command.param)
        elif isinstance(command, dali.gear.general.DTR1):
            self.on_dtr(DALI1, 1, command.param)
        elif isinstance(command, dali.gear.general.DTR2):
            self.on_dtr(DALI1, 2, command.param)
        elif isinstance(command, dali.device.general.DTR0):
            self.on_dtr(DALI2, 0, command._opcode)
        elif isinstance(command, dali.device.general.DTR1):
            self.on_dtr(DALI2, 1, command._opcode)
        elif isinstance(command, dali.device.general.DTR2):
            self.on_dtr(DALI2, 2, command._opcode)
        elif isinstance(command, dali.device.general.DTR1DTR0):
            self.on_dtr(DALI2, 0, command._opcode)
            self.on_dtr(DALI2, 1, command._instance)
        elif isinstance(command, dali.device.general.DTR2DTR1):
            self.on_dtr(DALI2, 1, command._opcode)
            self.on_dtr(DALI2, 2, command._instance)

    def on_dtr(self, dali_type: int, index: int, value: int):
        self.manager.set_dtr_value(dali_type, index, value)

    def handle_by_targets(self, command, response, *args):
        try:
            handler = ExternalHandlerAbstract.of(command)(self.manager, command)
        except (TypeError, AttributeError):
            pass
        else:
            handler.handle(response, *args)


class ExternalHandlerAbstract:
    @staticmethod
    def of(command: dali.command.Command):
        if not ExternalDispatcher.is_for_source(command):
            if isinstance(command.destination, dali.address.Short):
                return SingleExternalHandler
            if isinstance(command.destination, dali.address.Group):
                return GroupExternalHandler
            if isinstance(command.destination, dali.address.Broadcast):
                return BroadcastExternalHandler

        raise TypeError(f'{command.__class__.__name__} are not dispatched')

    def __init__(self, manager: Dali, command: dali.command.Command):
        self.manager = manager
        self.command = command

    @property
    def dali_type(self):
        if isinstance(self.command, dali.gear.general._GearCommand):
            return DALI1

        if isinstance(self.command, dali.device.general._DeviceCommand):
            return DALI2

        raise TypeError(f'{self.command} is unknown DALI district')

    @property
    def source(self):
        return self.command.destination

    @property
    def targets(self):
        raise NotImplementedError()

    def handle(self, response, *args):
        for provider in self.targets:
            provider.on_external(self.command, response, *args)


class SingleExternalHandler(ExternalHandlerAbstract):
    @property
    def targets(self):
        yield from self.manager.providers_by_address(self.dali_type, self.source)


class GroupExternalHandler(ExternalHandlerAbstract):
    @property
    def targets(self):
        if self.manager.has_group_provider(self.dali_type, self.source.group):
            yield from self.manager.providers_by_address(self.dali_type, self.source)
        else:
            yield from self.manager.get_group_members(self.dali_type, self.source.group)


class BroadcastExternalHandler(ExternalHandlerAbstract):
    @property
    def targets(self):
        broadcast = self.manager.broadcast(self.dali_type)
        if broadcast:
            yield broadcast
        else:
            yield from self.manager.get_all_devices_of_type(self.dali_type)
