import datetime
import logging
import math

import spread_core.mqtt as mqtt
import spread_dali as dali
from spread_core.mqtt.spread.variable import Variable
from spread_core.tools import utils
from spread_core.tools.timer import Timer
from spread_dali.gear import general as gear, led as led, colour_control as cc, emergency as em
from .providers import DaliProvider, DaliFunit, Groups, Validity, Discovery
from ..tuner import TunerLimited
from ... import const as C

DISCOVERY_DELAY = 0.5


class DaliLight(DaliProvider):
    _dali_type = 1

    def __init__(self, id: int, binding: dali.address.Address, **kwargs):
        super(DaliLight, self).__init__(id, binding, **kwargs)
        self._fadeTime = 2
        self._fadeRate = 0
        self._last_level = 0

        self._level_timer = None
        self.is_unaddress_tunned = False
        self._tuner = None
        self._level_timer = None

        self.__level_before_discovery = None

    @property
    def fadeTime(self): return self._fadeTime

    @property
    def fadeRate(self): return self._fadeRate

    @fadeTime.setter
    def fadeTime(self, value):
        self._fadeTime = value

        if self.is_device:
            for _group in self.groups():
                parts = [provider.fadeTime for provider in _group.get_members()]
                if parts:
                    _group.fadeTime = max(parts)

            broadcast = self.manager.broadcast(self._dali_type)
            if isinstance(broadcast, DaliLight):
                broadcast.fadeTime = max((provider.fadeTime for provider in broadcast.get_members()))

    @fadeRate.setter
    def fadeRate(self, value):
        self._fadeRate = value

        if self.is_device:
            for _group in self.groups():
                parts = [provider.fadeRate for provider in _group.get_members()]
                if parts:
                    _group.fadeRate = max(parts)

            broadcast = self.manager.broadcast(self._dali_type)
            if isinstance(broadcast, DaliLight):
                broadcast.fadeRate = max((provider.fadeRate for provider in broadcast.get_members()))

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type in (CurrentLevelRaw, BrightnessLevel, Tuning):
            return False

        return super(DaliLight, self).is_repaired_funit(funit_type)

    @property
    def tuner(self) -> TunerLimited:
        return self._tuner

    def on_survey(self):
        yield from super(DaliLight, self).on_survey()

        if self.is_valid_strong: yield self.request(gear.QueryActualLevel(self.short_addr))

        if self.is_valid_strong: yield self.request(gear.QueryFadeTimeFadeRate(self.short_addr))
        if self.is_valid_strong: yield self.request(gear.QueryMinLevel(self.short_addr))
        if self.is_valid_strong: yield self.request(gear.QueryMaxLevel(self.short_addr))
        if self.is_valid_strong and self.dali_device_type != 1: yield self.request(led.QueryDimmingCurve(self.short_addr))
        if self.is_valid_strong: yield self.request(gear.QueryPowerOnLevel(self.short_addr))
        if self.is_valid_strong: yield self.request(gear.QuerySystemFailureLevel(self.short_addr))
        if self.is_valid_strong: yield self.request(gear.QueryPhysicalMinimum(self.short_addr))
        if self.is_valid_strong: yield self.request(gear.QueryGroupsZeroToSeven(self.short_addr))
        if self.is_valid_strong: yield self.request(gear.QueryGroupsEightToFifteen(self.short_addr))

        for i in range(0, 16):
            if self.is_valid_strong: yield self.request(gear.QuerySceneLevel(self.short_addr, i))

    def request(self, cmd, *args, **kwargs):
        if isinstance(cmd, (TunerLimited.UP, TunerLimited.DOWN)):
            if self.tuner and not self.tuner.is_active:
                logging.debug(f'skip {cmd}')
                return

        if isinstance(cmd, gear.DTR0):
            self._manager.set_dtr_value(self._dali_type, 0, cmd.param)
        elif isinstance(cmd, gear.DTR1):
            self._manager.set_dtr_value(self._dali_type, 1, cmd.param)
        elif isinstance(cmd, gear.DTR2):
            self._manager.set_dtr_value(self._dali_type, 2, cmd.param)

        super(DaliLight, self).request(cmd, *args, **kwargs)

        if cmd.is_send_notify:
            self.on_command_sended(cmd)

    def on_request_answer(self, funit_type, response, *args, sig=None):
        if isinstance(response, gear.QueryFadeTimeAndRateResponse) and response:
            ft = None if response.fade_time is None else 'ft%s' % int(FadeTime.values[response.fade_time]*10)
            self.on_update(FadeTime, ft, sig=sig)

            fr = None if response.fade_rate is None else 'fr%s' % int(FadeRate.values[response.fade_rate]*10)
            self.on_update(FadeRate, fr, sig=sig)
        elif funit_type == CurrentLevelRaw and response is None:
            return
        elif funit_type == CurrentLevelRaw and response.value.as_integer == 255:
            self.on_update(funit_type, None, sig=sig)
        elif funit_type == Types and response:
            self.on_update(Types, [response.value.as_integer], sig=sig)
        elif funit_type == Types.part_funit and response:
            self.on_update(*Types.update_part(self, response.value.as_integer), sig=sig)
        elif isinstance(response, gear.QueryStatusResponse):
            self.on_update(Validity, not response.error, sig=sig)
        elif funit_type == SceneLevelsRaw:
            level = response.value.as_integer if response else 255
            self.on_update(SceneLevelsRaw, self._parse_scenes(level, args[0]), sig=sig)
        else:
            super(DaliLight, self).on_request_answer(funit_type, response, *args, sig=sig)

    def on_exit(self):
        if self._tuner:
            self._tuner.stop()
        self.stop_level_query()
        super(DaliLight, self).on_exit()

    def check_valid(self):
        super(DaliLight, self).check_valid()
        if self.is_device:
            self.request(gear.QueryStatus(self.short_addr))

    def get_info(self):
        super(DaliLight, self).get_info()

        if self.is_device:
            if FadeTime not in self.values or FadeRate not in self.values:
                self.request(gear.QueryFadeTimeFadeRate(self.short_addr))
            if MinLevelRaw not in self.values:
                self.request(gear.QueryMinLevel(self.short_addr))
            if MaxLevelRaw not in self.values:
                self.request(gear.QueryMaxLevel(self.short_addr))
            if DimmingCurve not in self.values:
                if self.dali_device_type != 1:
                    self.request(led.QueryDimmingCurve(self.short_addr))
            if SceneLevelsRaw not in self.values:
                self.GetSceneLevelsRaw()

            self.query_level_sync()

            if Types not in self.values:
                self.GetTypes()
            if PowerOnLevelRaw not in self.values:
                self.request(gear.QueryPowerOnLevel(self.short_addr))
            if SystemFailureLevelRaw not in self.values:
                self.request(gear.QuerySystemFailureLevel(self.short_addr))
            if PhysicalMinLevelRaw not in self.values:
                self.request(gear.QueryPhysicalMinimum(self.short_addr))
        else:
            # fadeTime меняется при изменении fadeTime e объектов
            self.on_update(FadeTime, self.get_value(FadeTime, FadeTime.index_of[1]))
            self.on_update(FadeRate, self.get_value(FadeRate, FadeRate.index_of[1]))
            self.on_update(MinLevelRaw, self.get_value(MinLevelRaw, 1))
            self.on_update(MaxLevelRaw, self.get_value(MaxLevelRaw, 254))
            if self.dali_device_type != 1:
                self.on_update(DimmingCurve, self.get_value(DimmingCurve, DimmingCurve.index_of[0]))
            self.on_update(SceneLevelsRaw, self.get_value(SceneLevelsRaw, []))
            self.on_update(CurrentLevelRaw, self.get_value(CurrentLevelRaw, 0))
            self.on_update(PowerOnLevelRaw, self.get_value(PowerOnLevelRaw, 0))
            self.on_update(SystemFailureLevelRaw, self.get_value(SystemFailureLevelRaw, 80))
            self.on_update(PhysicalMinLevelRaw, self.get_value(PhysicalMinLevelRaw, 1))

    def query_level_async(self, sig=None):
        if self.get_value(Discovery):
            return

        for _provider in self.get_members():
            _provider.stop_level_query()
            if self._manager.level_sync_delay:
                _provider._level_timer = Timer.set(max(_provider.fadeTime, self._manager.level_sync_delay), _provider.query_level, sig)

    def stop_level_query(self):
        if self._level_timer:
            try:
                self._level_timer.close()
            except ValueError:
                pass
            finally:
                self._level_timer = None

    def query_level_sync(self, sig=None):
        self.request(gear.QueryActualLevel(self.short_addr), sig=sig)

    def query_level(self, sig=None):
        self._level_timer = None
        if self.is_device:
            self.manager.put_to_stack(CurrentLevelRaw(self, Variable(None)).get)
        else:
            for _provider in self.get_members():
                _provider.query_level()

    def make_step(self, step):
        self.manager.put_to_stack(_Step(self, Variable(step)).set)

    def on_update(self, funit_type, value, sig=None, **kwargs):
        super(DaliLight, self).on_update(funit_type, value, sig, **kwargs)

        if funit_type in [CurrentLevelRaw, DimmingCurve]:
            self.GetBrightnessLevel(sig, kwargs.get(Variable.PRELIM, False))

        elif funit_type == FadeTime:
            self.fadeTime = FadeTime.values[FadeTime.index_of.index(value)]

        elif funit_type == FadeRate:
            self.fadeRate = FadeRate.values[FadeRate.index_of.index(value)]

    def on_command_sended(self, cmd):
        if cmd.is_levels_dependent:
            for _provider in self.get_members(True):
                value = None
                cl = _provider.get_value(CurrentLevelRaw, 0)

                if isinstance(cmd, gear.DAPC):
                    value = cmd.power
                elif isinstance(cmd, gear.Off):
                    value = 0
                elif isinstance(cmd, gear.RecallMinLevel):
                    value = 1
                elif isinstance(cmd, gear.RecallMaxLevel):
                    value = 254
                elif isinstance(cmd, gear.GoToLastActiveLevel):
                    value = _provider._last_level
                elif isinstance(cmd, gear.Up):
                    if cl > 0:
                        value = cl + max(1, round(self.fadeRate/5))
                elif isinstance(cmd, gear.Down):
                    if cl > 0:
                        value = max(1, cl - max(1, round(self.fadeRate/5)))
                elif isinstance(cmd, gear.StepUp):
                    if cl > 0:
                        value = cl + 1
                elif isinstance(cmd, gear.OnAndStepUp):
                    value = cl + 1
                elif isinstance(cmd, gear.StepDown):
                    if cl > 0:
                        value = cl - 1
                elif isinstance(cmd, gear.StepDownAndOff):
                    value = cl - 1
                    if value < _provider.get_value(MinLevelRaw, 1):
                        value = 0
                elif isinstance(cmd, gear.GoToScene):
                    scenes = _provider.get_value(SceneLevelsRaw, None)
                    scene_id = cmd.frame.pack[1] - 0x10
                    if scenes is not None and scene_id in scenes and scenes[scene_id] in range(0, 254):
                        value = scenes[scene_id]

                if value is not None:
                    if _provider.is_valid:
                        if value != 0:
                            value = max(_provider.get_value(MinLevelRaw, 0), value)
                            value = min(_provider.get_value(MaxLevelRaw, 254), value)
                            _provider._last_level = value
                        _provider.on_update(CurrentLevelRaw, value, **{Variable.PRELIM: True})

            if self._tuner:
                self._tuner.check_state(cmd)

            if isinstance(cmd, (Discovery.DOWN_CMD, Discovery.UP_CMD)) and self.get_value(Discovery):
                Timer.set(DISCOVERY_DELAY, self.switch_level_by_discovery, has_cmd=cmd.__class__)

            self.query_level_async(cmd.sig)

        elif isinstance(cmd, gear.SetFadeTime):
            if self.dtr(0) is not None:
                ft = 'ft%s' % int(FadeTime.values[self.dtr(0)] * 10)
                self.update_recursive(FadeTime, ft, **{Variable.PRELIM: True})
        elif isinstance(cmd, gear.SetFadeRate):
            if self.dtr(0) is not None:
                fr = 'fr%s' % int(FadeRate.values[self.dtr(0)] * 10)
                self.update_recursive(FadeRate, fr, **{Variable.PRELIM: True})
        elif isinstance(cmd, (gear.AddToGroup, gear.RemoveFromGroup)):
            included = isinstance(cmd, gear.AddToGroup)
            for part in self.get_members(return_self=True):
                if isinstance(part, DaliLight):
                    value = Groups.def_value(part._dali_type)
                    value = [{**i} for i in part.get_value(Groups, value)]
                    for i in value:
                        if i['index'] == cmd.param:
                            i['value'] = included
                            break
                    else:
                        value.append({'index': cmd.param, 'value': included})

                    part.on_update(Groups, value, **{Variable.PRELIM: True})
        elif isinstance(cmd, (gear.SetScene, gear.RemoveFromScene)):
            level = self.dtr(0) if isinstance(cmd, gear.SetScene) else 255
            for part in self.get_members(return_self=True):
                if isinstance(part, DaliLight):
                    value = SceneLevelsRaw.def_value()
                    value = [{**i} for i in part.get_value(SceneLevelsRaw, value)]

                    cpl = False

                    for i in value:
                        if i['index'] == cmd.param:
                            i['value'] = level
                            cpl = True
                            break

                    if cpl is False:
                        value.append({'index': cmd.param, 'value': level})

                    part.on_update(SceneLevelsRaw, value, **{Variable.PRELIM: True})
        elif isinstance(cmd, led.SelectDimmingCurve):
            if self.dtr(0) is not None:
                self.update_recursive(DimmingCurve, DimmingCurve.of_raw(self.dtr(0)), **{Variable.PRELIM: True})
        elif isinstance(cmd, gear.SetSystemFailureLevel):
            if self.dtr(0) is not None:
                self.update_recursive(SystemFailureLevelRaw, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, gear.SetPowerOnLevel):
            if self.dtr(0) is not None:
                self.update_recursive(PowerOnLevelRaw, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, gear.SetMaxLevel):
            if self.dtr(0) is not None:
                self.update_recursive(MaxLevelRaw, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, gear.SetMinLevel):
            if self.dtr(0) is not None:
                self.update_recursive(MinLevelRaw, self.dtr(0), **{Variable.PRELIM: True})
        else:
            super(DaliLight, self).on_command_sended(cmd)

    def prs2raw(self, p):
        if p == 0:
            return p

        try:
            p = abs(p)
        except (TypeError, ValueError) as ex:
            logging.error(ex)
            return 0

        _min = self.get_value(MinLevelRaw, 1)
        _max = self.get_value(MaxLevelRaw, 254)

        if self.get_value(DimmingCurve, None) == DimmingCurve.LINEAR:
            return min(max(_min, round(254 * p / 100)), _max)
        else:
            return min(max(_min, round(((math.log10(p) + 1) * 253 / 3 + 1))), _max)

    def raw2prs(self, r):
        if self.get_value(DimmingCurve, None) == DimmingCurve.LINEAR:
            return min(100, round(100 * r / 254))
        else:
            if r == 0:
                return 0
            elif r < 1:
                return 1
            else:
                p = 10 ** ((r - 1) * 3 / 253 - 1)
                if 0 < p < 1:
                    p = 1
                return round(p)

    def read_memory_location(self, bank_id, start_address, end_address) -> bytes:
        self.check_command_for_device()
        i = self.manager.send_interface

        i.send(gear.DTR1(bank_id))

        for _dtr0 in range(start_address, end_address):
            i.send(gear.DTR0(_dtr0))
            res = i.send(gear.ReadMemoryLocation(self.short_addr))
            if isinstance(res, dali.command.Response):
                yield res.value.as_integer
            else:
                break

    def SetDiscovery(self, value, sig=None):
        _discovery = self.get_value(Discovery, False)
        if _discovery != value:
            self.on_update(Discovery, value, sig=sig)
            if value:
                self.__level_before_discovery = self.get_value(BrightnessLevel, 0)
                self.switch_level_by_discovery()
            else:
                has_level = self.__level_before_discovery
                if has_level is not None:
                    self.SetBrightnessLevel(has_level)

    def switch_level_by_discovery(self, has_cmd=None):
        if not self.get_value(Discovery):
            return

        if has_cmd == Discovery.UP_CMD:
            cmd = Discovery.DOWN_CMD
        else:
            cmd = Discovery.UP_CMD

        self.request(cmd(self._binding))

    def _parse_scenes(self, value, index) -> list:
        scenes = [
            {key: value for key, value in item.items()}
            for item in self.get_value(SceneLevelsRaw, SceneLevelsRaw.def_value())
        ]
        if len(scenes) != SceneLevelsRaw.length:
            scenes = SceneLevelsRaw.def_value()

        scenes[index]['value'] = value

        return scenes

    def GetSceneLevelsRaw(self, _=None, sig=None, indexes=None):
        if self.is_device:
            if indexes is None:
                indexes = range(0, 16)
            for i in indexes:
                self.request(gear.QuerySceneLevel(self.short_addr, i))

        self.on_update(SceneLevelsRaw, self.get_value(SceneLevelsRaw, SceneLevelsRaw.def_value()), sig=sig)

    def SetSceneLevelsRaw(self, val, sig=None):
        indexes = []
        for item in val:
            index = item['index']
            self.request(gear.DTR0(max(0, min(255, item['value']))))
            self.request(gear.SetScene(self._binding, index))

            indexes.append(index)

        if self.is_device:
            self.GetSceneLevelsRaw(sig=sig, indexes=indexes)
        else:
            self.update_recursive(SceneLevelsRaw, val, sig=sig)

    def GetGroups(self, _=None, of_0_7=True, of_8_15=True, sig=None):
        if self.is_device:
            if of_0_7:
                self.request(gear.QueryGroupsZeroToSeven(self.short_addr))
            if of_8_15:
                self.request(gear.QueryGroupsEightToFifteen(self.short_addr))

        super(DaliLight, self).GetGroups(sig=sig)

    def SetGroups(self, val, sig=None):
        has_0_7 = False
        has_8_15 = False

        for item in val:
            index = item['index']

            if item['value']:
                self.request(gear.AddToGroup(self._binding, index))
            else:
                self.request(gear.RemoveFromGroup(self._binding, index))

            if index < 8:
                has_0_7 = True
            else:
                has_8_15 = True

        if self.is_device:
            self.GetGroups(sig=sig, of_0_7=has_0_7, of_8_15=has_8_15)
        else:
            self.update_recursive(Groups, val, sig=sig)

    def GetTypes(self, sig=None):
        _types = None

        if self.is_device:
            i = self._manager.send_interface

            _type = i.send(gear.QueryDeviceType(self.short_addr))
            if _type:
                if _type.value.as_integer == 0xff:
                    _types = []
                    while _type is not None:
                        _type = i.send(gear.QueryNextDeviceType(self.short_addr))
                        if _type:
                            if _type.value.as_integer == 0xfe:
                                break
                            else:
                                _types.append(_type.value.as_integer)

                    if len(_types) == 0:
                        _types.append(0xff)
                else:
                    _types = [_type.value.as_integer]

        self.on_update(Types, _types, sig=sig)

    def SetGroupLevel(self, value, sig=None):
        try:
            value = int(value)
        except:
            raise dali.exceptions.CommandGenerateError('Value({}) must be integer'.format(value))
        else:
            group_ids = self.get_value(Groups, None)
            group_id = None
            group_size = 999999999
            if group_ids is None:
                self.GetGroups()
                group_ids = self.get_value(Groups, [])

            for item in group_ids:
                if item['value'] is True and 'index' in item:
                    _group_id = item['index']
                    _group_size = self.manager.get_group_size(self._dali_type, _group_id)
                    if _group_size < group_size:
                        group_size = _group_size
                        group_id = _group_id

            if group_id is None:
                self.SetBrightnessLevel(value, sig=sig)

            elif self.manager.has_group_provider(self._dali_type, group_id):
                for group in self.manager.group_providers(self._dali_type, group_id):
                    group.SetBrightnessLevel(value, sig=sig)

            else:
                if value == 0:
                    raw_value = 0
                elif self.get_value(DimmingCurve, None) == DimmingCurve.LINEAR:
                    raw_value = round(254 * value / 100)
                else:
                    raw_value = round(((math.log10(value) + 1) * 253 / 3 + 1))
                cmd = gear.DAPC(dali.address.Group(group_id), raw_value)
                self.request(cmd)
                for _participant in self.manager.get_group_members(self.dali_type, group_id):
                    _participant.on_command_sended(cmd)

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        if self.is_valid is False:
            raise dali.exceptions.CommandGenerateError('Device is invalid!')

        method = kwargs.pop(mqtt.const.METHOD, None)
        if method == gear.RecallMinLevel.__name__:
            cmd = gear.RecallMinLevel(self._binding)
        elif method == gear.RecallMaxLevel.__name__:
            cmd = gear.RecallMaxLevel(self._binding)
        elif method == 'Scene':
            cmd = gear.GoToScene(self._binding, value)
        elif method == gear.Off.__name__:
            cmd = gear.Off(self._binding)
        elif method == gear.GoToLastActiveLevel.__name__:
            cmd = gear.GoToLastActiveLevel(self._binding)
        elif method == gear.StepUp.__name__:
            cmd = gear.StepUp(self._binding)
        elif method == gear.StepDown.__name__:
            cmd = gear.StepDown(self._binding)
        elif method == gear.OnAndStepUp.__name__:
            cmd = gear.OnAndStepUp(self._binding)
        elif method == gear.StepDownAndOff.__name__:
            cmd = gear.StepDownAndOff(self._binding)
        elif method == gear.Up.__name__:
            cmd = gear.Up(self._binding)
        elif method == gear.Down.__name__:
            cmd = gear.Down(self._binding)
        else:
            cmd = gear.DAPC(self._binding, self.prs2raw(value))
        cmd.sig = sig
        if self.is_device:
            if self.is_unaddress_tunned:
                for group in self.groups():
                    if group.tuner:
                        group.tuner.set_pause(2+self.fadeTime)

                bc = self.manager.broadcast(self._dali_type)
                if bc and bc.tuner:
                    bc.tuner.set_pause(2 + self.fadeTime)

        if self.tuner:
            self.tuner.set_pause(2+self.fadeTime)

        self.request(cmd, **kwargs)

    def GetBrightnessLevel(self, sig=None, preliminary=False):
        _level = self.get_value(CurrentLevelRaw, None)
        if _level is not None:
            _level = self.raw2prs(_level)
        self.on_update(BrightnessLevel, _level, sig=sig, preliminary=preliminary)

    def SetTuning(self, value, speed=None, sig=None):
        _c_val = self.get_value(Tuning, 0)
        speed = 0.1 if speed is None else speed

        if _c_val * value <= 0:
            self.on_update(Tuning, value, sig=sig)
            if value == 0:
                if self._tuner:
                    self._tuner.stop()
                    self._tuner = None
            else:
                if self._tuner is None:
                    self._tuner = TunerLimited(self)

                self._tuner.set_direction(TunerLimited.UP if value > 0 else TunerLimited.DOWN, speed)
        elif self._tuner:
            self._tuner.speed = speed

    def set_pause_for_group(self, group_id):
        if group_id == -1:
            bc = self.manager.broadcast(self._dali_type)
            if bc and bc.tuner:
                bc.tuner.set_pause(2+self.fadeTime)
        else:
            for group in self.manager.group_providers(self._dali_type, group_id):
                if group.tuner:
                    group.tuner.set_pause(2+self.fadeTime)


class DaliDimmer(DaliLight):
    def __init__(self, id: int, binding: dali.address.Address, **kwargs):
        super(DaliDimmer, self).__init__(id, binding, **kwargs)
        self.step_cmd = None


class _DaliColor(DaliDimmer):
    def __init__(self, *args, **kwargs):
        super(_DaliColor, self).__init__(*args, **kwargs)
        self.temperature_timer = None

    def stop_temperature_timer(self):
        if self.temperature_timer:
            try:
                self.temperature_timer.close()
            except ValueError:
                pass
            finally:
                self.temperature_timer = None

    def on_exit(self):
        self.stop_temperature_timer()
        super(_DaliColor, self).on_exit()


class DaliTunableWhite(_DaliColor):
    def __init__(self, id: int, binding: dali.address.Address, **kwargs):
        super(DaliTunableWhite, self).__init__(id, binding, **kwargs)

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == Temperature:
            return False

        return super(DaliTunableWhite, self).is_repaired_funit(funit_type)

    def get_info(self):
        super(DaliTunableWhite, self).get_info()
        if self.is_device:
            if Coolest not in self.values:
                self.GetCoolest()
            if Warmest not in self.values:
                self.GetWarmest()
            self.GetTemperature()
        else:
            self.on_update(Temperature, self.get_value(Temperature, 5500))
            self.on_update(Coolest, self.get_value(Coolest, 2000))
            self.on_update(Warmest, self.get_value(Warmest, 9000))

    def on_survey(self):
        yield from super(DaliTunableWhite, self).on_survey()

        if self.is_valid_strong: yield self.GetTemperature()
        if self.is_valid_strong: yield self.GetCoolest()
        if self.is_valid_strong: yield self.GetWarmest()

    def on_command_sended(self, cmd):
        if isinstance(cmd, cc.SetTemperature):
            Temperature.apply_extended(self, self.dtr(0), self.dtr(1))

        elif isinstance(cmd, cc.StoreColourTemperatureLimit):
            _TemperatureLimitFunit.apply_extended(self, self.dtr(0), self.dtr(1), self.dtr(2))

        else:
            super(DaliTunableWhite, self).on_command_sended(cmd)

    def _get_color_value(self, funit_type, value_type: int, sig=None):
        res = None
        try:
            self.request(gear.DTR0(value_type))
            res = self.manager.send_interface.send(cc.QueryColourValue(self.short_addr)).value.as_integer
            if res != 0xff:
                b0 = self.manager.send_interface.send(gear.QueryContentDTR0(self.short_addr)).value.as_integer
                b1 = self.manager.send_interface.send(gear.QueryContentDTR1(self.short_addr)).value.as_integer
                res = Temperature.mirek2kelvin(int.from_bytes([b0, b1], 'little'))
        except:
            pass
        finally:
            self.on_update(funit_type, res, sig=sig)

    def _set_color_limit(self, value_type: int, value: int):
        if not 1000 <= value <= 20000:
            raise ValueError("temperature must be in the range 1000..20000")
        mirek = Temperature.kelvin2mirek(value)
        bts = [int(b) for b in mirek.to_bytes(2, 'little')] + [value_type]
        self.request(gear.DTR0(bts[0]))
        self.request(gear.DTR1(bts[1]))
        self.request(gear.DTR2(bts[2]))
        self.request(cc.StoreColourTemperatureLimit(self._binding))

        _TemperatureLimitFunit.apply_extended(self, *bts)

    def query_temperature(self):
        self.stop_temperature_timer()
        self.temperature_timer = Timer.set(self.fadeTime, self.manager.put_to_stack, Temperature(self, Variable(None)).get)

    def SetTemperature(self, temperature=1, sig=None):
        if not 1000 <= temperature <= 20000:
            raise ValueError("temperature must be in the range 1000..20000")
        try:
            mirek = Temperature.kelvin2mirek(temperature)
            bts = [int(b) for b in mirek.to_bytes(2, 'little')]
            self.request(gear.DTR0(bts[0]))
            self.request(gear.DTR1(bts[1]))
            self.request(cc.SetTemperature(self._binding))
            self.request(cc.Activate(self._binding))
            Temperature.apply_extended(self, *bts, sig=sig)
        except BaseException as ex:
            logging.error(ex)
        else:
            self.query_temperature()

    def GetTemperature(self, sig=None):
        self._get_color_value(Temperature, cc.QueryColourValue.TEMPERATURE, sig=sig)

    def SetCoolest(self, value=0, sig=None):
        try:
            self._set_color_limit(cc.StoreColourTemperatureLimit.COOLEST, value)
        except BaseException as ex:
            logging.error(ex)
        else:
            self.GetCoolest(sig)

    def GetCoolest(self, sig=None):
        self._get_color_value(Coolest, cc.QueryColourValue.COOLEST, sig=sig)

    def SetWarmest(self, value=0, sig=None):
        try:
            self._set_color_limit(cc.StoreColourTemperatureLimit.WARMEST, value)
        except BaseException as ex:
            logging.error(ex)
        else:
            self.GetWarmest(sig)

    def GetWarmest(self, sig=None):
        self._get_color_value(Warmest, cc.QueryColourValue.WARMEST, sig=sig)


class DaliRgbwaf(_DaliColor):
    def __init__(self, id: int, binding: dali.address.Address, **kwargs):
        super(DaliRgbwaf, self).__init__(id, binding, **kwargs)
        self.colour_timer = None

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == Color:
            return False

        return super(DaliRgbwaf, self).is_repaired_funit(funit_type)

    def get_info(self):
        super(DaliRgbwaf, self).get_info()
        if self.is_device:
            if ChannelsCount not in self.values:
                self.request(cc.QueryColourTypeFeatures(self.binding))
            self.GetColor()
        else:
            self.on_update(ChannelsCount, 6)
            self.on_update(Color, self.get_value(Color, Color.def_value(6)))

    def on_survey(self):
        yield from super(DaliRgbwaf, self).on_survey()

        if self.is_valid_strong:
            yield self.GetColor()

    def on_command_sended(self, cmd):
        if isinstance(cmd, (cc.SetRgbColour, cc.SetWafColour)):
            Color.apply_extended(self, cmd, (self.dtr(i) for i in range(3)))

        else:
            super(DaliRgbwaf, self).on_command_sended(cmd)

    def on_request_answer(self, funit_type, response, *args, sig=None):
        if isinstance(response, cc.ColourTypeFeaturesResponse):
            self.on_update(ChannelsCount, response.rgb_channels)
        # if response and funit_type == 'ColorValue':
        #     color_value = response.value.as_integer
        #
        #     if color_value != 255:
        #
        #         value = self.get_value(Color, Color.def_value())
        #         color_index = args[0] - cc.QueryColourValue.RED
        #
        #         if color_index < 3:
        #             color_name = Color.RGB_KEYS[color_index]
        #
        #         else:
        #             color_name = Color.WAF_KEYS[color_index-3]
        #
        #         if value.get(color_name, None) != color_value:
        #             self.on_update(Color, {**value, color_name: color_value}, sig=sig)
        else:
            super(DaliRgbwaf, self).on_request_answer(funit_type, response, *args, sig=sig)

    def query_color(self, sig=None):
        self.stop_temperature_timer()
        if self.fadeTime:
            self.temperature_timer = Timer.set(
                self.fadeTime,
                self.manager.put_to_stack,
                Color(self, Variable(None, sig=sig)).get
            )
        else:
            self.GetColor(sig=sig)

    def GetColor(self, sig=None):
        self.check_command_for_device()
        channels_count = self.get_value(ChannelsCount, 0)

        value = None

        if channels_count:
            value = dict()
            for index, color_name in enumerate(Color.RGB_KEYS + Color.WAF_KEYS):
                try:
                    if index >= channels_count:
                        break

                    self.manager.send_interface.send(gear.DTR0(cc.QueryColourValue.RED + index))
                    r = self.manager.send_interface.send(cc.QueryColourValue(self.short_addr))
                    if r:
                        value[color_name] = Color.from_raw(r.value.as_integer)

                except Exception as ex:
                    logging.warning('Color(%s) reading error: %s' % (color_name, str(ex)))
                    value[color_name] = None

        self.on_update(Color, value, sig=sig)

    def SetColor(self, rgb_waf, sig=None):
        for CommandClass, components in zip((cc.SetRgbColour, cc.SetWafColour), rgb_waf):
            if components:
                dtr0 = gear.DTR0(components[0])
                dtr1 = gear.DTR1(components[1])
                dtr2 = gear.DTR2(components[2])

                self.request(dtr0)
                self.request(dtr1)
                self.request(dtr2)
                self.request(CommandClass(self.binding))

        self.request(cc.Activate(self.binding))

        for CommandClass, components in zip((cc.SetRgbColour, cc.SetWafColour), rgb_waf):
            if components:
                Color.apply_extended(self, CommandClass(self.binding), components, sig=sig)


class DaliRelay(DaliLight):
    def on_update(self, funit_type, value, sig=None, **kwargs):
        if funit_type == BrightnessLevel:
            self.on_update(On, None if value is None else value == 100, sig=sig, **kwargs)
        else:
            super(DaliRelay, self).on_update(funit_type, value, sig=sig, **kwargs)


COSINE_DRIVER = '#COSINE_1_1'


class DaliEmergency(DaliLight):
    _dali_device_type = 1   # Emergency

    def __init__(self, id: int, binding: dali.address.Address, **kwargs):
        super(DaliEmergency, self).__init__(id, binding, **kwargs)

    @property
    def standard(self) -> bool:
        if str(self.name).lower().rfind(COSINE_DRIVER.lower()) >= 0:
            return False  # не стандартный провайдер (ЮАРовский)
        else:
            return True

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type in (EmergencyMode, ChargeLevelRaw):
            return False

        return super(DaliEmergency, self).is_repaired_funit(funit_type)

    def get_info(self):
        if self.standard:
            self._get_info_full()
        else:
            self._get_info_short()

    def _get_info_full(self):
        # !!! do not skip super.get_info !!!
        super(DaliEmergency, self).get_info()

        if self.is_device:
            if Types not in self.values:
                self.GetTypes()

            if RatedDuration not in self.values:
                self.request(em.QueryRatedDuration(self.short_addr))

            self.request(em.QueryBatteryCharge(self.short_addr))
            self.request(em.QueryEmergencyMode(self.short_addr))
            self.request(em.QueryEmergencyStatus(self.short_addr))
            self.request(em.QueryDurationTestResult(self.short_addr))
            self.request(em.QueryEmergencyLevel(self.short_addr))
            self.request(em.QueryEmergencyMaxLevel(self.short_addr))
            self.request(em.QueryEmergencyMinLevel(self.short_addr))
            DaysBetweenFunctionTest(self, None).get()
            WeeksBetweenDurationTest(self, None).get()
            UntilNextFunctionTest(self, None).get()
            UntilNextDurationTest(self, None).get()
            LastFunctionTestInfo(self, None).get()
            LastDurationTestInfo(self, None).get()
            Mode(self, None).get()
        else:
            self.on_update(Mode, 'Unknown')
            self.on_update(EmergencyMode, 'Normal')
            self.on_update(EmergencyStatus, '')
            self.on_update(DurationTestResult, 0)
            self.on_update(ChargeLevelRaw, 0)
            self.on_update(EmergencyLevelRaw, 0)
            self.on_update(EmergencyMaxLevelRaw, 0)
            self.on_update(EmergencyMinLevelRaw, 0)
            self.on_update(DaysBetweenFunctionTest, 0)
            self.on_update(WeeksBetweenDurationTest, 0)
            self.on_update(UntilNextFunctionTest, 0)
            self.on_update(UntilNextDurationTest, 0)
            self.on_update(LastFunctionTestInfo, {'timestamp': '', 'result': False})
            self.on_update(LastDurationTestInfo, {'timestamp': '', 'duration': 0, 'ratedDuration': 0, 'result': False})

    def _get_info_short(self):
        if self.is_device:
            if SystemFailureLevelRaw not in self.values:
                self.request(gear.QuerySystemFailureLevel(self.short_addr))
            if PhysicalMinLevelRaw not in self.values:
                self.request(gear.QueryPhysicalMinimum(self.short_addr))

            if Types not in self.values:
                self.GetTypes()

            Mode(self, None).get()
        else:
            self.on_update(Mode, 'Unknown')
            self.on_update(EmergencyMode, 'Normal')
            self.on_update(EmergencyStatus, '')
            self.on_update(LastFunctionTestInfo, {'timestamp': '', 'result': False})
            self.on_update(LastDurationTestInfo, {'timestamp': '', 'duration': 0, 'ratedDuration': 0, 'result': False})

    def on_survey(self):
        if self.standard:
            yield from self._on_survey_full()
        else:
            yield from self._on_survey_short()

    def _on_survey_full(self):
        # !!! do not skip super.on_survey !!!
        yield from super(DaliEmergency, self).on_survey()

        if self.is_valid_strong: yield self.request(em.QueryBatteryCharge(self.short_addr))
        if self.is_valid_strong: yield self.request(em.QueryEmergencyMode(self.short_addr))
        if self.is_valid_strong: yield self.request(em.QueryEmergencyStatus(self.short_addr))
        if self.is_valid_strong: yield self.request(em.QueryEmergencyLevel(self.short_addr))
        if self.is_valid_strong: yield self.request(em.QueryEmergencyMaxLevel(self.short_addr))
        if self.is_valid_strong: yield self.request(em.QueryEmergencyMinLevel(self.short_addr))
        if self.is_valid_strong: yield DaysBetweenFunctionTest(self, None).get()
        if self.is_valid_strong: yield WeeksBetweenDurationTest(self, None).get()
        if self.is_valid_strong: yield UntilNextFunctionTest(self, None).get()
        if self.is_valid_strong: yield UntilNextDurationTest(self, None).get()
        if self.is_valid_strong: yield LastFunctionTestInfo(self, None).get()
        if self.is_valid_strong: yield LastDurationTestInfo(self, None).get()
        if self.is_valid_strong: yield Mode(self, None).get()

    def _on_survey_short(self):
        if self.is_valid_strong: yield self.request(gear.QuerySystemFailureLevel(self.short_addr))
        if self.is_valid_strong: yield self.request(gear.QueryPhysicalMinimum(self.short_addr))
        if self.is_valid_strong: yield self.request(gear.QueryGroupsZeroToSeven(self.short_addr))
        if self.is_valid_strong: yield self.request(gear.QueryGroupsEightToFifteen(self.short_addr))

        if self.is_valid_strong: yield self.request(em.QueryEmergencyStatus(self.short_addr))
        if self.is_valid_strong: yield Mode(self, None).get()

    def check_valid(self):
        # !!! skip super.on_survey !!!
        # super(DaliEmergency, self).check_valid()

        if self.is_device:
            self.request(em.QueryEmergencyFailureStatus(self.short_addr))
            self.build_mode_topic()
            if self.standard:
                self.request(em.QueryDurationTestResult(self.short_addr))

    def on_command_sended(self, cmd):
        super(DaliEmergency, self).on_command_sended(cmd)

    def on_request_answer(self, funit_type, response, *args, sig=None):
        if funit_type == Validity and response is None:
            self.on_update(Validity, False, sig=sig)

        elif isinstance(response, em.QueryEmergencyFailureStatusResponse):
            self.on_update(Validity, True)
            self.on_update(FailureStatus, response.status_lower_case, sig=sig)

        elif isinstance(response, em.QueryEmergencyModeResponse) and response:
            self.on_update(EmergencyMode, response.mode.title().replace(' ', ''), sig=sig)

        elif isinstance(response, em.QueryEmergencyStatusResponse) and response:
            self.on_update(EmergencyStatus, response.status_lower_case, sig=sig)

        # elif isinstance(response, em.QueryDurationTestResult) and response:
        elif funit_type == DurationTestResult and response:
            self.on_update(DurationTestResult,
                           int(datetime.timedelta(minutes=response.value.as_integer * 2).total_seconds()), sig=sig)

        # elif isinstance(response, em.QueryEmergencyLevel) and response:
        elif funit_type == EmergencyLevelRaw and response:
            self.on_update(EmergencyLevelRaw, response.value.as_integer, sig=sig)

        # elif isinstance(response, em.QueryEmergencyMaxLevel) and response:
        elif funit_type == EmergencyMaxLevelRaw and response:
            self.on_update(EmergencyMaxLevelRaw, response.value.as_integer, sig=sig)

        # elif isinstance(response, em.QueryEmergencyMinLevel) and response:
        elif funit_type == EmergencyMinLevelRaw and response:
            self.on_update(EmergencyMinLevelRaw, response.value.as_integer, sig=sig)

        elif funit_type == 'Timing' and response:
            if args[0] == DaysBetweenFunctionTest.DTR_VAL:
                self.on_update(DaysBetweenFunctionTest, response.value.as_integer, sig=sig)

            elif args[0] == WeeksBetweenDurationTest.DTR_VAL:
                self.on_update(WeeksBetweenDurationTest, response.value.as_integer, sig=sig)

        elif funit_type == RatedDuration and response:
            self.on_update(RatedDuration,
                           int(datetime.timedelta(minutes=response.value.as_integer * 2).total_seconds()), sig=sig)

        else:
            super(DaliEmergency, self).on_request_answer(funit_type, response, *args, sig=sig)

    def on_update(self, funit_type, value, sig=None, **kwargs):
        super(DaliEmergency, self).on_update(funit_type, value, sig=sig, **kwargs)

        if funit_type == ChargeLevelRaw:
            _lvl = ChargeLevel.of_raw(value)
            self.on_update(ChargeLevel, _lvl, sig=sig, nullable=True if _lvl is None else False)

    def query_emergency_function_test_timing(self, sig=None):
        self._query_emergency_test_timing(UntilNextFunctionTest, 0, sig=sig)

    def query_emergency_duration_test_timing(self, sig=None):
        self._query_emergency_test_timing(UntilNextDurationTest, 2, sig=sig)

    def _query_emergency_test_timing(self, funit_type, dtr, sig=None):
        try:
            value = int(self._until_next_test(dtr).total_seconds())
            self.on_update(funit_type, value, sig=sig)
        except BaseException as ex:
            logging.error(ex)

    def _until_next_test(self, dtr: int) -> datetime.timedelta:
        """ Определяет время, оставшееся до старта теста """
        self.manager.send_interface.send(gear.DTR0(dtr))
        value = self.manager.send_interface.send(em.QueryTestTiming(self.short_addr)).value.as_integer

        high = value
        low = self.manager.send_interface.send(gear.QueryContentDTR1(self.short_addr)).value.as_integer
        value = int.from_bytes(bytes([high, low]), 'big')

        return datetime.timedelta(minutes=value * 15)

    def _test_period(self, dtr: int) -> int:
        """Определяет периодичность преведения теста
        Для Функционального теста данное значение представляет кол-во дней
        Для теста Аккумулятора данное значение представляет кол-во недель
        """
        self.manager.send_interface.send(gear.DTR0(dtr))
        value = self.manager.send_interface.send(em.QueryTestTiming(self.short_addr)).value.as_integer
        return value

    def _last_test_datetime(self, dtr1, dtr2) -> str:
        """Определяет дату и время последнего теста
        Т.к. дата и время последнего теста пока нигде не хранится, то значение определяется так:
        1. Узнаем кол-во секунд до старта теста
        2. Узнаем периодичность в секундах
        3. К текущей дате прибавляем секунды до старта (1), получем дату и время следующего старта, вычитаем из нее секунды периодичности (2)
        """
        until_test = self._until_next_test(dtr1).total_seconds()
        if dtr2 == 4:
            period = datetime.timedelta(days=self._test_period(dtr2)).total_seconds()
        else:
            period = datetime.timedelta(weeks=self._test_period(dtr2)).total_seconds()

        if period > 0:
            return (datetime.datetime.now() +
                    datetime.timedelta(seconds=until_test) - datetime.timedelta(seconds=period)).astimezone().isoformat()
        else:
            return None

    def _last_test_status(self, b) -> bool:
        value = self.manager.send_interface.send(em.QueryEmergencyStatus(self.short_addr)).value.as_integer
        return bool(value & b)

    def _publish_test_info(self, funit_type, last, result, duration=None, rated_duration=None, sig=None):
        if funit_type == LastDurationTestInfo:
            d = {"timestamp": last, "duration": duration, "ratedDuration": rated_duration, "result": result}
        else:
            d = {"timestamp": last, "result": result}

        if d:
            self.on_update(funit_type, d, sig=sig)

    def function_test_info(self, sig=None):
        """ Заполняет топик информации о тесте функциональности """
        last_dt = None
        result = None
        try:
            last_dt = self._last_test_datetime(0, 4)
            result = self._last_test_status(0x02)
        except BaseException as ex:
            logging.error(ex)
        finally:
            self._publish_test_info(LastFunctionTestInfo, last_dt, result, duration=None, sig=sig)

    def duration_test_info(self, sig=None):
        """ Заполняет топик информации о тесте аккумулятора """
        last_dt = None
        result = None
        duration = None
        rated_duration = None
        try:
            rated_duration = self.get_value(RatedDuration, None)

            last_dt = self._last_test_datetime(2, 5)
            result = self._last_test_status(0x04)

            duration = self.manager.send_interface.send(em.QueryDurationTestResult(self.short_addr)).value.as_integer
            duration = int(datetime.timedelta(seconds=duration*120).total_seconds())

            result = result and duration >= rated_duration
        except BaseException as ex:
            logging.error(ex)
        finally:
            self._publish_test_info(LastDurationTestInfo, last_dt, result,
                                    duration=duration, rated_duration=rated_duration, sig=sig)

    def StartFunctionTest(self, value, *args, sig=None):
        """ Запускает тест функциональности """
        try:
            if value:
                if self.standard:
                    # стандартная последовательность команд запуска теста
                    period = self._test_period(4)
                    if period is None or period <= 0:
                        period = 7  # дней

                    self.manager.send_interface.send(em.StopTest(self.short_addr))

                    # сначала выключаем автоматическое тестирование (период = 0)
                    self.manager.send_interface.send(gear.DTR0(0))
                    self.manager.send_interface.send(em.StoreFunctionTestInterval(self.short_addr))

                    # зануляем TEST DELAY TIME, чтобы исключить отложенный запуск теста (pending test)
                    self.set_test_delay_time(0)

                    # включаем, чтобы тест гарантировано запустился
                    self.manager.send_interface.send(gear.DTR0(period))
                    self.manager.send_interface.send(em.StoreFunctionTestInterval(self.short_addr))

                    # self.manager.send_interface.send(em.StartFunctionTest)
                    self.query_emergency_function_test_timing()
                else:
                    # просто запускаем тест
                    self.manager.send_interface.send(em.StartFunctionTest(self.short_addr))
                    self.on_update(LastFunctionTestInfo,
                                   {'timestamp': datetime.datetime.now().astimezone().isoformat(), 'result': False})
            else:
                self.manager.send_interface.send(em.StopTest(self.short_addr))
                if self.standard:
                    self.query_emergency_function_test_timing()
                    self.query_emergency_duration_test_timing()
        except BaseException as ex:
            logging.error(ex)
        finally:
            failure_status = self.manager.send_interface.send(em.QueryEmergencyFailureStatus(self.short_addr))
            self.on_update(FailureStatus, failure_status.status_lower_case if failure_status else None, sig=sig)
            self.build_mode_topic(*args, sig=sig)

    def StartDurationTest(self, value, *args, sig=None):
        '''Запускает тест аккумулятора (продолжительности)'''
        try:
            if value:
                if self.standard:
                    # стандартная последовательность команд запуска теста
                    period = self._test_period(5)
                    if period is None or period <= 0:
                        period = 52 # недели

                    self.manager.send_interface.send(em.StopTest(self.short_addr))

                    # сначала выключаем автоматическое тестирование (период = 0)
                    self.manager.send_interface.send(gear.DTR0(0))
                    self.manager.send_interface.send(em.StoreDurationTestInterval(self.short_addr))

                    # зануляем TEST DELAY TIME, чтобы исключить отложенный запуск теста (pending test)
                    self.set_test_delay_time(0)

                    # включаем, чтобы тест гарантировано запустился
                    self.manager.send_interface.send(gear.DTR0(period))
                    self.manager.send_interface.send(em.StoreDurationTestInterval(self.short_addr))

                    # self.manager.send_interface.send(em.StartDurationTest)
                    self.query_emergency_duration_test_timing()
                else:
                    # просто запускаем тест
                    self.manager.send_interface.send(em.StartDurationTest(self.short_addr))
                    self.on_update(LastDurationTestInfo,
                                   {'timestamp': datetime.datetime.now().astimezone().isoformat(), 'result': False})
            else:
                self.manager.send_interface.send(em.StopTest(self.short_addr))
                if self.standard:
                    self.query_emergency_function_test_timing()
                    self.query_emergency_duration_test_timing()
        except BaseException as ex:
            logging.error(ex)
        finally:
            failure_status = self.manager.send_interface.send(em.QueryEmergencyFailureStatus(self.short_addr))
            self.on_update(FailureStatus, failure_status.status_lower_case if failure_status else None, sig=sig)
            self.build_mode_topic(*args, sig=sig)

    def build_mode_topic(self, *args, sig=None):
        """ Заполняет топик Mode на основе данных топика EmergencyStatus и топика EmergencyMode """
        mode = 'Unknown'
        emode = None
        estatus = None
        try:
            if self.standard:
                _mode = self.manager.send_interface.send(em.QueryEmergencyMode(self.short_addr)).info
            else:
                _mode = {'mode': None, 'value': 0}
            _status = self.manager.send_interface.send(em.QueryEmergencyStatus(self.short_addr)).info
            _em = _mode['value']
            _es = _status['value']
            emode = _mode['mode']
            estatus = _status['status']

            if utils.is_bit(_em, 0):
                mode = 'Rest'

            elif utils.is_bit(_em, 1):
                if utils.is_bit(_es, 0):
                    mode = 'InhibitMode'
                else:
                    mode = 'Normal'

            elif utils.is_bit(_em, 2):
                if not utils.is_bit(_es, 0):
                    mode = 'Emergency'

            elif utils.is_bit(_em, 3):
                if not utils.is_bit(_es, 0):
                    mode = 'ExtendedEmergency'

            elif utils.is_bit(_em, 4):
                if not utils.is_bits(_es, [0, 1, 4]):
                    mode = 'FunctionTest'

            elif utils.is_bit(_em, 5):
                if not utils.is_bits(_es, [0, 2, 5]):
                    mode = 'DurationTest'

        except BaseException as ex:
            logging.error(ex)
        finally:
            self.on_update(Mode, mode, sig=sig)
            self.on_update(EmergencyMode, emode, sig=sig)
            self.on_update(EmergencyStatus, estatus, sig=sig)

    def set_test_delay_time(self, delay: int):
        low, high = int(delay).to_bytes(2, 'little')

        self.manager.send_interface.send(gear.DTR0(high))
        self.manager.send_interface.send(em.StoreTestDelayTimeHighByte(self.short_addr))
        self.manager.send_interface.send(gear.DTR0(low))
        self.manager.send_interface.send(em.StoreTestDelayTimeLowByte(self.short_addr))


class DaliLightFunit(DaliFunit):
    entity_type = DaliLight
    _dtr0 = gear.DTR0
    _dtr1 = gear.DTR1
    _dtr2 = gear.DTR2

    @property
    def entity(self) -> DaliLight:
        return self._entity

    def _prepare_cmd(self, provider: DaliLight, cmd, *args):
        if cmd._hasparam:
            return cmd(provider.binding, args[0])
        else:
            return cmd(provider.binding)


DaliLight.funit_class = DaliLightFunit


class BrightnessLevel(DaliLightFunit):
    set_func = DaliLight.SetBrightnessLevel
    get_func = DaliLight.GetBrightnessLevel


class Types(DaliLightFunit):
    part_funit = 'NextType'
    get_func = DaliLight.GetTypes

    @classmethod
    def update_part(cls, entity: DaliLight, value: int) -> (DaliFunit, str):
        v = entity.get_value(cls, [])
        v = list(v) if isinstance(v, list) else []

        if value != 254 and value not in v:
            v.append(value)

        if value == 254 and 255 in v and len(v) > 1:
            v.remove(255)

        return cls, sorted(v)


class CurrentLevelRaw(DaliLightFunit):
    get_after_set = False
    set_cmd = gear.DAPC
    get_cmd = gear.QueryActualLevel


class PhysicalMinLevelRaw(DaliLightFunit):
    get_cmd = gear.QueryPhysicalMinimum


class MinLevelRaw(DaliLightFunit):
    set_cmd = gear.SetMinLevel
    get_cmd = gear.QueryMinLevel

    def set(self):
        super(MinLevelRaw, self).set()
        self.entity.query_level()


class MaxLevelRaw(DaliLightFunit):
    set_cmd = gear.SetMaxLevel
    get_cmd = gear.QueryMaxLevel

    def set(self):
        super(MaxLevelRaw, self).set()
        self.entity.query_level()


class PowerOnLevelRaw(DaliLightFunit):
    set_cmd = gear.SetPowerOnLevel
    get_cmd = gear.QueryPowerOnLevel


class SystemFailureLevelRaw(DaliLightFunit):
    set_cmd = gear.SetSystemFailureLevel
    get_cmd = gear.QuerySystemFailureLevel


class SceneLevelsRaw(DaliLightFunit):
    length = 16
    set_func = DaliLight.SetSceneLevelsRaw
    get_func = DaliLight.GetSceneLevelsRaw

    @staticmethod
    def def_value() -> list:
        return [{'index': i, 'value': 255} for i in range(0, 16)]


class FadeTime(DaliLightFunit):
    values = [0, 0.7, 1, 1.4, 2, 2.8, 4, 5.7, 8, 11.3, 16, 22.6, 32, 45.3, 64, 90.5]
    index_of = ["ft0", "ft7", "ft10", "ft14", "ft20", "ft28", "ft40", "ft57", "ft80", "ft113", "ft160", "ft226",
                "ft320", "ft453", "ft640", "ft905"]
    set_cmd = gear.SetFadeTime
    get_cmd = gear.QueryFadeTimeFadeRate


class FadeRate(DaliLightFunit):
    values = [0, 358, 253, 179, 127, 89.4, 63.3, 44.7, 31.6, 22.4, 15.8, 11.2, 7.9, 5.6, 4, 2.8]
    index_of = ["", "fr3580", "fr2530", "fr1790", "fr1270", "fr894", "fr633", "fr447", "fr316", "fr224", "fr158",
                "fr112", "fr79", "fr56", "fr40", "fr28"]
    set_cmd = gear.SetFadeRate
    get_cmd = gear.QueryFadeTimeFadeRate


class DimmingCurve(DaliLightFunit):
    LOGARITHMIC = 'Logarithmic'
    LINEAR = 'Linear'
    index_of = [LOGARITHMIC, LINEAR]
    set_cmd = led.SelectDimmingCurve
    get_cmd = led.QueryDimmingCurve


class GroupLevel(DaliLightFunit):
    set_func = DaliLight.SetGroupLevel


class Tuning(DaliLightFunit):
    entity_type = DaliLight
    set_func = entity_type.SetTuning


class On(BrightnessLevel):
    entity_type = DaliRelay

    @property
    def set_value(self): return 100 if super(On, self).set_value else 0


class Temperature(DaliLightFunit):
    entity_type = DaliTunableWhite
    set_func = entity_type.SetTemperature
    get_func = entity_type.GetTemperature

    @staticmethod
    def mirek2kelvin(mirek: int):
        return int(1000000 / mirek)

    @staticmethod
    def kelvin2mirek(kelvin: int):
        return int(1000000 / kelvin)

    @classmethod
    def apply_extended(cls, entity, dtr0, dtr1, sig=None):
        if dtr0 is not None and dtr1 is not None:
            mirek = int.from_bytes(bytes([dtr0, dtr1]), 'little')
            temperature = cls.mirek2kelvin(mirek)
            entity.update_recursive(Temperature, temperature, **{Variable.PRELIM: True}, sig=sig)
            entity.query_temperature()


class _TemperatureLimitFunit(DaliLightFunit):
    entity_type = DaliTunableWhite

    @classmethod
    def apply_extended(cls, entity, dtr0, dtr1, dtr2):
        if dtr2 == cc.StoreColourTemperatureLimit.COOLEST:
            limit_funit = Coolest
        elif dtr2 == cc.StoreColourTemperatureLimit.WARMEST:
            limit_funit = Warmest
        else:
            return

        mirek = int.from_bytes(bytes([dtr0, dtr1]), 'little')
        temperature = Temperature.mirek2kelvin(mirek)
        entity.update_recursive(limit_funit, temperature, **{Variable.PRELIM: True})
        entity.manager.put_to_stack(limit_funit(entity, Variable(None)).get)


class Warmest(DaliLightFunit):
    entity_type = DaliTunableWhite
    set_func = entity_type.SetWarmest
    get_func = entity_type.GetWarmest


class Coolest(DaliLightFunit):
    entity_type = DaliTunableWhite
    set_func = entity_type.SetCoolest
    get_func = entity_type.GetCoolest


class RgbLightFunit(DaliLightFunit):
    entity_type = DaliRgbwaf


class ChannelsCount(RgbLightFunit):
    get_cmd = cc.QueryColourTypeFeatures


class Color(RgbLightFunit):
    get_func = DaliRgbwaf.GetColor
    set_func = DaliRgbwaf.SetColor

    RGB_KEYS = [C.RED, C.GREEN, C.BLUE]
    WAF_KEYS = [C.WHITE, C.AMBER, C.FREE_COLOR]

    @classmethod
    def from_raw(cls, raw_value):
        if raw_value != 255:
            return round(raw_value * 10000 / 254)

        return None

    @classmethod
    def to_raw(cls, level):
        if level is None:
            return 255
        elif not 0 <= level <= 10000:
            raise ValueError('Components level expected as [0..10000]: %s' % level)

        return round(level * 254 / 10000)

    def get_components(self, channel: list):
        keys = set(channel).intersection(self._variable.value.keys())
        if keys:
            return [self.to_raw(self._variable.value.get(key)) for key in channel]

        return None

    @classmethod
    def def_value(cls, channels_count: int):
        return dict.fromkeys(list(cls.RGB_KEYS + cls.WAF_KEYS)[:channels_count])

    @property
    def set_value(self):
        if self._set_value is None:
            rgb = self.get_components(self.RGB_KEYS)
            waf = self.get_components(self.WAF_KEYS)

            if rgb is None and waf is None:
                raise ValueError('RGB and WAF is absent')

            self._set_value = rgb, waf

        return self._set_value

    @staticmethod
    def apply_extended(entity, cmd, dtrs, sig=None):
        value = dict(entity.get_value(Color) or Color.def_value(entity.get_value(ChannelsCount, 6)))

        components = Color.RGB_KEYS if isinstance(cmd, cc.SetRgbColour) else Color.WAF_KEYS
        for color_name, raw_value in zip(components, dtrs):
            level = Color.from_raw(raw_value)
            if level is not None:
                value[color_name] = level

        entity.on_update(Color, value, sig=sig, preliminary=True)
        entity.query_color()


class _Step(DaliLightFunit):
    def set(self):
        self.entity.request(self.set_value)


class _EmergencyFunit(DaliLightFunit):

    entity_type = DaliEmergency


class EmergencyMode(_EmergencyFunit):
    get_cmd = em.QueryEmergencyMode

    FUNCTION_TEST = 'FunctionTest'
    DURATION_TEST = 'DurationTest'
    INHIBIT_MODE = 'Inhibit'
    RESET_INHIBIT = 'ResetInhibit'

    variants = [FUNCTION_TEST, DURATION_TEST, INHIBIT_MODE, RESET_INHIBIT]

    def set(self):
        value = self.set_value
        if value == self.FUNCTION_TEST:
            self.set_cmd = em.StartFunctionTest
        elif value == self.DURATION_TEST:
            self.set_cmd = em.StartDurationTest
        elif value == self.INHIBIT_MODE:
            self.set_cmd = em.Inhibit
        else:
            self.set_cmd = em.ReLightResetInhibit

        if value == self.INHIBIT_MODE or value == self.RESET_INHIBIT:
            # QueryEmergencyStatus allows to find out if Inhibit mode is active or not
            self.get_cmd = em.QueryEmergencyStatus

        super(EmergencyMode, self).set()


class FunctionTest(_EmergencyFunit):
    # def set(self):
    #     self.get_after_set = True
    #     value = self.set_value
    #     if value:
    #         self.set_cmd = em.StartFunctionTest
    #         self.get_cmd = em.QueryEmergencyMode
    #     else:
    #         self.set_cmd = em.StopTest
    #         self.get_cmd = em.QueryEmergencyStatus
    #
    #     super(FunctionTest, self).set()
    set_func = DaliEmergency.StartFunctionTest


class DurationTest(_EmergencyFunit):
    # def set(self):
    #     self.get_after_set = True
    #     value = self.set_value
    #     if value:
    #         self.set_cmd = em.StartDurationTest
    #         self.get_cmd = em.QueryEmergencyMode
    #     else:
    #         self.set_cmd = em.StopTest
    #         self.get_cmd = em.QueryEmergencyStatus
    #
    #     super(DurationTest, self).set()
    set_func = DaliEmergency.StartDurationTest


class ChargeLevelRaw(_EmergencyFunit):
    get_cmd = em.QueryBatteryCharge


class ChargeLevel(_EmergencyFunit):
    get_cmd = em.QueryBatteryCharge

    @classmethod
    def of_raw(cls, raw_value):
        # if raw_value is None:
        #     return None
        #
        # if raw_value == 255:
        #     return 'NULL'   # special case to set 'null' in topic
        if raw_value is None or raw_value == 255:
            return None

        if raw_value == 0:
            return 0

        return round(10000 * raw_value / 254)


class DurationTestResult(_EmergencyFunit):
    get_cmd = em.QueryDurationTestResult


class FailureStatus(_EmergencyFunit):
    get_cmd = em.QueryEmergencyFailureStatus


class EmergencyStatus(_EmergencyFunit):
    get_cmd = em.QueryEmergencyStatus


class _TestPeriod(_EmergencyFunit):
    get_cmd = em.QueryTestTiming

    DTR_VAL = None

    def get(self):
        for cmd in self._apply_cmd(self.entity, self.get_cmd, self.DTR_VAL):
            self.entity.request(cmd, sig=self.signature)

    def set(self):
        # Если периодичность теста не 0, включаем автотестирование
        # Первый запуск теста откладываем на время, равное периодичности тестирования
        # Для этого вычисляем TEST DELAY TIME в 15-минутных интервалах
        if self.set_value != 0:
            if isinstance(self, DaysBetweenFunctionTest):
                _delay = self.set_value * 24 * 4
            elif isinstance(self, WeeksBetweenDurationTest):
                _delay = self.set_value * 7 * 24 * 4
            else:
                _delay = 0
        else:
            _delay = 0

        if _delay:
            getattr(self.entity, DaliEmergency.set_test_delay_time.__name__)(_delay)

        # Установка периодичности теста
        super().set()

        if _delay:
            getattr(self.entity, DaliEmergency.set_test_delay_time.__name__)(0)

        # Вызов функции получения нового срока до следующего теста
        if isinstance(self, DaysBetweenFunctionTest):
            func_name = DaliEmergency.query_emergency_function_test_timing
        elif isinstance(self, WeeksBetweenDurationTest):
            func_name = DaliEmergency.query_emergency_duration_test_timing
        else:
            return

        for _provider in self.entity.get_members():
            if isinstance(_provider, self.entity_type):
                assert func_name
                getattr(_provider, func_name.__name__)(**self.params, sig=self.signature)


class DaysBetweenFunctionTest(_TestPeriod):
    value_range = range(255)
    set_cmd = em.StoreFunctionTestInterval
    DTR_VAL = 4


class WeeksBetweenDurationTest(_TestPeriod):
    value_range = range(98)
    set_cmd = em.StoreDurationTestInterval
    DTR_VAL = 5


class UntilNextFunctionTest(_EmergencyFunit):
    set_cmd = None
    DTR_VAL = 0

    get_func = DaliEmergency.query_emergency_function_test_timing


class UntilNextDurationTest(_EmergencyFunit):
    set_cmd = None
    DTR_VAL = 2

    get_func = DaliEmergency.query_emergency_duration_test_timing


class LastFunctionTestInfo(_EmergencyFunit):
    set_cmd = None
    get_cmd = None
    get_func = DaliEmergency.function_test_info


class LastDurationTestInfo(_EmergencyFunit):
    set_cmd = None
    get_cmd = None
    get_func = DaliEmergency.duration_test_info


class EmergencyLevelRaw(_EmergencyFunit):
    get_cmd = em.QueryEmergencyLevel


class EmergencyMaxLevelRaw(_EmergencyFunit):
    get_cmd = em.QueryEmergencyMaxLevel


class EmergencyMinLevelRaw(_EmergencyFunit):
    get_cmd = em.QueryEmergencyMinLevel


class Mode(_EmergencyFunit):
    """Топик, содержащий данные топика EmergencyStatus и два флага топика EmergencyMode"""
    set_cmd = None
    get_cmd = None
    get_func = DaliEmergency.build_mode_topic


class RatedDuration(_EmergencyFunit):
    get_cmd = em.QueryRatedDuration
