import struct

import spread_core.tools.settings_const as const
import spread_dali as dali
from spread_core.mqtt.spread.variable import Variable
from spread_core.tools.timer import Timer
from spread_dali.device import general as device, push_button_ext as device1, presence_sensor_ext as device3, \
    light_sensor_ext as device4
from .providers import DaliProvider, DaliFunit, Groups, GtinOem, Gtin, Serial, SerialOem, HardwareVersion, \
    FirmwareVersion, Validity, Discovery


class DaliSensor(DaliProvider):
    _type_id = None
    _dali_type = 2
    _query_event_on_info = True

    def __init__(self, id: int, binding: dali.address.Address, instance: dali.address.Instance = None, **kwargs):
        super(DaliSensor, self).__init__(id, binding, **kwargs)
        self._instance = instance or dali.address.InstanceType(self.type_id)

    @property
    def rank(self):
        return self.dali_type, 32 - self._type_id, self.id

    @property
    def instance(self):
        return self._instance

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == On:
            return False

        return super(DaliSensor, self).is_repaired_funit(funit_type)

    def compare_instance(self, instance: dali.address.Instance):
        if isinstance(instance, dali.address.InstanceType):
            return self.type_id == instance.value
        elif isinstance(instance, dali.address.InstanceNumber):
            return self.get_value(InstanceIndex) == instance.value or self._instance == instance

        return isinstance(instance, dali.address.InstanceBroadcast)

    @property
    def type_id(self):
        return self._type_id

    def check_valid(self):
        super(DaliSensor, self).check_valid()
        if self.is_device:
            self.request(device.QueryInstanceStatus(self.short_addr, self.instance))

    def on_survey(self):
        yield from super(DaliSensor, self).on_survey()

        if self.is_valid_strong: yield self.request(device.QueryDeviceGroupsZeroToSeven(self.short_addr))
        if self.is_valid_strong: yield self.request(device.QueryDeviceGroupsEightToFifteen(self.short_addr))
        if self.is_valid_strong: yield self.request(device.QueryDeviceGroupsSixteenToTwentyThree(self.short_addr))
        if self.is_valid_strong: yield self.request(device.QueryDeviceGroupsTwentyFourToThirtyOne(self.short_addr))
        if self.is_valid_strong: yield self.request(device.QueryInstanceEnabled(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device.QueryNumberOfInstances(self.short_addr))
        if self.is_valid_strong: yield self.request(device.QueryEventScheme(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device.QueryResolution(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device.QueryEventPriority(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device.QueryEventFilterL(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device.QueryOperatingMode(self.short_addr))
        if self.is_valid_strong: yield self.request(device.QueryPrimaryInstanceGroup(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device.QueryInstanceGroup1(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device.QueryInstanceGroup2(self.short_addr, self.instance))

    def on_external(self, command: dali.command.Command, response, *args):
        if isinstance(command, dali.device.general._StandardInstanceCommand):
            if not self.compare_instance(command.instance):
                return

        super(DaliSensor, self).on_external(command, response, *args)

    def get_info(self):
        super(DaliSensor, self).get_info()
        self.on_update(InstanceType, self.type_id)

        if self.is_device:
            self.request(device.QueryInstanceEnabled(self.short_addr, self.instance))

            if InstancesNumber not in self.values:
                self.request(device.QueryNumberOfInstances(self.short_addr))
            if InstanceIndex not in self.values:
                self.GetInstanceIndex()
            if EventScheme not in self.values:
                self.request(device.QueryEventScheme(self.short_addr, self.instance))
            if self.get_value(Resolution, 0) == 0:
                self.request(device.QueryResolution(self.short_addr, self.instance))
            if self._query_event_on_info and self.get_value(On, False):
                self.get_event_value()
            if EventPriority not in self.values:
                self.request(device.QueryEventPriority(self.short_addr, self.instance))
            if EventFilter not in self.values:
                self.request(device.QueryEventFilterL(self.short_addr, self.instance))
            if OperationMode not in self.values:
                self.request(device.QueryOperatingMode(self.short_addr))
            if Group0 not in self.values:
                self.request(device.QueryPrimaryInstanceGroup(self.short_addr, self.instance))
            if Group1 not in self.values:
                self.request(device.QueryInstanceGroup1(self.short_addr, self.instance))
            if Group2 not in self.values:
                self.request(device.QueryInstanceGroup2(self.short_addr, self.instance))
            if FeatureTypes not in self.values:
                self.fill_feature_types()
        else:
            self.on_update(On, False)

            for funit_type in (InstancesNumber, InstanceIndex, EventScheme, Resolution, OperationMode,
                               Group0, Group1, Group2, EventPriority):
                self.on_update(funit_type, 0)

            for funit_type in (EventScheme, FeatureTypes):
                self.on_update(funit_type, [])

    def get_event_value(self, sig=None):
        resolution = self.get_value(Resolution, 0)
        _val = None

        if resolution > 0:
            r = self.manager.send_interface.send(device.QueryInputValue(self.short_addr, self.instance))
            if r:
                _val = r.value.as_integer
                if resolution > 8:
                    r = self.manager.send_interface.send(device.QueryInputValueLatch(self.short_addr, self.instance))
                    if r:
                        _val = (_val << 2) + (r.value.as_integer >> 6)
                else:
                    _val = _val >> (8 - resolution)

        return _val

    def fill_feature_types(self, sig=None):
        self.check_command_for_device()
        resp = self.manager.send_interface.send(device.QueryFeatureType(self.short_addr, self.instance))
        _arr = []
        if resp:
            if resp.value.as_integer == 0xff:
                _r = None
                while _r is None or _r.value.as_integer == 0xfe or _r.value.as_integer == 0xff:
                    _r = self.request(device.QueryNextFeatureType(self.short_addr, self.instance))
                    if _r is None or _r.value.as_integer == 0xfe:
                        break
                    _arr.append(_r.value.as_integer)
            elif resp.value.as_integer != 0xfe:
                _arr = [resp.value.as_integer]

        self.on_update(FeatureTypes, _arr, sig=sig)

    def on_command_sended(self, cmd):
        if isinstance(cmd, (device.AddToDeviceGroupsZeroToFifteen,
                            device.AddToDeviceGroupsSixteenToThirtyOne,
                            device.RemoveFromDeviceGroupsZeroToFifteen,
                            device.RemoveFromDeviceGroupsSixteenToThirtyOne)):
            start_index = 0 if isinstance(cmd, (device.AddToDeviceGroupsZeroToFifteen,
                                                device.RemoveFromDeviceGroupsZeroToFifteen)) else 16

            adding = isinstance(cmd, (device.AddToDeviceGroupsZeroToFifteen,
                                      device.AddToDeviceGroupsSixteenToThirtyOne))
            raw_value = self.dtr(1) + (self.dtr(2) << 8)

            value = Groups.def_value(self._dali_type)
            value = [{**i} for i in self.get_value(Groups, value)]
            for i in range(0, 16):
                if 1 << i & raw_value:
                    value[start_index+i]['value'] = adding
            self.update_recursive(Groups, value, **{Variable.PRELIM: True})
        elif isinstance(cmd, (device.EnableInstance, device.DisableInstance)):
            self.update_recursive(On, isinstance(cmd, device.EnableInstance), **{Variable.PRELIM: True})
        elif isinstance(cmd, device.SetEventScheme):
            self.update_recursive(EventScheme, EventScheme.of_raw(self.dtr(0)), **{Variable.PRELIM: True})
        elif isinstance(cmd, device.SetEventPriority):
            self.update_recursive(EventPriority, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, device.SetEventFilter):
            value = 0
            for i in range(0, 3):
                value += (self.dtr(i) >> i)
            mask = EventFilter.FILTERS.get(self.type_id, [])
            self.update_recursive(EventFilter, [mask[i] for i in range(0, 8) if value & 2 ** i != 0], **{Variable.PRELIM: True})
        else:
            super(DaliSensor, self).on_command_sended(cmd)

    def on_request_answer(self, funit_type, response, *args, sig=None):
        if isinstance(response, device.QueryInstanceStatusResponse):
            self.update_recursive(Validity, not response.error, sig=sig)
        elif funit_type == EventFilter and response:
            mask = EventFilter.FILTERS.get(self.type_id, [])
            self.update_recursive(funit_type, [mask[i] for i in range(0, 8) if response.value.as_integer & 2 ** i != 0], sig=sig)
        elif funit_type == HoldTime:
            self.update_recursive(funit_type, response.value.as_integer if response else None, sig=sig)
            self.update_recursive(EngineryHoldTime, (response.value.as_integer * 10) if response else None)
        else:
            super(DaliSensor, self).on_request_answer(funit_type, response, *args, sig=sig)

    def on_update(self, funit_type, value, sig=None, **kwargs):
        super(DaliSensor, self).on_update(funit_type, value, sig, **kwargs)

        if funit_type in [InstancesNumber, Groups, OperationMode, Gtin, GtinOem, Serial, SerialOem, FirmwareVersion, HardwareVersion]:
            for _provider in self.manager.providers_by_address(self._dali_type, self.short_addr):
                if isinstance(_provider, DaliSensor):
                    if self.id != _provider.id:
                        if _provider.get_value(funit_type) != value:
                            _provider.on_update(funit_type, value, **kwargs)

    def on_event(self, event: dali.device.events.Event):
        raise NotImplementedError()

    def read_memory_location(self, bank_id, start_address, end_address) -> bytes:
        self.check_command_for_device()
        i = self.manager.send_interface

        i.send(device.DTR1(bank_id))

        for _dtr0 in range(start_address, end_address):
            i.send(device.DTR0(_dtr0))
            res = i.send(device.ReadMemoryLocation(self.short_addr))
            if isinstance(res, dali.command.Response):
                yield res.value.as_integer
            else:
                break

    def SetEnable(self, value, sig=None):
        if self.get_value('On') != value:
            if value is True:
                self.request(device.EnableInstance(self._binding, self.instance))
            else:
                self.request(device.DisableInstance(self._binding, self.instance))

            for _provider in self.get_members():
                cmd = device.QueryInstanceEnabled(_provider.short_addr, _provider.instance)
                cmd.sig = sig
                self.request(cmd)

    def SetDiscovery(self, value, sig=None):
        _discovery = self.get_value(Discovery, False)
        if _discovery != value:
            self.on_update(Discovery, value, sig=sig)

            if _discovery is False and value is True:
                self.check_discovery()
            else:
                self.request(device.Terminate())

    def check_discovery(self):
        if self.get_value(Discovery) is True:
            self.request(device.IdentifyDevice(self._binding))
            Timer.set(7, self.check_discovery)

    def GetGroups(self, _=None, of_0_7=True, of_8_15=True, of_16_23=True, of_24_31=True, sig=None):
        if self.is_device:
            if of_0_7:
                self.request(device.QueryDeviceGroupsZeroToSeven(self.short_addr))
            if of_8_15:
                self.request(device.QueryDeviceGroupsEightToFifteen(self.short_addr))
            if of_16_23:
                self.request(device.QueryDeviceGroupsSixteenToTwentyThree(self.short_addr))
            if of_24_31:
                self.request(device.QueryDeviceGroupsTwentyFourToThirtyOne(self.short_addr))

        super(DaliSensor, self).GetGroups(sig=sig)

    def SetGroups(self, val, sig=None):
        dtrs = [0, 0, 0, 0]
        has_0_7 = False
        has_8_15 = False
        has_16_23 = False
        has_24_31 = False

        for item in val:
            index = item['index']
            if item['value']:
                dtrs[int(index / 16)] += 1 << (index % 16)
            else:
                dtrs[2+int(index / 16)] += 1 << (index % 16)

            if index < 8:
                has_0_7 = True
            elif index < 16:
                has_8_15 = True
            elif index < 24:
                has_16_23 = True
            else:
                # index < 16
                has_24_31 = True

        for i in range(0, 4):
            if dtrs[i]:
                _dtr1, _dtr2 = struct.pack('<H', dtrs[i])
                self.request(device.DTR2DTR1(_dtr2, _dtr1))
                self.request(
                    (
                        device.AddToDeviceGroupsZeroToFifteen,
                        device.AddToDeviceGroupsSixteenToThirtyOne,
                        device.RemoveFromDeviceGroupsZeroToFifteen,
                        device.RemoveFromDeviceGroupsSixteenToThirtyOne
                    )[i](self._binding)
                )

        if self.is_device:
            self.GetGroups(of_0_7=has_0_7, of_8_15=has_8_15, of_16_23=has_16_23, of_24_31=has_24_31, sig=sig)
        else:
            self.update_recursive(Groups, val)

    def GetInstanceIndex(self, sig=None):
        self.check_command_for_device()
        val = None
        if isinstance(self.instance, dali.address.InstanceType):
            if InstancesNumber in self.values:
                for instance_num in range(0, self.values[InstancesNumber]):
                    tp = self.manager.send_interface.send(
                        device.QueryInstanceType(self.short_addr, dali.address.InstanceNumber(instance_num)))
                    if tp:
                        if tp.value.as_integer == self.instance._value:
                            val = instance_num
                            break
        elif isinstance(self.instance, dali.address.InstanceNumber):
            val = self.instance._value

        self.on_update(InstanceIndex, self.get_value(InstanceIndex, val), sig=sig)


class DaliLightSensor(DaliSensor):
    _type_id = 4

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == CurrentLuminosity:
            return False

        return super(DaliLightSensor, self).is_repaired_funit(funit_type)

    def on_event(self, event: dali.device.events.Event):
        self.on_update(CurrentLuminosity, event.value)

    def get_info(self):
        super(DaliLightSensor, self).get_info()
        if self.is_device:
            if DeadTime not in self.values:
                self.request(device4.QueryDeadTime(self.short_addr, self.instance))
            if ReportTime not in self.values:
                self.request(device4.QueryReportTime(self.short_addr, self.instance))
            if Hysteresis not in self.values:
                self.request(device4.QueryHysteresis(self.short_addr, self.instance))
            if HysteresisMin not in self.values:
                self.request(device4.QueryHysteresisMin(self.short_addr, self.instance))
        else:
            for funit_type in (DeadTime, ReportTime, Hysteresis, HysteresisMin):
                self.on_update(funit_type, 0)

    def on_command_sended(self, cmd):
        if isinstance(cmd, device4.SetHysteresis):
            self.update_recursive(Hysteresis, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, device4.SetHysteresisMin):
            self.update_recursive(HysteresisMin, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, device4.SetDeadTime):
            self.update_recursive(DeadTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, device4.SetReportTime):
            self.update_recursive(ReportTime, self.dtr(0), **{Variable.PRELIM: True})
        else:
            super(DaliLightSensor, self).on_command_sended(cmd)
                
    def on_survey(self):
        yield from super(DaliLightSensor, self).on_survey()
        
        if self.is_valid_strong: yield self.request(device4.QueryDeadTime(self.short_addr, self.instance))
        if self.is_valid_strong: yield device4.QueryReportTime(self.short_addr, self.instance)
        if self.is_valid_strong: yield device4.QueryHysteresis(self.short_addr, self.instance)
        if self.is_valid_strong: yield device4.QueryHysteresisMin(self.short_addr, self.instance)

    def get_event_value(self, sig=None):
        self.on_update(CurrentLuminosity, super().get_event_value(), sig=sig)


class DaliPresenceSensor(DaliSensor):
    _type_id = 3

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == CurrentPresence:
            return False

        return super(DaliPresenceSensor, self).is_repaired_funit(funit_type)

    def on_event(self, event: dali.device.events.Event):
        self.on_update(CurrentPresence, dali.device.events.PresenceEvent(event.value).value)

    def get_info(self):
        super(DaliPresenceSensor, self).get_info()
        if self.is_device:
            if DeadTime not in self.values:
                self.request(device3.QueryDeadTime(self.short_addr, self.instance))
            if ReportTime not in self.values:
                self.request(device3.QueryReportTime(self.short_addr, self.instance))
            if HoldTime not in self.values:
                self.request(device3.QueryHoldTime(self.short_addr, self.instance))
        else:
            for funit_type in (DeadTime, ReportTime, HoldTime):
                self.on_update(funit_type, 0)

    def on_survey(self):
        yield from super(DaliPresenceSensor, self).on_survey()

        if self.is_valid_strong: yield self.request(device3.QueryDeadTime(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device3.QueryReportTime(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device3.QueryHoldTime(self.short_addr, self.instance))

    def get_event_value(self, sig=None):
        value = super().get_event_value(sig)
        if value is not None:
            value = dali.device.events.PresenceEvent(value).value

        self.on_update(CurrentPresence, value, sig=sig)

    def on_command_sended(self, cmd):
        if isinstance(cmd, device3.SetHoldTime):
            self.update_recursive(HoldTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, device3.SetDeadTime):
            self.update_recursive(DeadTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, device3.SetReportTime):
            self.update_recursive(ReportTime, self.dtr(0), **{Variable.PRELIM: True})
        else:
            super(DaliPresenceSensor, self).on_command_sended(cmd)


class DaliPushButton(DaliSensor):
    _type_id = 1
    _query_event_on_info = False

    def __init__(self, id: int, binding: dali.address.Address, instance: dali.address.InstanceNumber, **kwargs):
        super(DaliPushButton, self).__init__(id, binding, instance, **kwargs)

    @staticmethod
    def parse_args(attributes: dict, **kwargs) -> dict:
        return dict(
            **DaliSensor.parse_args(attributes, **kwargs),
            instance=dali.address.InstanceNumber(attributes[const.INSTANCE_INDEX])
        )

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == LastEvent:
            return False

        return super(DaliPushButton, self).is_repaired_funit(funit_type)

    def on_event(self, event: dali.device.events.Event):
        self.values.pop(LastEvent, None)
        self.on_update(LastEvent, dali.device.events.ButtonEvent(event.value).value)

    def get_info(self):
        super(DaliPushButton, self).get_info()
        if self.is_device:
            if ShortTimeMin not in self.values:
                self.request(device1.QueryShortTimeMin(self.short_addr, self.instance))
            if ShortTime not in self.values:
                self.request(device1.QueryShortTime(self.short_addr, self.instance))
            if DoubleTime not in self.values:
                self.request(device1.QueryDoubleTime(self.short_addr, self.instance))
            if DoubleTimeMin not in self.values:
                self.request(device1.QueryDoubleTimeMin(self.short_addr, self.instance))
            if RepeatTime not in self.values:
                self.request(device1.QueryRepeatTime(self.short_addr, self.instance))
            if StuckTime not in self.values:
                self.request(device1.QueryStuckTime(self.short_addr, self.instance))
        else:
            for funit_type in (ShortTimeMin, ShortTime, DoubleTime, DoubleTimeMin, RepeatTime, StuckTime):
                self.on_update(funit_type, 0)
    
    def on_survey(self):
        yield from super(DaliPushButton, self).on_survey()
        
        if self.is_valid_strong: yield self.request(device1.QueryShortTimeMin(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device1.QueryShortTime(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device1.QueryDoubleTime(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device1.QueryDoubleTimeMin(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device1.QueryRepeatTime(self.short_addr, self.instance))
        if self.is_valid_strong: yield self.request(device1.QueryStuckTime(self.short_addr, self.instance))

    def get_event_value(self, sig=None):
        value = super().get_event_value()
        if value is not None:
            value = dali.device.events.ButtonEvent(value).value

        self.on_update(LastEvent, value, sig=sig)

    def on_command_sended(self, cmd):
        if isinstance(cmd, device1.SetShortTime):
            self.update_recursive(ShortTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, device1.SetDoubleTime):
            self.update_recursive(DoubleTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, device1.SetDoubleTime):
            self.update_recursive(DoubleTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, device1.SetRepeatTime):
            self.update_recursive(RepeatTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, device1.SetStuckTime):
            self.update_recursive(StuckTime, self.dtr(0), **{Variable.PRELIM: True})
        else:
            super(DaliPushButton, self).on_command_sended(cmd)


class DaliSensorFunit(DaliFunit):
    entity_type = DaliSensor
    _dtr0 = device.DTR0
    _dtr1 = device.DTR1
    _dtr2 = device.DTR2

    @property
    def entity(self) -> DaliSensor:
        return self._entity

    def _prepare_cmd(self, provider: DaliSensor, cmd, *args):
        if issubclass(cmd, device._StandardInstanceCommand):
            return cmd(provider.short_addr, provider.instance, *args)
        elif issubclass(cmd, device._StandardDeviceCommand):
            return cmd(provider.short_addr, *args)
        else:
            return cmd(*args)


DaliSensor.funit_class = DaliSensorFunit


class On(DaliSensorFunit):
    get_after_set = False
    set_func = DaliSensor.SetEnable
    get_cmd = device.QueryInstanceEnabled


class OperationMode(DaliSensorFunit):
    set_cmd = device.SetOperatingMode
    get_cmd = device.QueryOperatingMode


class Group0(DaliSensorFunit):
    set_cmd = device.SetPrimaryInstanceGroup
    get_cmd = device.QueryPrimaryInstanceGroup


class Group1(DaliSensorFunit):
    set_cmd = device.SetInstanceGroup1
    get_cmd = device.QueryInstanceGroup1


class Group2(DaliSensorFunit):
    set_cmd = device.SetInstanceGroup2
    get_cmd = device.QueryInstanceGroup2


class EventScheme(DaliSensorFunit):
    set_cmd = device.SetEventScheme
    get_cmd = device.QueryEventScheme
    index_of = ["esIndexAndType", "esAddressAndType", "esAddressAndIndex", "esLowestGroupAndType", "esPrimaryGroupAndType"]


class EventPriority(DaliSensorFunit):
    set_cmd = device.SetEventPriority
    get_cmd = device.QueryEventPriority


class EventFilter(DaliSensorFunit):
    FILTERS = {
        1: ["buttonReleasedEvents", "buttonPressedEvents", "shortPressEvents", "doublePressEvents",
            "longPressStartEvents", "longPressRepeatEvents", "longPressStopEvents", "buttonStuckFreeEvents"],
        3: ["occupiedEvents", "vacantEvents", "repeatEvents", "movementEvents", "noMovementEvents"],
        4: ["illuminanceLevelEvents"],
    }
    set_cmd = device.SetEventFilter
    get_cmd = device.QueryEventFilterL
    @property
    def mask(self): return self.FILTERS.get(self.entity.type_id, [])

    @property
    def params(self) -> dict: return [0, 0]


class InstancesNumber(DaliSensorFunit):
    get_cmd = device.QueryNumberOfInstances


class InstanceIndex(DaliSensorFunit):
    get_func = DaliSensor.GetInstanceIndex


class InstanceType(DaliSensorFunit):
    pass


class FeatureTypes(DaliSensorFunit):
    get_func = DaliSensor.fill_feature_types


class Resolution(DaliSensorFunit):
    get_cmd = device.QueryResolution


class CurrentLuminosity(DaliSensorFunit):
    entity_type = DaliLightSensor
    get_func = DaliLightSensor.get_event_value


class CurrentPresence(DaliSensorFunit):
    entity_type = DaliPresenceSensor
    get_func = DaliPresenceSensor.get_event_value


class Hysteresis(DaliSensorFunit):
    entity_type = DaliLightSensor
    set_cmd = device4.SetHysteresis
    get_cmd = device4.QueryHysteresis


class HysteresisMin(DaliSensorFunit):
    entity_type = DaliLightSensor
    set_cmd = device4.SetHysteresisMin
    get_cmd = device4.QueryHysteresisMin


class HoldTime(DaliSensorFunit):
    entity_type = DaliPresenceSensor
    set_cmd = device3.SetHoldTime
    get_cmd = device3.QueryHoldTime


class EngineryHoldTime(DaliSensorFunit):
    entity_type = DaliPresenceSensor
    set_cmd = device3.SetHoldTime
    get_cmd = device3.QueryHoldTime
    value_range = range(0, 254*10+1)

    @property
    def set_value(self): return int(super(EngineryHoldTime, self).set_value / 10)


class DeadTime(DaliSensorFunit):
    @property
    def set_cmd(self):
        if isinstance(self.entity, DaliPresenceSensor):
            return device3.SetDeadTime
        elif isinstance(self.entity, DaliLightSensor):
            return device4.SetDeadTime

    @property
    def get_cmd(self):
        if isinstance(self.entity, DaliPresenceSensor):
            return device3.QueryDeadTime
        elif isinstance(self.entity, DaliLightSensor):
            return device4.QueryDeadTime


class ReportTime(DaliSensorFunit):
    @property
    def set_cmd(self):
        if isinstance(self.entity, DaliPresenceSensor):
            return device3.SetReportTime
        elif isinstance(self.entity, DaliLightSensor):
            return device4.SetReportTime

    @property
    def get_cmd(self):
        if isinstance(self.entity, DaliPresenceSensor):
            return device3.QueryReportTime
        elif isinstance(self.entity, DaliLightSensor):
            return device4.QueryReportTime


class _DaliPushButtonFunit(DaliSensorFunit):
    entity_type = DaliPushButton


class LastEvent(_DaliPushButtonFunit):
    retained = False
    get_func = DaliPushButton.get_event_value


class ShortTime(_DaliPushButtonFunit):
    value_range = range(10, 256)
    get_cmd = device1.QueryShortTime
    set_cmd = device1.SetShortTime


class ShortTimeMin(_DaliPushButtonFunit):
    get_cmd = device1.QueryShortTimeMin


class DoubleTime(_DaliPushButtonFunit):
    value_range = range(10, 101)
    get_cmd = device1.QueryDoubleTime
    set_cmd = device1.SetDoubleTime


class DoubleTimeMin(_DaliPushButtonFunit):
    get_cmd = device1.QueryDoubleTimeMin


class RepeatTime(_DaliPushButtonFunit):
    value_range = range(5, 101)
    get_cmd = device1.QueryRepeatTime
    set_cmd = device1.SetRepeatTime


class StuckTime(_DaliPushButtonFunit):
    value_range = range(5, 256)
    get_cmd = device1.QueryStuckTime
    set_cmd = device1.SetStuckTime
