from spread_core.bam.recipe import Recipe

from spread_core.mqtt.spread.topic import State, Set
from spread_core.mqtt.spread.variable import Variable
from .enginery import Enginery, _EngineryFunit
from .. import const as C


class Light(Enginery):
    def __init__(self, _id: int, power=0, **kwargs):
        super(Light, self).__init__(_id, **kwargs)
        self.power = power

    @staticmethod
    def parse_args(**kwargs) -> dict:
        return dict(
            power=int(kwargs.get('attributes', {}).get('power', 0))
        )

    @property
    def current_power(self):
        return int(self.power * self.get_value(BrightnessLevel, 0) / 100)

    def calc_on(self, variable):
        return None if variable.invalid else variable.value > 0

    def on_state(self, topic: State, variable: Variable):
        super(Light, self).on_state(topic, variable)

        if topic.entity_address.funit_type == BrightnessLevel:
            self.publish_state(On, self.calc_on(variable))
            self.on_level(variable)

    def on_level(self, variable):
        for _subginery in self._subgineries:
            _subginery.on_level_changed(
                eng=self,
                level=None if variable.invalid else variable.value,
                preliminary=variable.preliminary
            )

        if not variable.preliminary:
            power = None if variable.invalid else self.current_power
            self.publish_state(Power, power, delay=Power.delay)

    def set_on(self, value, sig=None):
        level = 100 if value else 0
        self.publish_provider_command(BrightnessLevel, level, sig=sig)


class SwitchingLight(Light):
    def on_state(self, topic: State, variable: Variable):
        if topic.entity_address.funit_type == BrightnessLevel:
            return

        super(Light, self).on_state(topic, variable)

        if topic.entity_address.funit_type == On:
            self.on_level(variable)

    def calc_on(self, variable):
        return None if variable.invalid else variable.value

    @property
    def current_power(self):
        return self.power * self.get_value(On, 0)

    def set_on(self, value, sig=None):
        self.publish_provider_command(On, value, sig=sig)


class DimmingLight(Light):
    pass


class TunableWhiteLight(Light):
    pass


class RgbLight(Light):
    components = [C.RED, C.GREEN, C.BLUE]

    def on_state(self, topic: State, variable: Variable):
        if topic.entity_address.funit_type == Color:
            if not variable.invalid:
                value = dict.fromkeys(self.components)
                for key in self.components:
                    component_value = variable.value.get(key, None)
                    if component_value is None:
                        value = None
                        break
                    else:
                        value[key] = component_value

            else:
                value = None

            variable = Variable(value)

        super(RgbLight, self).on_state(topic, variable)


class RgbwLight(RgbLight):
    components = RgbLight.components + [C.WHITE]


class RgbLightComplex(RgbLight):
    def __init__(self, _id: int, power=0, **kwargs):
        super(RgbLight, self).__init__(_id, power, **kwargs)
        self.entity_type = RgbLight.entity_type

        self._recipe = [None] * len(self.components)

    def set_recipe(self, ingredient_type, recipe: Recipe):
        try:
            recipe_index = self.components.index(ingredient_type.casefold())
            self._recipe[recipe_index] = recipe

        except (IndexError, ValueError):
            raise ValueError('unknown ingredient type: "%s"' % ingredient_type)

    def on_state(self, topic: State, variable: Variable):
        if topic.entity_address.funit_type == BrightnessLevel:
            for index, _ingredient in enumerate(self._recipe):
                if _ingredient.id == topic.entity_address.id:
                    if variable.invalid:
                        color = None
                    else:
                        color = dict(self.get_value(Color, {}))
                        color[self.components[index]] = Color.from_brightness(variable.value)

                        brightness = Color.to_brightness(max(color.values())) if color else None
                        super(RgbLightComplex, self).on_state(topic, Variable(brightness))

                        if len(color) < len(self.components):
                            self.set_value(Color, color)
                            return

                    self.publish_state(Color, color)
                    break

        else:
            super(RgbLightComplex, self).on_state(topic, variable)

    def publish_provider_command(self, funit_type, value, params=None, sig=None):
        if self.adapter:
            if funit_type == BrightnessLevel:
                if self.get_value(BrightnessLevel) == 0 and value:
                    params['method'] = 'GoToLastActiveLevel'

                levels = self._shift_brightnesses(value)
                self._set_brightnesses(levels, params=params, sig=sig)

            elif funit_type == Color:
                levels = map(lambda _component: Color.to_brightness(value[_component]), self.components)
                self._set_brightnesses(levels, params=params, sig=sig)

    def _set_brightnesses(self, levels, params=None, sig=None):
        for recipe, level in zip(self._recipe, levels):
            self.adapter.publish(
                Set(recipe.get_address(BrightnessLevel)),
                Variable(level, params=params, sig=sig)
            )

    def _shift_brightnesses(self, brightness: int):
        if brightness == 0:
            return [0] * len(self.components)

        color = self.get_value(Color, {})
        levels = tuple(map(Color.to_brightness, (color[component] for component in self.components)))

        diff = brightness - max(levels)

        if diff < 0:
            diff = 1 - min(levels)

        return map(lambda l: l + diff, levels)


class EmergencyUnit(Enginery):
    pass


class RgbwLightComplex(RgbLight):
    components = RgbLightComplex.components + [C.WHITE]


class DynamicLight(Light):
    pass


class _LightFunit(_EngineryFunit):
    entity_type = Light


Light.funit_class = _LightFunit


class On(_LightFunit):
    value_type = bool

    def set(self):
        self.entity.set_on(self.set_value, sig=self.signature)


class BrightnessLevel(_LightFunit):
    value_range = range(0, 101)


class Power(_LightFunit):
    delay = 0
    def set(self): raise self.NotImplemented()


class GroupLevel(_LightFunit):
    value_range = range(0, 101)


class GroupOn(_LightFunit):
    value_type = bool

    @property
    def set_value(self): return 100 if super(GroupOn, self).set_value else 0

    def set(self): self.entity.publish_provider_command(GroupLevel, self.set_value, sig=self.signature)


class Temperature(_LightFunit):
    value_range = range(1000, 20001)
    entity_type = TunableWhiteLight


class Color(_LightFunit):
    value_type = dict
    entity_type = RgbLight

    @property
    def set_value(self):
        value = super(Color, self).set_value

        unexpected = set(set(value.keys())).difference(self.entity.components)
        if unexpected:
            raise ValueError('Unexpected component: %s' % unexpected)

        for component in value:
            if not 0 <= value[component] <= 10000:
                raise ValueError('Component "%s" not in [0..10000]: %s' % (component, value[component]))

        return value

    @staticmethod
    def to_brightness(color: int):
        return round(color * 100 / 10_000)

    @staticmethod
    def from_brightness(brightness: int):
        return round(brightness * 10_000 / 100)


class _EmergencyFunit(_EngineryFunit):
    entity_type = EmergencyUnit


EmergencyUnit.funit_class = _EmergencyFunit


class ChargeLevel(_EmergencyFunit):
    nullable = True


class EmergencyMode(_EmergencyFunit):
    pass


class EmergencyStatus(_EmergencyFunit):
    pass


class FunctionTest(_EmergencyFunit):
    value_type = bool


class DurationTest(_EmergencyFunit):
    value_type = bool


class Mode(_EmergencyFunit):
    pass


class LastFunctionTestInfo(_EmergencyFunit):
    pass


class LastDurationTestInfo(_EmergencyFunit):
    pass


Light.settable_units = [On, BrightnessLevel, GroupLevel, GroupOn]
SwitchingLight.settable_units = [_ for _ in Light.settable_units if _ != BrightnessLevel]
TunableWhiteLight.settable_units = Light.settable_units + [Temperature]
RgbLight.settable_units = Light.settable_units + [Color]
RgbwLight.settable_units = Light.settable_units + [Color]
EmergencyUnit.settable_units = [EmergencyMode, FunctionTest, DurationTest]
