import logging

import spread_knx as knx
import spread_core.bam.const as BAM
from ..entities import Funit
from ..providers import Provider


class KnxProvider(Provider):
    def_state = None

    def __init__(self, id: int, address: knx.address.Address, **kwargs):
        super(KnxProvider, self).__init__(id, **kwargs)
        self._address = address
        self._status = self.SUSPICIOUS

    @property
    def binding(self) -> knx.address.Address:
        return self._address

    def parse_bindings(self, data):
        pass

    def on_survey(self):
        yield from super(KnxProvider, self).on_survey()
        # yield self.check_valid()

    def request(self, cmd, *args, sig=None, state_address=None, **kwargs):
        for response in self._manager.send_interface.send(cmd, state_address):
            self.on_request_answer(response, *args, sig=sig)

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None, **kwargs):
        logging.warning('Unknown response %s' % response)

    def check_valid(self):
        raise NotImplementedError()

    def get_info(self):
        self.check_valid()

    def prs2raw(self, p):
        if p == 0:
            return 0
        return min(max(1, round(255 * p / 100)), 255)

    def raw2prs(self, r):
        return None if r is None else min(100, round(100 * r / 255))

    def SetDiscovery(self, value, sig=None):
        pass

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        raise NotImplemented()

    def GetBrightnessLevel(self, sig=None):
        raise NotImplemented()


class KnxLights(KnxProvider):
    @staticmethod
    def parse_args(attributes, **kwargs) -> dict:
        return dict(
            **kwargs,
            address=knx.address.LightAddress.deserialize(attributes)
        )

    @property
    def binding(self) -> knx.address.LightAddress:
        return self._address

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == BrightnessLevel:
            return False

        return super(KnxLights, self).is_repaired_funit(funit_type)

    def check_valid(self):
        self.GetBrightnessLevel()

        level = self.get_value(BrightnessLevel)
        self.set_valid(self.SUSPICIOUS if level is None else self.VALID)


class KnxDimmer(KnxLights):
    def_state = knx.commands.Level

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None):
        if isinstance(response, knx.commands.Level):
            self.on_update(BrightnessLevel, self.raw2prs(response.value), sig=sig)

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        self.on_update(BrightnessLevel, value, preliminary=True, sig=sig)
        self.request(knx.commands.Level.set(self.binding.control, self.prs2raw(value)), sig=sig)

    def GetBrightnessLevel(self, sig=None):
        self.request(knx.commands.Level.get(self.binding.state), state_address=self.binding.state, sig=sig)


class KnxRelay(KnxProvider):
    def_state = knx.commands.Switching

    @staticmethod
    def parse_args(attributes, **kwargs) -> dict:
        return dict(
            **kwargs,
            address=knx.address.Address.deserialize(attributes)
        )

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == On:
            return False

        return super(KnxRelay, self).is_repaired_funit(funit_type)

    def check_valid(self):
        self.GetOn()

        on = self.get_value(On)
        self.set_valid(self.SUSPICIOUS if on is None else self.VALID)

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None):
        if isinstance(response, knx.commands.Switching):
            value = response.value
            self.on_update(On, value, sig=sig)

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        self.SetOn(None if value is None else value == 100, sig=sig, **kwargs)

    def SetOn(self, value, sig=None, **kwargs):
        self.on_update(On, value, preliminary=True, sig=sig)
        self.request(knx.commands.Switching.set(self.binding.control, value), sig=sig)

    def GetOn(self, sig=None):
        self.request(knx.commands.Switching.get(self.binding.state), state_address=self.binding.state, sig=sig)


class KnxMotor(KnxProvider):
    def_state = knx.commands.Level

    @staticmethod
    def parse_args(attributes, **kwargs) -> dict:
        return dict(
            **kwargs,
            address=knx.address.MotorAddress.deserialize(attributes)
        )

    @property
    def binding(self) -> knx.address.MotorAddress:
        return self._address

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == PositionLevel:
            return False

        return super(KnxMotor, self).is_repaired_funit(funit_type)

    def check_valid(self):
        self.GetLevel()

        level = self.get_value(PositionLevel)
        self.set_valid(self.SUSPICIOUS if level is None else self.VALID)

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None):
        if isinstance(response, knx.commands.Level):
            self.on_update(PositionLevel, PositionLevel.of_raw(response.value), sig=sig)

    def on_update(self, funit_type, value, sig=None, **kwargs):
        if funit_type == PositionLevel:
            self.on_update(Motion, Motion.of_position(self.get_value(PositionLevel), value), **kwargs)

        super(KnxMotor, self).on_update(funit_type, value, sig=sig, **kwargs)

    def SetLevel(self, raw_value, sig=None, **kwargs):
        if raw_value is None:
            self.request(knx.commands.Step.set(self.binding.stop, 1), sig=sig)
            # Значение возвращается само по завершении дижения
            # self.GetLevel()
        else:
            level = PositionLevel.of_raw(raw_value)
            has_level = self.get_value(PositionLevel)

            self.on_update(PositionLevel, level, preliminary=True, sig=sig)

            if level == has_level:
                self.GetLevel(sig=sig)
            else:
                self.request(knx.commands.Level.set(self.binding.control, raw_value), sig=sig)

    def GetLevel(self, sig=None):
        self.request(knx.commands.Level.get(self.binding.state), state_address=self.binding.state, sig=sig)


class KnxTemperatureSensor(KnxProvider):
    def_state = knx.commands.Float

    @staticmethod
    def parse_args(attributes, **kwargs) -> dict:
        return dict(
            **kwargs,
            address=knx.address.SensorAddress.deserialize(attributes)
        )

    @property
    def binding(self) -> knx.address.SensorAddress:
        return self._address

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == Temperature:
            return False

        return super(KnxTemperatureSensor, self).is_repaired_funit(funit_type)

    def check_valid(self):
        self.GetTemperature()

        temperature = self.get_value(Temperature)
        self.set_valid(self.SUSPICIOUS if temperature is None else self.VALID)

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None):
        if isinstance(response, knx.commands.Float):
            value = response.value
            if value is not None:
                value = 27315 + int(value * 100)
            self.on_update(Temperature, value, sig=sig)

    def GetTemperature(self, sig=None):
        self.request(knx.commands.Float.get(self.binding.state), state_address=self.binding.state, sig=sig)


class KNXFunit(Funit):
    entity_type = KnxProvider
    easy_get = False
    get_after_set = False
    set_cmd = None
    get_cmd = None
    set_func = None
    get_func = None

    @property
    def entity(self) -> KnxProvider:
        return super(KNXFunit, self).entity

    def set(self):
        if self.set_cmd:
            self.entity.request(self.set_cmd.set(self.entity.binding.control, self.set_value), sig=self.signature)
        elif self.set_func:
            getattr(self.entity, self.set_func.__name__)(self.set_value, **self.params, sig=self.signature)

        if self.get_cmd and self.get_after_set:
            self.get()

    def get(self):
        if self.easy_get:
            self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)
        elif self.get_cmd:
            self.entity.request(self.get_cmd.get(self.entity.binding.state), sig=self.signature)
        elif self.get_func:
            getattr(self.entity, self.get_func.__name__)(**self.params, sig=self.signature)
        else:
            raise self.NotImplemented()


KnxProvider.funit_class = KNXFunit


class KnxLightFunit(KNXFunit):
    pass


class BrightnessLevel(KnxLightFunit):
    set_func = KnxProvider.SetBrightnessLevel
    get_func = KnxProvider.GetBrightnessLevel
    value_range = range(0, 101)


class On(KNXFunit):
    set_func = KnxRelay.SetOn
    get_func = KnxRelay.GetOn
    entity_type = KnxRelay
    value_type = bool


class GroupLevel(BrightnessLevel):
    pass


class KnxMotorFunit(KNXFunit):
    pass


class PositionLevel(KnxMotorFunit):
    set_func = KnxMotor.SetLevel
    get_func = KnxMotor.GetLevel
    value_range = range(0, 10_001)

    @classmethod
    def of_raw(cls, raw_level):
        if raw_level is None:
            return None

        try:
            return int(round(raw_level / 254 * 10000))
        except:
            return None

    @staticmethod
    def to_raw(level):
        if level in (0, None):
            return level

        return int(round(level * 254 / 10000))

    @property
    def set_value(self):
        return self.to_raw(super(PositionLevel, self).set_value)


class Motion(KnxMotorFunit):
    variants = (BAM.MOTOR_OPEN, BAM.MOTOR_CLOSE, BAM.MOTOR_STOP)
    state_variants = (BAM.MOTOR_OPENING, BAM.MOTOR_CLOSING, BAM.MOTOR_STOPPED)

    def set(self):
        position = self.to_position(self.set_value)

        if position is None:
            assert self.entity.get_value(self.__class__) != BAM.MOTOR_STOPPED, "Already stopped"

        PositionLevel(self.entity, position).set()

    @classmethod
    def to_position(cls, motion: str) -> [None, int]:
        if motion == BAM.MOTOR_CLOSE:
            return PositionLevel.value_range.stop - 1
        elif motion == BAM.MOTOR_OPEN:
            return PositionLevel.value_range.start
        elif motion == BAM.MOTOR_STOP:
            return None

        raise ValueError(f'"direction" expected as one of [{BAM.MOTOR_OPEN}, {BAM.MOTOR_CLOSE}, {BAM.MOTOR_STOP}] but {motion} is given')

    @staticmethod
    def of_position(has_position, position):
        """
        Определяет направление движения по предыдущей и текущей позициям
        Разница в 50 единиц обусловлена необходимостью иметь погрешность

        :param has_position: последнее известное значение позиции
        :param position: полученное значение позиции
        :return: "Opening", "Closing" или "Stopped"

        """
        if position is None:
            return None
        elif has_position is None:
            return BAM.MOTOR_STOPPED
        else:
            if position > has_position:
                return BAM.MOTOR_CLOSING
            elif position < has_position:
                return BAM.MOTOR_OPENING
            else:
                return BAM.MOTOR_STOPPED


class KnxSensorFunit(KNXFunit):
    def set(self):
        raise self.NotImplemented()


class Temperature(KnxSensorFunit):
    get_func = KnxTemperatureSensor.GetTemperature
