import json
import logging
import os

import spread_lom as lom
from spread_core.mqtt.project import Preset
from spread_core.errors.project_errors import ProjectError, InitError
from spread_core.tools.settings import config, PROFILES, TEMP_DIMMING_PERIOD
from spread_core.tools.timer import Timer
from ..entities import Funit
from ..managers import Manager

RETAINED = [lom.const.PROFILES]
SET_PROFILES_TIMEOUT = 5


class Lom(Manager):
    def_period = None
    LOCAL_PATH = '/var/project'

    def __init__(self, p_id, m_id):
        super().__init__(p_id, m_id)
        self.profiles = {}
        self.coordinates = (0, 0, 0)
        self.mc = None
        self.profiles_to_update = None
        self.profiles_timer = None

    def parse(self, data):
        try:
            data['attributes']['bus'] = dict(
                    host=data['attributes']['host'],
                    port=data['attributes']['port'],
                    login=data['attributes']['login'],
                    pwd=data['attributes']['password'],
                    send_topic=data['attributes']['downLink'],
                    dump_topic=data['attributes']['upLink']
                )

            super(Lom, self).parse(data)
            self.coordinates = (
                data['attributes']['latitude'],
                data['attributes']['longitude'],
                data['attributes']['altitude'],
            )

            self.mc = data['attributes']['broadcast']
        except KeyError as ex:
            # ToDo надо райсить InitError
            raise ProjectError(f'Некорректная конфигурация для менеджера {self} ({",".join(ex.args)})')

        # self.add_provider({
        #     'id': -1,
        #     'type': 'LomMulticast',
        #     'managerID': self.id,
        #     'name': lom.const.MULTICAST,
        #     'attributes': {
        #         'devEui': data['attributes']['broadcast']
        #     }
        # }, self)

        self.read_profiles()

    def set_interface(self):
        self.send_interface = lom.interface.LomInterface()

    def update_config(self):
        super(Lom, self).update_config()
        self.def_period = config[TEMP_DIMMING_PERIOD]

    def _publish_profiles(self, sig=None):
        arr = {}
        for profile in self.profiles.values():
            arr[profile.id] = profile.serialize()
        self.on_update(Profiles, arr, sig=sig)

    def read_profiles(self):
        self.profiles = {}

        dest = Preset.destination(self.LOCAL_PATH, self.project_id, entity_type=lom.profiles.DimmingProfile.__name__)
        for dirpath, dirnames, filenames in os.walk(dest):
            for file in filenames:
                try:
                    preset_id = int(os.path.splitext(file)[0])
                    topic = Preset(self.project_id, lom.profiles.DimmingProfile.__name__, preset_id)
                    with open(os.path.join(dirpath, file), 'r') as f:
                        data = json.load(f)

                    profile = lom.profiles.GlobalDimmingProfile.deserialize({'id': topic.entity_id, **data})
                    self.profiles[profile.id] = profile
                except Exception as ex:
                    logging.warning("Profile reading error: %s" % ex)
                    continue

        self.subscribe(Preset.broadcast(self.project_id, lom.profiles.DimmingProfile.__name__))

    def add_profile(self, profile: lom.profiles.GlobalDimmingProfile):
        need_update = False
        if profile.id not in self.profiles:
            self.profiles[profile.id] = profile
            need_update = True
        else:
            items = self.profiles[profile.id].items
            for item in profile.items:
                if item not in items:
                    self.profiles[profile.id].add_item(item)
                    need_update = True
        if need_update:
            self._publish_profiles()

    def profile_changed(self, topic: Preset, data: str):
        assert topic.entity_type == lom.profiles.DimmingProfile.__name__

        if self.profiles_to_update is None:
            self.profiles_to_update = {}

        self.profiles_to_update[topic.entity_id] = json.loads(data) if data else None
        self.start_profiles_timer()

    def stop_profiles_timer(self):
        if self.profiles_timer:
            try:
                self.profiles_timer.close()
            except ValueError:
                pass
            finally:
                self.profiles_timer = None

    def start_profiles_timer(self):
        self.stop_profiles_timer()
        self.profiles_timer = Timer.set(SET_PROFILES_TIMEOUT, self.SetProfiles, self.profiles_to_update)

    def on_external_command(self, command, *args, **kwargs):
        try:
            for provider in self.providers_by_address(command.dev_eui):
                provider.on_command_sended(command)
        except lom.errors.UnknownProfile as ex:
            logging.error(ex)
        except BaseException as ex:
            logging.exception(ex)

    def on_event(self, event):
        try:
            for provider in self.providers_by_address(event.dev_eui):
                if isinstance(event, lom.commands.LEvent):
                    for funit_type, data in event.value.items():
                        provider.on_update(funit_type, data)
                else:
                    provider.on_request_answer(None, event)

        except lom.errors.UnknownProfile as ex:
            logging.error(ex)
        except BaseException as ex:
            logging.exception(ex)

    def SetProfiles(self, profiles: dict, sig=None):
        need_update = False
        for _id, _data in profiles.items():
            if isinstance(_id, str) and _id.isdigit():
                _id = int(_id)
            if _data is None:
                if _id in self.profiles:
                    self.profiles.pop(_id)
                    need_update = True
                    for mc in self.providers_by_address(self.mc):
                        mc.request(lom.commands.ClearProfile(self.mc, [_id]))
            else:
                try:
                    profile = lom.profiles.GlobalDimmingProfile.deserialize({'id': _id, **_data})
                    if profile.id not in self.profiles or profile != self.profiles[profile.id]:
                        self.profiles[_id] = profile
                        for mc in self.providers_by_address(self.mc):
                            mc.request(lom.commands.SetProfile.of(self.mc, profile, mc=True))
                        need_update = True
                except BaseException as ex:
                    raise lom.errors.ParameterError(f'<<{_data}>>: {ex}')

        if need_update:
            self._publish_profiles(sig=sig)


class LMFunit(Funit):
    @property
    def entity(self) -> Lom:
        return super(LMFunit, self).entity

    def get(self):
        self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)


Lom.funit_class = LMFunit


class Profiles(LMFunit):
    def set(self):
        self.entity.SetProfiles(self.set_value)
