import logging

import spread_core.mqtt as mqtt
from spread_core.tools import settings
from .entities import Entity


class Provider(Entity):
    VALID = 'valid'
    SUSPICIOUS = 'suspicious'
    INVALID = 'invalid'

    def __init__(self, id, **kwargs):
        super(Provider, self).__init__(id, **kwargs)
        self._manager = None
        self._status = self.VALID

    @property
    def binding(self):
        raise NotImplementedError()

    @property
    def manager(self): return self._manager

    @property
    def project_id(self):
        return self.manager.project_id

    def publish(self, topic, data, retain=False, **kwargs):
        self._manager.publish(topic, data, retain, **kwargs)

    def get_address(self, funit_type):
        return mqtt.spread.address.ProviderAddress(
            self.manager.project_id,
            self.manager.__class__.__name__,
            self.manager.id,
            self.__class__.__name__,
            self.id,
            funit_type
        )

    @classmethod
    def get_address_broadcast(cls, p_id, funit_type='+'):
        pass

    @property
    def surveyable(self):
        return True

    @property
    def infoble(self):
        return True

    @property
    def is_device(self):
        return True

    @property
    def is_group(self):
        return False

    @property
    def is_broadcast(self):
        return False

    @property
    def is_valid(self):
        return self._status != self.INVALID

    @property
    def is_valid_strong(self):
        return self._status == self.VALID

    def set_manager(self, manager):
        self._manager = manager

    def on_command_sended(self, cmd):
        pass

    def on_queue(self):
        pass

    def request(self, cmd, *args, **kwargs):
        raise NotImplementedError()

    def on_request_answer(self, funit_type, response, *args, sig=None):
        self.on_update(funit_type, None, sig=sig)

    def set_valid(self, validity):
        if validity == self._status == self.SUSPICIOUS:
            validity = self.INVALID

        has_validity = self.is_valid

        if self._status != validity:
            logging.info(f'{self} is {validity}')
            self._status = validity

        if has_validity != self.is_valid:
            if self.is_valid:
                self.restore_states()
                self.manager.put_to_stack(
                    self.get_info
                )
            else:
                self.remove_states()

    def on_status_survey(self):
        has_valid = self.is_valid
        self.check_valid()
        yield self.is_valid != has_valid

    def on_survey(self):
        yield from ()

    def get_info(self):
        raise NotImplemented()

    def restore_states(self):
        for funit_type, value in self._invalid_values.copy().items():
            self.make_state(funit_type, mqtt.spread.variable.Variable(value))

    def remove_states(self):
        for funit_type, value in self.values.copy().items():
            self.make_state(funit_type, mqtt.spread.variable.Variable(value, invalid=True))

    def on_state(self, funit_type, variable: mqtt.spread.variable.Variable):
        if variable.value is not None:
            if not self.is_device or not variable.preliminary:
                if self.is_repaired_funit(funit_type):
                    self._invalid_values[funit_type] = variable.value

    def check_valid(self):
        """ Проверка валидности провайдера """
        pass

    def on_exit(self):
        self.remove_states()

    def override_info(self):
        for funit_type, value in settings.dump.get_dump_entity(self).items():
            self.on_update(funit_type, value)
