import logging
import socket
import sys

logging.basicConfig(format=u'%(levelname)-8s [%(asctime)s]  %(message)s', level=logging.DEBUG, stream=sys.stdout)

HOST = '10.10.1.254'
PORT = 4001
TIMEOUT = 1


CMD_REQ     = 0x0C
CMD_RESP    = 0x0D
CMD_WRT     = 0x16
CMD_SUC     = 0x17
CMD_ERR     = 0x06

# Настройки:
PAR_TIME        = 60000
PAR_ID          = 60004
PAR_POWER       = 60006
PAR_MODE        = 60007
PAR_PERIOD      = 60008
PAR_PCL_MODE    = 60009
PAR_BAUD_RATE   = 60010
PAR_INFO        = 60012
PAR_PIN         = 60076
PAR_PIN_USED    = 60080
PAR_RESET       = 65500


def crc16(data):
    crc = 0xFFFF
    i = 0
    while i < len(data):
        j = 0
        crc = crc ^ data[i]
        while j < 8:
            if crc & 0x1:
                mask = 0xA001
            else:
                mask = 0x00
            crc = ((crc >> 1) & 0x7FFF) ^ mask
            j += 1
        i += 1
    if crc < 0:
        crc -= 256

    return crc.to_bytes(2, 'little')


def input_reader():
    addr = 1
    while True:
        data = input()
        while '  ' in data:
            data = data.replace('  ', ' ')

        try:
            data = bytes.fromhex(data)
            yield data + crc16(data)
        except ValueError:
            data = data.split()

            if len(data) == 0:
                continue

            elif data[0].lower() in ('stop', 'exit', 'break'):
                break

            elif data[0].lower() == 'address':
                if len(data) > 1:
                    addr = int(data[1])
                logging.debug('Address %s in use', addr)

            elif data[0].lower() == 'reset':
                yield pack(addr, CMD_WRT, PAR_RESET, 0x2a)

            elif data[0].lower() == 'level':
                if len(data) > 1:
                    yield pack(addr, CMD_WRT, PAR_POWER, int(data[1]))
                else:
                    yield pack(addr, CMD_REQ, PAR_POWER)

            else:
                try:
                    yield pack(*data)
                except ValueError:
                    logging.warning('Bad data!!')


def pack(addr, cmd, par, *args):
    data = b(addr, 2) + b(cmd, 1) + b(par, 2)

    if len(args) > 0:
        pl = b(args[0])
    else:
        pl = b''

    data += b(len(pl), 2) + pl

    return data + crc16(data)


def g():
    try:
        yield from input_reader()
    except StopIteration:
        pass
    except BaseException as ex:
        pass


def run():
    _socket = None
    i = 0
    while not _socket:
        try:
            sys.stdout.write('\rConnecting to %s:%s %s' % (HOST, PORT, '.'*i))
            _socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            _socket.settimeout(TIMEOUT)
            _socket.connect((HOST, PORT))
        except socket.timeout:
            _socket = None
            i = (i+1) % 5
    else:
        print()
        logging.info('Connected')

    for forward_frame in g():
        if forward_frame:
            _socket.send(forward_frame)
            logging.debug('-> %s' % bytes2str(forward_frame))

        r_size = 1024
        backward_frame = b''
        try:
            while len(backward_frame) < r_size:
                backward_frame += _socket.recv(r_size - len(backward_frame))
        except socket.timeout:
            pass

        if backward_frame:
            if backward_frame[-2:] != crc16(backward_frame[:-2]):
                logging.info('Неверная контрольная сумма ответа!')

            logging.debug('  <- %s' % bytes2str(backward_frame))

    logging.debug('Kill socket')
    _socket.close()


def listen():
    _socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    _socket.settimeout(TIMEOUT)
    _socket.connect((HOST, PORT))

    data = b''

    try:
        while True:
            try:
                data += _socket.recv(1)
                continue
            except socket.timeout:
                data = b''
                backward_frame = data

            if backward_frame:
                logging.debug('  <- %s' % bytes2str(backward_frame))
    except KeyboardInterrupt:
        logging.debug('Kill socket')
        _socket.close()


def bytes2str(data: bytes) -> str:
    return data.hex().upper()


def b(value, length=None):
    if isinstance(value, str):
        value = int(value, 10 if value.isdigit() else 16)

    if length is None:
        length = max(1, (value.bit_length() + 7) // 8)

    return value.to_bytes(length, 'little')


if __name__ == '__main__':
    if 'listen' in sys.argv:
        listen()
    else:
        run()
