from typing import Union

from spread_core.errors.project_errors import ClassifierError
from spread_core.mqtt.classifier import classifier
from ..base import TopicError


class EntityAddress:
    S_TYPE = 'AppServer'
    family = 'Hardware'

    def __init__(self, e_type: str, e_id: Union[int, str], funit_type: Union[int, str]):
        self._type = e_type
        self._id = e_id
        self._funit_type = funit_type

    type = property(fget=lambda self: self._type, doc='type')
    entity_id = property(fget=lambda self: to_int(self._id), doc='entity_id')
    funit_type = property(fget=lambda self: self._funit_type, doc='funit_type')
    funit_id = property(fget=lambda self: self._funit_type, doc='funit_id')

    @property
    def funit(self):
        if self.type not in classifier:
            raise ClassifierError.of_funit(self.type)
        elif self.funit_type not in classifier[self.type]:
            raise ClassifierError.of_funit(self.funit_type, self.type)
        else:
            return classifier[self.type][self.funit_type]

    def __str__(self):
        return f'{self.family}/{self.type}/{self.entity_id}/{self.funit_type}'

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return str(self) == str(other)

    @classmethod
    def of(cls, entity, funit):
        return cls(entity.__class__.__name__, entity.id, funit)

    @staticmethod
    def deserialize(family, *args):

        try:
            if family == ManagerAddress.family:
                if len(args) == 5:
                    return ManagerAddress(*args)

                elif len(args) == 7:
                    return ProviderAddress(*args)

            elif family == EngineryAddress.family:
                return EngineryAddress(*args)

        except TypeError:
            raise TopicError('Invalid address: "%s"' % '/'.join([family, '/'.join(args)]))

        raise TopicError('Unknown family: "%s"' % family)


class ManagerAddress(EntityAddress):
    def __init__(self, server_type='+', server_id='+', manager_type='+', manager_id='+', funit_type='+'):
        super().__init__(manager_type, manager_id, funit_type)
        self._server_type = server_type
        self._server_id = server_id

    server_type = property(fget=lambda self: self._server_type, doc='server_type')
    server_id = property(fget=lambda self: to_int(self._server_id), doc='server_id')

    def __str__(self):
        return f'{self.family}/{self.server_type}/{self.server_id}/{self.type}/{self.entity_id}/{self.funit_type}'

    @classmethod
    def of(cls, manager, funit_type='+'):
        return cls(EngineryAddress.S_TYPE, manager.server_id, manager.__class__.__name__, manager.id, funit_type)


class ProviderAddress(EntityAddress):
    def __init__(self, server_type='+', server_id='+', manager_type='+', manager_id='+', provider_type='+', provider_id='+', funit_type='+'):
        super().__init__(provider_type, provider_id, funit_type)
        self._server_type = server_type
        self._server_id = server_id
        self._manager_type = manager_type
        self._manager_id = manager_id

    server_type = property(fget=lambda self: self._server_type, doc='server_type')
    server_id = property(fget=lambda self: to_int(self._server_id), doc='server_id')
    manager_type = property(fget=lambda self: self._manager_type, doc='manager_type')
    manager_id = property(fget=lambda self: to_int(self._manager_id), doc='manager_id')

    def __str__(self):
        return f'{self.family}/{self.server_type}/{self.server_id}/{self.manager_type}/{self.manager_id}/{self.type}/{self.entity_id}/{self.funit_type}'

    @classmethod
    def of(cls, entity, funit_type='#'):
        return cls(
            cls.S_TYPE,
            entity.manager.server_id, entity.manager.__class__.__name__, entity.manager.id,
            entity.__class__.__name__, entity.id, funit_type
        )

    @property
    def funit(self):
        if self.manager_type not in classifier:
            raise ClassifierError.of_funit(self.manager_type)
        elif self.type not in classifier[self.manager_type]:
            raise ClassifierError.of_funit(self.type, self.manager_type)
        elif self.funit_type not in classifier[self.manager_type][self.type]:
            raise ClassifierError.of_funit(self.funit_type, self.manager_type, self.type)
        else:
            return classifier[self.manager_type][self.type][self.funit_type]


class SubgineryAddress(EntityAddress):
    family = 'Equipment'

    @property
    def funit(self):
        if self.type not in classifier['Lighting']:
            raise ClassifierError.of_funit(self.type, 'Lighting')
        elif self.funit_type not in classifier['Lighting'][self.type]:
            raise ClassifierError.of_funit(self.funit_type, 'Lighting', self.type)
        else:
            return classifier['Lighting'][self.type][self.funit_type]


class EngineryAddress(EntityAddress):
    family = 'Equipment'

    @property
    def funit(self):
        if self.type not in classifier['Lighting']:
            raise ClassifierError.of_funit(self.type, 'Lighting')
        elif self.funit_type not in classifier['Lighting'][self.type]:
            raise ClassifierError.of_funit(self.funit_type, 'Lighting', self.type)
        else:
            return classifier['Lighting'][self.type][self.funit_type]


def to_int(value):
    if isinstance(value, str) and value.isdigit():
        return int(value)

    return value
