from ..base import TopicError


HARDWARE = "Hardware"
EQUIPMENT = "Equipment"
LOCATION = "Location"


class EntityAddress:
    family = None

    def __init__(self, project_id: int, e_type: str, e_id: int, funit_type):
        self._project_id = int(project_id)
        self._type = e_type
        self._id = int(e_id)
        self._funit_type = funit_type
    
    @property
    def project_id(self):
        return self._project_id
    
    @property
    def type(self):
        return self._type
        
    @property
    def id(self):
        return self._id
    
    @property
    def funit_type(self):
        return self._funit_type

    def __str__(self):
        return f'{self.project_id}/{self.family}/{self.type}/{self.id}/{self.funit_type}'

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return str(self) == str(other)

    @classmethod
    def broadcast(cls, project_id, e_type='+', e_id='+', funit_type='+'):
        return f'{project_id}/{cls.family}/{e_type}/{e_id}/{funit_type}'


class ManagerAddress(EntityAddress):
    family = HARDWARE


class ProviderAddress(EntityAddress):
    family = HARDWARE

    def __init__(self, project_id, manager_type, manager_id, e_type, e_id, funit_type):
        super(ProviderAddress, self).__init__(project_id, e_type, e_id, funit_type)
        self._manager_type = manager_type
        self._manager_id = int(manager_id)
    
    @property
    def manager_type(self):
        return self._manager_type
    
    @property
    def manager_id(self):
        return self._manager_id

    def __str__(self):
        return f'' \
            f'{self.project_id}/{self.family}/' \
            f'{self.manager_type}/{self.manager_id}/' \
            f'{self.type}/{self.id}/{self.funit_type}'

    @classmethod
    def broadcast(cls, project_id, manager_type='+', manager_id='+', e_type='+', e_id='+', funit_type='+'):
        return f'{project_id}/{cls.family}/{manager_type}/{manager_id}/{e_type}/{e_id}/{funit_type}'


class EngineryAddress(EntityAddress):
    family = EQUIPMENT


class SubgineryAddress(EntityAddress):
    family = LOCATION

    def __init__(self, project_id, location_id: int, subginery_type: str, funit_type):
        super(SubgineryAddress, self).__init__(project_id, subginery_type, location_id, funit_type)

    def __str__(self):
        return f'{self.project_id}/{self.family}/{self.id}/{self.type}/{self.funit_type}'

    @classmethod
    def broadcast(cls, project_id, location_id='+', subginery_type='+', funit_type='+'):
        return f'{project_id}/{cls.family}/{location_id}/{subginery_type}/{funit_type}'


def of(project_id, family, *parts):

    try:
        if family == HARDWARE:
            if len(parts) == 5:
                return ProviderAddress(project_id, *parts)

            return ManagerAddress(project_id, *parts)

        elif family == EQUIPMENT:
            return EngineryAddress(project_id, *parts)

        elif family == LOCATION:
            return SubgineryAddress(project_id, *parts)

    except TypeError:
        raise TopicError('Invalid address: "%s"' % '/'.join([project_id, family, *parts]))

    raise TopicError('Unknown family: "%s"' % family)
