from unittest import TestCase

from spread_core.mqtt.spread.address import ProviderAddress
from spread_core.errors.project_errors import CommandError

from spread_core.bam.engineries import alarm
from spread_core.mqtt.spread.topic import State
from spread_core.mqtt.spread.variable import Variable
from spread_core.tests.equipment.utils import TestEquipmentAdapter, TestEngineryRecipe


class _TestAlarmSensors(TestCase):
    enginery_type = alarm._AlarmSensor

    def setUp(self) -> None:
        self.adapter = TestEquipmentAdapter()
        self.recipe = TestEngineryRecipe(3)

        enginery = self.enginery_type(999)
        enginery.set_adapter(self.adapter)

        self.enginery = enginery

    def _emulate_state(self, funit, value, provider_id):
        self.enginery.on_state(State(ProviderAddress(1, 'Manager', 2, 'Provider', provider_id, funit)), Variable(value))

    def _check_state(self, funit, value, provider_id=3, expect_state=True):
        self._emulate_state(funit, value, provider_id)
        self._check_value(funit, value)

        if expect_state:
            self._check_publish(f'Spread/State/1/Equipment/{self.enginery_type.__name__}/999/{funit}', value)

        self._check_end_of_publishes()

    def _check_value(self, funit, expected_value):
        self.assertEqual(expected_value, self.enginery.values.get(funit))

    def _check_end_of_publishes(self):
        if self.adapter.queue:
            unexpected_publishes = f'Unexpected publishes ({len(self.adapter.queue)}):\n'
            unexpected_publishes += "\n".join(str(i) for i in self.adapter.queue)
            self.fail(unexpected_publishes)

    def _check_publish(self, _topic, _value, _retain=True, _delay=0):
        topic, variable, retain, delay = self.adapter.queue.popleft()
        self.assertEqual(topic, _topic)
        self.assertEqual(variable.value, _value)
        self.assertEqual(retain, _retain)
        self.assertEqual(delay, _delay)

    def _set_solid_recipe(self):
        self.recipe = TestEngineryRecipe(3)
        self.enginery.set_recipe('Solid', self.recipe)

    def _set_isGuarded_recipe(self):
        self.recipe1 = TestEngineryRecipe(4)
        self.recipe2 = TestEngineryRecipe(5)
        self.enginery.set_recipe('isGuarded', self.recipe1)
        self.enginery.set_recipe('isTriggered', self.recipe2)

    def _set_isTriggered_recipe(self):
        self.recipe2 = TestEngineryRecipe(5)
        self.enginery.set_recipe('isTriggered', self.recipe2)

    def test_solid_recipe_appointment(self):
        self._set_solid_recipe()
        self.assertRaises(ValueError, self.enginery.set_recipe, 'SomeString', self.recipe)
        self.assertIs(self.enginery.recipe, self.recipe)
        self.assertEqual(('Guard', self.recipe), self.enginery.recipe_for_funit('Guard'))

    def test_isGuarded_recipe_appointment(self):
        self._set_isGuarded_recipe()
        self.assertIs(self.enginery.recipe, self.recipe1)
        self.assertIs(self.enginery._trigger, self.recipe2)
        self.assertEqual(('Guard', self.recipe1), self.enginery.recipe_for_funit('Guard'))

    def test_isTriggered_recipe_appointment(self):
        self._set_isTriggered_recipe()
        self.assertIs(self.enginery.recipe, None)
        self.assertIs(self.enginery._trigger, self.recipe2)
        self.assertEqual(('Guard', None), self.enginery.recipe_for_funit('Guard'))

    def test_address(self):
        self.assertEqual(self.enginery.get_address('SomeFunit'), f'1/Equipment/{self.enginery_type.__name__}/999/SomeFunit')

    def test_address_broadcast(self):
        self.assertEqual(self.enginery.get_address_broadcast(1, 'SomeFunit'), f'1/Equipment/{self.enginery_type.__name__}/+/SomeFunit')

    def test_solid_state_handler(self):
        self._set_solid_recipe()
        self._check_state('Guard', True)
        self._check_state('Guard', True, expect_state=False)
        self._check_state('Guard', False)
        self._check_state('Guard', None)
        self._check_state('Triggered', True)
        self._check_state('Triggered', False)
        self._check_state('Triggered', None)

    def test_isGuarded_state_handler(self):
        self._set_solid_recipe()
        self._check_state('Guard', True, 4)
        self._check_state('Guard', True, 4, expect_state=False)
        self._check_state('Guard', False, 4)
        self._check_state('Guard', None, 4)
        self._check_state('Triggered', True, 5)
        self._check_state('Triggered', False, 5)
        self._check_state('Triggered', None, 5)

    def test_isTriggered_state_handler(self):
        self._set_solid_recipe()
        self._check_state('Triggered', True, 5)
        self._check_state('Triggered', True, 5, expect_state=False)
        self._check_state('Triggered', False, 5)
        self._check_state('Triggered', None, 5)

    def test_invalidation(self):
        self.enginery.values['Funit1'] = 1
        self.enginery.values['Funit2'] = 'Funit2'
        self.enginery.values['Funit3'] = [1, 2, 3]

        self.enginery.set_invalid()

        self._check_publish(f'Spread/State/1/Equipment/{self.enginery_type.__name__}/999/Funit1', None)
        self._check_publish(f'Spread/State/1/Equipment/{self.enginery_type.__name__}/999/Funit2', None)
        self._check_publish(f'Spread/State/1/Equipment/{self.enginery_type.__name__}/999/Funit3', None)

        self._check_end_of_publishes()

    def test_solid_funit(self):
        self._set_solid_recipe()
        alarm.Guard(self.enginery, Variable(True)).set()
        self._check_publish('Spread/Set/Provider/3/Guard', True, False, 0)
        self._check_end_of_publishes()

        alarm.Guard(self.enginery, Variable(False)).set()
        self._check_publish('Spread/Set/Provider/3/Guard', False, False, 0)
        self._check_end_of_publishes()

    def test_isGuarded_funit(self):
        self._set_isGuarded_recipe()
        alarm.Guard(self.enginery, Variable(True)).set()
        self._check_publish('Spread/Set/Provider/4/Guard', True, False, 0)
        self._check_end_of_publishes()

        alarm.Guard(self.enginery, Variable(False)).set()
        self._check_publish('Spread/Set/Provider/4/Guard', False, False, 0)
        self._check_end_of_publishes()

    def test_isTriggered_funit(self):
        self._set_isTriggered_recipe()
        self.assertRaises(CommandError, alarm.Guard(self.enginery, Variable(True)).set)
        self.assertRaises(CommandError, alarm.Guard(self.enginery, Variable(False)).set)

    def test_unsettable_funit(self):
        self.assertRaises(alarm.Triggered.NotImplemented, alarm.Triggered(self.enginery, Variable(True)).set)


class TestFireSensor(_TestAlarmSensors):
    enginery_type = alarm.FireSensor


class TestIntruderSensor(_TestAlarmSensors):
    enginery_type = alarm.IntruderSensor


class TestLeakageSensor(_TestAlarmSensors):
    enginery_type = alarm.LeakageSensor
