from unittest import TestCase

from spread_core.bam.engineries import Enginery

from spread_core.bam.engineries.alarm import FireSensor, LeakageSensor, IntruderSensor

from spread_core.bam.subgineries import alarming as alarming
from spread_core.mqtt.spread.variable import Variable
from spread_core.tests.equipment.utils import TestEquipmentAdapter, TestEngineryRecipe


class TestAlarm(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.adapter = TestEquipmentAdapter()

    def setUp(self) -> None:
        ENGS = [FireSensor, IntruderSensor, LeakageSensor, Enginery]

        self.adapter.add_manager(2, "Manager")

        for i in range(4):
            self.adapter.add_provider(TestEngineryRecipe(1000+i))
            self.adapter.add_enginery(ENGS[i](i*4), 1000+i, "Solid")

            self.adapter.add_provider(TestEngineryRecipe(1100+i))
            self.adapter.add_enginery(ENGS[i](1+i*4), 1100+i, "Solid")

            e = ENGS[i](2+i*4)
            self.adapter.add_provider(TestEngineryRecipe(1200+i, 'DaliRelay'))
            self.adapter.add_enginery(e, 1200+i, "isGuarded")
            self.adapter.add_provider(TestEngineryRecipe(1300+i, 'DaliRelay'))
            self.adapter.add_enginery(e, 1300+i, "isTriggered")

            self.adapter.add_provider(TestEngineryRecipe(1400+i, 'DaliRelay'))
            self.adapter.add_enginery(ENGS[i](3+i*4), 1400+i, "isTriggered")

        self._sb = alarming.Alarm(99, 88, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15])
        self._sb.set_adapter(self.adapter)

    def _check_publish_state(self, funit, value, delay=0, params=None, attrs=None):
        self.__check_publish('Spread/State/1/Location/88/Alarm/%s' % funit, value, _retain=True, _delay=delay, params=params, attrs=attrs)

    def _check_publish_command(self, p_id, funit, value, params=None):
        self.__check_publish('Spread/Set/Provider/%s/%s' % (p_id, funit), value, _retain=False, _delay=0, params=params)

    def _check_end_of_publishes(self):
        if self.adapter.queue:
            unexpected_publishes = f'Unexpected publishes({len(self.adapter.queue)} pcs):\n'
            unexpected_publishes += "\n".join(str(i) for i in self.adapter.queue)
            self.fail(unexpected_publishes)

    def __check_publish(self, _topic, _value, _retain=True, _delay=0, params=None, attrs=None):
        if len(self.adapter.queue) == 0:
            self.fail('Expected publish is absent: %s' % _topic)
        else:
            topic, variable, retain, delay = self.adapter.queue.popleft()
            self.assertEqual(_topic, topic)
            self.assertEqual(_value, variable.value)
            self.assertEqual(_retain, retain)
            self.assertEqual(_delay, delay)

            if params:
                for param_name in params:
                    try:
                        self.assertEqual(params[param_name], variable.params[param_name])
                    except KeyError:
                        self.fail('Expected variable parameter "%s" is absent' % param_name)

            if attrs:
                for attrs_name in attrs:
                    try:
                        self.assertEqual(attrs[attrs_name], getattr(variable, attrs_name))
                    except KeyError:
                        self.fail('Expected variable attribute "%s" is absent' % attrs_name)

    @property
    def sb(self) -> alarming.Alarm:
        return self._sb

    def test_init(self):
        self.assertEqual(1, self.sb.project_id)
        self.assertEqual(88, self.sb.location_id)
        self.assertEqual(99, self.sb.id)
        self.assertEqual(self.sb._enginery_ids, [e.id for e in self.sb.engineries])

        for e in self.sb.engineries:
            assert isinstance(e, Enginery)
            self.assertIn(self.sb, e._subgineries)

    def test_address(self):
        self.assertEqual('1/Location/88/Alarm/SomeFunit', self.sb.get_address("SomeFunit"))

    def test_state_publishing(self):
        self.sb.publish_state('SomeFunit', 'SomeValue')
        self._check_publish_state('SomeFunit', 'SomeValue')
        self._check_end_of_publishes()

    def test_invalidation(self):
        self.sb.values['SomeFunit'] = 2
        self.sb.set_invalid()

        self._check_publish_state('SomeFunit', None)
        self._check_end_of_publishes()

    def _set_enginery_guard(self, enginery_id, value, expected_unit=None, expected_value=None, **attrs):
        e = self.adapter._entities[enginery_id]
        assert isinstance(e, Enginery)
        e.values['Guard'] = value
        self.sb.on_level_changed(e, 'Guard')
        if expected_unit:
            self._check_publish_state(expected_unit, expected_value, attrs=attrs)

        self._check_end_of_publishes()

    def test_fire_guard_changing(self):
        self._set_enginery_guard(0, True, 'FireSensorsGuard', True, invalid=True, inconsistent=False)
        self._set_enginery_guard(1, True)
        self._set_enginery_guard(2, True, 'FireSensorsGuard', True, invalid=False, inconsistent=False)

        self._set_enginery_guard(4, False, 'IntruderSensorsGuard', None, invalid=True, inconsistent=True)
        self._set_enginery_guard(10, False, 'LeakageSensorsGuard', None, invalid=True, inconsistent=True)

        self._set_enginery_guard(2, False, 'FireSensorsGuard', None, invalid=False, inconsistent=True)
        self._set_enginery_guard(0, None, 'FireSensorsGuard', None, invalid=True, inconsistent=True)
        self._set_enginery_guard(1, None)
        self._set_enginery_guard(2, None, 'FireSensorsGuard', True, invalid=True, inconsistent=False)

    def test_intruder_guard_changing(self):
        self._set_enginery_guard(4, True, 'IntruderSensorsGuard', True, invalid=True, inconsistent=False)
        self._set_enginery_guard(5, True)
        self._set_enginery_guard(6, True, 'IntruderSensorsGuard', True, invalid=False, inconsistent=False)

        self._set_enginery_guard(0, False, 'FireSensorsGuard', None, invalid=True, inconsistent=True)
        self._set_enginery_guard(10, False, 'LeakageSensorsGuard', None, invalid=True, inconsistent=True)

        self._set_enginery_guard(6, False, 'IntruderSensorsGuard', None, invalid=False, inconsistent=True)
        self._set_enginery_guard(4, None, 'IntruderSensorsGuard', None, invalid=True, inconsistent=True)
        self._set_enginery_guard(5, None)
        self._set_enginery_guard(6, None, 'IntruderSensorsGuard', True, invalid=True, inconsistent=False)

    def test_leakage_guard_changing(self):
        self._set_enginery_guard(8, True, 'LeakageSensorsGuard', True, invalid=True, inconsistent=False)
        self._set_enginery_guard(9, True)
        self._set_enginery_guard(10, True, 'LeakageSensorsGuard', True, invalid=False, inconsistent=False)

        self._set_enginery_guard(0, False, 'FireSensorsGuard', None, invalid=True, inconsistent=True)
        self._set_enginery_guard(6, False, 'IntruderSensorsGuard', None, invalid=True, inconsistent=True)

        self._set_enginery_guard(10, False, 'LeakageSensorsGuard', None, invalid=False, inconsistent=True)
        self._set_enginery_guard(8, None, 'LeakageSensorsGuard', None, invalid=True, inconsistent=True)
        self._set_enginery_guard(9, None)
        self._set_enginery_guard(10, None, 'LeakageSensorsGuard', True, invalid=True, inconsistent=False)

    def _set_enginery_triggered(self, enginery_id, value, expected_unit=None, expected_value=None, **attrs):
        e = self.adapter._entities[enginery_id]
        assert isinstance(e, Enginery)
        e.values['Triggered'] = value
        self.sb.on_level_changed(e, 'Triggered')
        if expected_unit:
            self._check_publish_state(expected_unit, expected_value, attrs=attrs)

        self._check_end_of_publishes()

    def test_fire_triggered_changing(self):
        self._set_enginery_triggered(0, False, 'FireSensorsTriggered', False, invalid=True, inconsistent=False)
        self._set_enginery_triggered(1, False)
        self._set_enginery_triggered(2, False)
        self._set_enginery_triggered(3, False, 'FireSensorsTriggered', False, invalid=False, inconsistent=False)

        self._set_enginery_triggered(4, True, 'IntruderSensorsTriggered', True, invalid=True, inconsistent=False)
        self._set_enginery_triggered(10, True, 'LeakageSensorsTriggered', True, invalid=True, inconsistent=False)

        self._set_enginery_triggered(3, True, 'FireSensorsTriggered', True, invalid=False, inconsistent=True)
        self._set_enginery_triggered(0, None, 'FireSensorsTriggered', True, invalid=True, inconsistent=True)
        self._set_enginery_triggered(1, None)
        self._set_enginery_triggered(2, None, 'FireSensorsTriggered', True, invalid=True, inconsistent=False)
        self._set_enginery_triggered(3, None, 'FireSensorsTriggered', None, invalid=True, inconsistent=False)

    def test_intruder_triggered_changing(self):
        self._set_enginery_triggered(4, False, 'IntruderSensorsTriggered', False, invalid=True, inconsistent=False)
        self._set_enginery_triggered(5, False)
        self._set_enginery_triggered(6, False)
        self._set_enginery_triggered(7, False, 'IntruderSensorsTriggered', False, invalid=False, inconsistent=False)

        self._set_enginery_triggered(0, True, 'FireSensorsTriggered', True, invalid=True, inconsistent=False)
        self._set_enginery_triggered(10, True, 'LeakageSensorsTriggered', True, invalid=True, inconsistent=False)

        self._set_enginery_triggered(7, True, 'IntruderSensorsTriggered', True, invalid=False, inconsistent=True)
        self._set_enginery_triggered(4, None, 'IntruderSensorsTriggered', True, invalid=True, inconsistent=True)
        self._set_enginery_triggered(5, None)
        self._set_enginery_triggered(6, None, 'IntruderSensorsTriggered', True, invalid=True, inconsistent=False)
        self._set_enginery_triggered(7, None, 'IntruderSensorsTriggered', None, invalid=True, inconsistent=False)

    def test_leakage_triggered_changing(self):
        self._set_enginery_triggered(8, False, 'LeakageSensorsTriggered', False, invalid=True, inconsistent=False)
        self._set_enginery_triggered(9, False)
        self._set_enginery_triggered(10, False)
        self._set_enginery_triggered(11, False, 'LeakageSensorsTriggered', False, invalid=False, inconsistent=False)

        self._set_enginery_triggered(0, True, 'FireSensorsTriggered', True, invalid=True, inconsistent=False)
        self._set_enginery_triggered(6, True, 'IntruderSensorsTriggered', True, invalid=True, inconsistent=False)

        self._set_enginery_triggered(11, True, 'LeakageSensorsTriggered', True, invalid=False, inconsistent=True)
        self._set_enginery_triggered(8, None, 'LeakageSensorsTriggered', True, invalid=True, inconsistent=True)
        self._set_enginery_triggered(9, None)
        self._set_enginery_triggered(10, None, 'LeakageSensorsTriggered', True, invalid=True, inconsistent=False)
        self._set_enginery_triggered(11, None, 'LeakageSensorsTriggered', None, invalid=True, inconsistent=False)

    def _check_guard(self, funit, i):
        for value in True, False:
            funit(self.sb, Variable(value)).set()
            self._check_publish_command(1000+i, 'Guard', value)
            self._check_publish_command(1100+i, 'Guard', value)
            self._check_publish_command(1200+i, 'On', value)

            self._check_end_of_publishes()

    def test_fire_guard_setting(self):
        self._check_guard(alarming.FireSensorsGuard, 0)

    def test_intruder_guard_setting(self):
        self._check_guard(alarming.IntruderSensorsGuard, 1)

    def test_leakage_guard_setting(self):
        self._check_guard(alarming.LeakageSensorsGuard, 2)

    def test_fire_triggered_setting(self):
        funit = alarming.FireSensorsTriggered(self.adapter._entities[0], Variable(True))
        self.assertRaises(alarming.FireSensorsTriggered.NotImplemented, funit.set)

    def test_intruder_triggered_setting(self):
        funit = alarming.IntruderSensorsTriggered(self.adapter._entities[3], Variable(True))
        self.assertRaises(alarming.IntruderSensorsTriggered.NotImplemented, funit.set)

    def test_leakage_triggered_setting(self):
        funit = alarming.LeakageSensorsTriggered(self.adapter._entities[6], Variable(True))
        self.assertRaises(alarming.LeakageSensorsTriggered.NotImplemented, funit.set)
