import copy
from time import sleep
from typing import Any
from unittest import TestCase, SkipTest

from spread_core.bam.subgineries.subginery import Subginery

from spread_core.bam.engineries import lighting_area as LA
from spread_core.bam.engineries.lighting_area import ButtonPreset, ButtonAction, Pause
from spread_core.mqtt.spread.address import ProviderAddress
from spread_core.mqtt.spread.topic import State
from spread_core.mqtt.spread.variable import Variable
from spread_core.tests.equipment.utils import TestEquipmentAdapter, TestEngineryRecipe

DEFAULT_PROFILES = {
    "Profile1": {
        "TuningType": "Combo",
        "OccupancyLevel": 100,
        "VacancyLevel": 0,
        "TargetLuminosity": 0,
        "Hysteresis": 10,
        "TuningSpeed": 500,
        "OccupancyAction": "MaxLevel",
        "OccupancyScene": 2,
        "VacancyAction": "Off",
        "VacancyScene": 1,
        "HoldTime": 600
    },
    "Profile2": {
        "TuningType": "Combo",
        "OccupancyLevel": 100,
        "VacancyLevel": 0,
        "TargetLuminosity": 0,
        "Hysteresis": 10,
        "TuningSpeed": 500,
        "OccupancyAction": "MaxLevel",
        "OccupancyScene": 2,
        "VacancyAction": "Off",
        "VacancyScene": 1,
        "HoldTime": 600
    },
    "Profile3": {
        "TuningType": "Combo",
        "OccupancyLevel": 100,
        "VacancyLevel": 0,
        "TargetLuminosity": 0,
        "Hysteresis": 10,
        "TuningSpeed": 500,
        "OccupancyAction": "MaxLevel",
        "OccupancyScene": 2,
        "VacancyAction": "Off",
        "VacancyScene": 1,
        "HoldTime": 600
    },
    "Profile4": {
        "TuningType": "Combo",
        "OccupancyLevel": 100,
        "VacancyLevel": 0,
        "TargetLuminosity": 0,
        "Hysteresis": 10,
        "TuningSpeed": 500,
        "OccupancyAction": "MaxLevel",
        "OccupancyScene": 2,
        "VacancyAction": "Off",
        "VacancyScene": 1,
        "HoldTime": 600
    },
    "Profile5": {
        "TuningType": "Combo",
        "OccupancyLevel": 100,
        "VacancyLevel": 0,
        "TargetLuminosity": 0,
        "Hysteresis": 10,
        "TuningSpeed": 500,
        "OccupancyAction": "MaxLevel",
        "OccupancyScene": 2,
        "VacancyAction": "Off",
        "VacancyScene": 1,
        "HoldTime": 600
    }
}


def profiles_with_value(profile_name, item_name, item_value, profiles=None):
    if profiles is None:
        profiles = copy.deepcopy(DEFAULT_PROFILES)

    profiles.setdefault(profile_name, {})[item_name] = item_value

    return profiles


class _TestLightingArea(TestCase):
    @property
    def values(self):
        return self.la.values

    @property
    def is_on(self):
        return self.la.is_on

    @property
    def actuators(self):
        return self.la.actuators

    @property
    def luminosities(self):
        return self.la.luminosities

    @property
    def presences(self):
        return self.la.presences

    @property
    def buttons(self):
        return self.la.buttons

    def set_recipe(self, ingredient_type, recipe):
        self.la.set_recipe(ingredient_type, recipe)

    def load_scene(self, scene_type):
        self.la.load_scene(scene_type)

    def setUp(self) -> None:
        self.adapter = TestEquipmentAdapter()

        la = LA.LightingArea(1, name="Test LA")
        la.get_info()

        la.adapter = self.adapter

        la.actuators = {i: TestEngineryRecipe(i) for i in range(5)}
        la.presences = {i: TestEngineryRecipe(i) for i in range(5, 10)}
        la.luminosities = {i: TestEngineryRecipe(i) for i in range(10, 15)}
        la.buttons = {i: TestEngineryRecipe(i) for i in range(15, 20)}

        self.la = la

    def _check_value(self, funit, expected_value):
        self.assertEqual(self.values.get(funit), expected_value)

    def _check_end_of_publishes(self):
        if self.adapter.queue:
            unexpected_publishes = f'Unexpected publishes({len(self.adapter.queue)} pcs):\n'
            unexpected_publishes += "\n".join(str(i) for i in self.adapter.queue)
            self.fail(unexpected_publishes)

    def __check_publish(self, _topic, _value, _retain=True, _delay=0, params=None):
        if len(self.adapter.queue) == 0:
            self.fail('Expected publish is absent: %s' % _topic)
        else:
            topic, variable, retain, delay = self.adapter.queue.popleft()
            self.assertEqual(_topic, topic)
            self.assertEqual(_value, variable.value)
            self.assertEqual(_retain, retain)
            self.assertEqual(_delay, delay)

            if params:
                for param_name in params:
                    try:
                        self.assertEqual(params[param_name], variable.params[param_name])
                    except KeyError:
                        self.fail('Expected variable parameter "%s" is absent' % param_name)

    def _check_publish_state(self, funit, value, delay=0, params=None):
        self.__check_publish('Spread/State/1/Equipment/LightingArea/1/%s' % funit, value, _retain=True, _delay=delay, params=params)

    def _check_publish_state_la2(self, funit, value, delay=0, params=None):
        self.__check_publish('Spread/State/1/Equipment/LightingArea/2/%s' % funit, value, _retain=True, _delay=delay, params=params)

    def _check_publish_command(self, p_id, funit, value, params=None):
        self.__check_publish('Spread/Set/Provider/%s/%s' % (p_id, funit), value, _retain=False, _delay=0, params=params)

    def _write_value_to_profile(self, funit_name, value, profile_name='Profile1'):
        if value is None:
            self.la.values['Profiles'][profile_name].pop(funit_name, None)
        else:
            self.la.values['Profiles'][profile_name][funit_name] = value

    def _compare_value_from_profile(self, profile_name, funit_name, value):
        self.assertEqual(value, self.la.values['Profiles'][profile_name][funit_name])

    def _compare_value_from_current_profile(self, funit_name, value):
        self.assertEqual(value, self.la.get_value(funit_name))


class TestLightingArea(_TestLightingArea):
    def test_on_selector(self):
        self.values.pop('On', None)
        self.values.pop('Pause', None)
        self.assertEqual(self.is_on, False)

        self.values['Pause'] = True
        self.assertEqual(self.is_on, False)

        self.values['Pause'] = False
        self.assertEqual(self.is_on, False)

        self.values['On'] = True
        self.values.pop('Pause', None)
        self.assertEqual(self.is_on, True)

        self.values['Pause'] = True
        self.assertEqual(self.is_on, False)

        self.values['Pause'] = False
        self.assertEqual(self.is_on, True)

        self.values['On'] = False
        self.values.pop('Pause', None)
        self.assertEqual(self.is_on, False)

        self.values['Pause'] = True
        self.assertEqual(self.is_on, False)

        self.values['Pause'] = False
        self.assertEqual(self.is_on, False)

    def test_subginery_setting(self):
        sb = Subginery(99, 999, [1])
        self.la.set_subginery(sb)

        self.assertIn(sb, self.la._subgineries)

    def test_event_storing(self):
        is_registered = self.la.is_button_event_registered
        register = self.la.register_button_event

        register(999, "SomeEvent0")

        for i in range(4):
            register(999, "SomeEvent")
            self.assertEqual(i % 2 == 0, is_registered(999, "SomeEvent"))
            self.assertEqual(True, is_registered(999, "SomeEvent0"))
            self.assertEqual(False, is_registered(999, "SomeEvent1"))

        register(999, "LongPressRepeat")
        self.assertEqual(False, is_registered(999, "LongPressRepeat"))

        register(999, "LongPressStart")
        self.assertEqual(True, is_registered(999, "LongPressRepeat"))

        register(999, "LongPressStop")
        self.assertEqual(False, is_registered(999, "LongPressRepeat"))

    def test_la_bonding(self):
        def is_light_bonded():
            return self.la.is_light_bonded

        self.actuators[0]._engineries = []
        self.assertEqual(False, is_light_bonded())
        self.actuators[0]._engineries = [1]
        self.assertEqual(True, is_light_bonded())

    def test_la_actuators_state(self):
        def is_any_actuator_is_on():
            return self.la.is_any_actuator_is_on

        self.actuators[0]._level = None
        self.assertEqual(False, is_any_actuator_is_on())
        self.actuators[0]._level = 0
        self.assertEqual(False, is_any_actuator_is_on())
        self.actuators[0]._level = 1
        self.assertEqual(True, is_any_actuator_is_on())

    def test_recipe_appending_actuation(self):
        recipe = TestEngineryRecipe(1001)
        self.set_recipe('Actuation', recipe)
        self.assertIs(self.actuators[1001], recipe)

    def test_recipe_appending_luminosity(self):
        recipe = TestEngineryRecipe(1001)
        self.set_recipe('Luminosity', recipe)
        self.assertIs(self.luminosities[1001], recipe)

    def test_recipe_appending_presence(self):
        recipe = TestEngineryRecipe(1001)
        self.set_recipe('Presence', recipe)
        self.assertIs(self.presences[1001], recipe)

    def test_recipe_appending_button(self):
        recipe = TestEngineryRecipe(1001)
        self.set_recipe('Button', recipe)
        self.assertIs(self.buttons[1001], recipe)

    def test_recipe_appending_unknown(self):
        recipe = TestEngineryRecipe(1001)
        with self.assertRaises(ValueError) as ex:
            self.set_recipe('UnknownRecipeType', recipe)

        self.assertEqual(ex.exception.args[0], 'unknown ingredient type: "UnknownRecipeType"')

    def test_provider_command_sending(self):
        recipe = TestEngineryRecipe(1001)
        self.la.provider_command(recipe, 'SomeFunit', 123)
        self._check_publish_command(1001, "SomeFunit", 123)
        self._check_end_of_publishes()

    def test_info(self):
        self.la.values = {}
        self.la.get_info()

        self._check_publish_state("On", False)
        self._check_publish_state("Discovery", False)
        self._check_publish_state("CurrentProfile", "Profile1")
        self._check_publish_state("Pause", False)
        self._check_publish_state("Buttons", [])
        self._check_publish_state("Profiles", DEFAULT_PROFILES)
        self._check_publish_state("TuningType", "Combo")
        self._check_publish_state("OccupancyLevel", 100)
        self._check_publish_state("VacancyLevel", 0)
        self._check_publish_state("TargetLuminosity", 0)
        self._check_publish_state("Hysteresis", 10)
        self._check_publish_state("TuningSpeed", 500)
        self._check_publish_state("OccupancyAction", "MaxLevel")
        self._check_publish_state("OccupancyScene", 2)
        self._check_publish_state("VacancyAction", "Off")
        self._check_publish_state("VacancyScene", 1)
        self._check_publish_state("HoldTime", 600)

        self._check_end_of_publishes()

    def test_profile_loading(self):
        new_profile = {
            "TuningType": "SomeType",
            "OccupancyLevel": 991,
            "VacancyLevel": 992,
            "TargetLuminosity": 993,
            "Hysteresis": 994,
            "TuningSpeed": 995,
            "OccupancyAction": "SomeAction",
            "OccupancyScene": 996,
            "VacancyAction": "SomeAction2",
            "VacancyScene": 997,
            "HoldTime": 998

        }

        self.values['Profiles']['Profile3'].update(new_profile)
        self.la.load_profile("Profile3")

        self._check_publish_state("CurrentProfile", "Profile3")
        for key, value in new_profile.items():
            self._check_publish_state(key, value)

        self._check_publish_command(5, 'EngineryHoldTime', 998)
        self._check_publish_command(6, 'EngineryHoldTime', 998)
        self._check_publish_command(7, 'EngineryHoldTime', 998)
        self._check_publish_command(8, 'EngineryHoldTime', 998)
        self._check_publish_command(9, 'EngineryHoldTime', 998)

        self._check_end_of_publishes()

    def test_level_setter(self):
        self.la.set_level(99, method='SomeMethod')
        for recipe_id in self.actuators:
            self._check_publish_command(recipe_id, 'BrightnessLevel', 99, params={'method': 'SomeMethod'})
        self._check_end_of_publishes()

    def test_tuning_setter(self):
        set_tuning = self.la.set_tuning

        def check_tuning(value, expected_commands=True):
            if expected_commands:
                for recipe_id in self.actuators:
                    self._check_publish_command(recipe_id, 'Tuning', value, params={'speed': 10})

            self._check_end_of_publishes()
            return value == self.la.tuning

        self._write_value_to_profile('TuningSpeed', 10000)

        set_tuning(1)
        check_tuning(1)

        set_tuning(1)
        check_tuning(1, expected_commands=False)

        set_tuning(2)
        check_tuning(1, expected_commands=False)

        set_tuning(-1)
        check_tuning(-1)


class TestSceneSaving(_TestLightingArea):
    def save_scene(self, scene_type):
        self.la.save_scene(scene_type)

    def setUp(self) -> None:
        super(TestSceneSaving, self).setUp()
        self.SCENES = []

        provider_id = 0
        for recipe in self.actuators.values():
            if provider_id == 3:
                recipe._level = None
            else:
                recipe._level = 25 + provider_id
                self.SCENES.append({'id': provider_id, 'level': recipe._level})

            provider_id += 1

    def test_occupancy(self):
        self.values["Profiles"]["Profile1"]['OccupancyScene'] = 14
        self.save_scene('OccupancyScene')

        self._check_publish_command(0, 'SceneLevelsRaw', [{'index': 14, 'value': 25}])
        self._check_publish_command(1, 'SceneLevelsRaw', [{'index': 14, 'value': 26}])
        self._check_publish_command(2, 'SceneLevelsRaw', [{'index': 14, 'value': 27}])
        self._check_publish_command(4, 'SceneLevelsRaw', [{'index': 14, 'value': 29}])
        self._check_end_of_publishes()

    def test_vacancy(self):
        self.values["Profiles"]["Profile1"]['VacancyScene'] = 13
        self.save_scene('VacancyScene')

        self._check_publish_command(0, 'SceneLevelsRaw', [{'index': 13, 'value': 25}])
        self._check_publish_command(1, 'SceneLevelsRaw', [{'index': 13, 'value': 26}])
        self._check_publish_command(2, 'SceneLevelsRaw', [{'index': 13, 'value': 27}])
        self._check_publish_command(4, 'SceneLevelsRaw', [{'index': 13, 'value': 29}])
        self._check_end_of_publishes()

    def test_index(self):
        self.save_scene(12)

        self._check_publish_command(0, 'SceneLevelsRaw', [{'index': 12, 'value': 25}])
        self._check_publish_command(1, 'SceneLevelsRaw', [{'index': 12, 'value': 26}])
        self._check_publish_command(2, 'SceneLevelsRaw', [{'index': 12, 'value': 27}])
        self._check_publish_command(4, 'SceneLevelsRaw', [{'index': 12, 'value': 29}])
        self._check_end_of_publishes()

    def test_occupancy_custom(self):
        self.values["Profiles"]["Profile1"]['OccupancyScene'] = 255
        self.save_scene('OccupancyScene')
        profiles = profiles_with_value('Profile1', 'OccupancyScene', 255)
        profiles = profiles_with_value('Profile1', 'OccupancySceneLevels', self.SCENES, profiles=profiles)

        self._check_publish_state('OccupancySceneLevels', self.SCENES)
        self._check_publish_state('Profiles', profiles)

        self._check_end_of_publishes()

    def test_vacancy_custom(self):
        self.values["Profiles"]["Profile1"]['VacancyScene'] = 255
        self.save_scene('VacancyScene')
        profiles = profiles_with_value('Profile1', 'VacancyScene', 255)
        profiles = profiles_with_value('Profile1', 'VacancySceneLevels', self.SCENES, profiles=profiles)

        self._check_publish_state('VacancySceneLevels', self.SCENES)
        self._check_publish_state('Profiles', profiles)

        self._check_end_of_publishes()


class TestSceneLoading(_TestLightingArea):
    def setUp(self) -> None:
        super(TestSceneLoading, self).setUp()
        self.SCENES = [
            {'id': 0, 'level': 25}, {'id': 1, 'level': 26}, {'id': 2, 'level': 27}, {'id': 4, 'level': 29}
        ]

    def check_custom(self, scene_name, scenes: [list, None]):
        self.values["Profiles"]["Profile1"][scene_name] = 255

        if scenes is None:
            self.values["Profiles"]["Profile1"].pop(f'{scene_name}Levels', None)
        else:
            self.values["Profiles"]["Profile1"][f'{scene_name}Levels'] = scenes

        self.load_scene(scene_name)

        for scene_item in scenes or []:
            self._check_publish_command(scene_item["id"], 'BrightnessLevel', scene_item["level"])

        self._check_end_of_publishes()

    def test_occupancy_empty(self):
        self.check_custom('OccupancyScene', None)
        self.check_custom('OccupancyScene', [])
        self.check_custom('OccupancyScene', self.SCENES)
        self.check_custom('C1', self.SCENES)

    def test_vacancy(self):
        self.check_custom('VacancyScene', None)
        self.check_custom('VacancyScene', [])
        self.check_custom('VacancyScene', self.SCENES)
        self.check_custom('C2', self.SCENES)

    def test_index_invalid(self):
        self.assertRaises(AssertionError, self.load_scene, 'qwerty')
        self.assertRaises(AssertionError, self.load_scene, -1)
        self.assertRaises(AssertionError, self.load_scene, 16)

    def test_index(self):
        self.load_scene(12)

        self._check_publish_command(0, 'BrightnessLevel', 12, params={'method': 'Scene'})
        self._check_publish_command(1, 'BrightnessLevel', 12, params={'method': 'Scene'})
        self._check_publish_command(2, 'BrightnessLevel', 12, params={'method': 'Scene'})
        self._check_publish_command(3, 'BrightnessLevel', 12, params={'method': 'Scene'})
        self._check_publish_command(4, 'BrightnessLevel', 12, params={'method': 'Scene'})
        self._check_end_of_publishes()


class TestStateHandling(_TestLightingArea):
    def emulate_state(self, provider_id, funit, value):
        topic = State(ProviderAddress(1, 'Manager', 1, 'Provider', provider_id, funit))
        variable = Variable(value)
        self.la.on_state(topic, variable)

    def test_presence_event(self):
        self.emulate_state(6, 'CurrentPresence', True)
        self._check_publish_state('Presence', True)
        self._check_end_of_publishes()

        self.emulate_state(7, 'CurrentPresence', True)
        self._check_end_of_publishes()

        self.emulate_state(6, 'CurrentPresence', False)
        self._check_end_of_publishes()

        self.emulate_state(7, 'CurrentPresence', False)
        self._check_publish_state('Presence', False)
        self._check_end_of_publishes()

    def test_luminosity_event(self):
        self.emulate_state(11, 'CurrentLuminosity', 123)
        self._check_publish_state('Luminosity', 123)
        self._check_end_of_publishes()

        self.emulate_state(12, 'CurrentLuminosity', 236)
        self._check_publish_state('Luminosity', round((123+236) / 2))
        self._check_end_of_publishes()

        self.emulate_state(12, 'On', False)
        self._check_publish_state('Luminosity', 123)
        self._check_end_of_publishes()

        self.emulate_state(11, 'On', False)
        self._check_publish_state('Luminosity', None)
        self._check_end_of_publishes()


class TestPresenceHandling(_TestLightingArea):
    def emulate_presence(self, presence_value: bool):
        self.la.presence_action(presence_value)

    def _check_scene_calling(self, presence_value, scene_name):
        def compare_scene_loader(*args, **kwargs):
            self.assertEqual(args[0], scene_name)
            self.assertEqual(kwargs['session'], 'PresenceHandling')

        self.la.load_scene = compare_scene_loader
        self.emulate_presence(presence_value)
        self._check_end_of_publishes()

    def _check_level_calling(self, presence_value, level, params=None):
        self.emulate_presence(presence_value)

        for provider_id in self.actuators:
            self._check_publish_command(provider_id, 'BrightnessLevel', level, params)

        self._check_end_of_publishes()

    def test_scene(self):
        self._write_value_to_profile('OccupancyAction', 'Scene')
        self._check_scene_calling(True, 'OccupancyScene')

        self._write_value_to_profile('VacancyAction', 'Scene')
        self._check_scene_calling(False, 'VacancyScene')

    def test_level_exact(self):
        self._write_value_to_profile('OccupancyAction', 'Level')
        self._write_value_to_profile('OccupancyLevel', 56)
        self._check_level_calling(True, 56)

        self._write_value_to_profile('VacancyAction', 'Level')
        self._write_value_to_profile('VacancyLevel', 65)
        self._check_level_calling(False, 65)

    def test_level_border(self):
        self._write_value_to_profile('OccupancyAction', 'MaxLevel')
        self._check_level_calling(True, 100, {'method': 'RecallMaxLevel'})

        self._write_value_to_profile('VacancyAction', 'MinLevel')
        self._check_level_calling(False, 1, {'method': 'RecallMinLevel'})

    def test_level_last(self):
        self._write_value_to_profile('OccupancyAction', 'LastLevel')
        self._check_level_calling(True, 50, {'method': 'GoToLastActiveLevel'})

        self._write_value_to_profile('VacancyAction', 'Off')
        self._check_level_calling(False, 0, {'method': 'Off'})


class TestLuminosityHandling(_TestLightingArea):
    def setUp(self) -> None:
        super(TestLuminosityHandling, self).setUp()
        self.la.values['On'] = True
        self._write_value_to_profile('TuningSpeed', 10000)
        self._write_value_to_profile('TargetLuminosity', 50)

    def _emulate_state(self, provider_id, funit, value):
        topic = State(ProviderAddress(1, 'Manager', 1, 'Provider', provider_id, funit))
        variable = Variable(value)
        self.la.on_state(topic, variable)

    def _check_tuning_switching(self, value, expected_commands=True):
        if expected_commands:
            for recipe_id in self.actuators:
                self._check_publish_command(recipe_id, 'Tuning', value, params={'speed': 10})

        self.assertEqual(value, self.la.tuning)

    def _check_level_calling(self, level, params=None):
        for provider_id in self.actuators:
            self._check_publish_command(provider_id, 'BrightnessLevel', level, params)

    def test_luminosity_handling_of_vacancy(self):
        self._write_value_to_profile('TuningType', 'Combo')

        self._emulate_state(10, 'On', True)
        self._check_end_of_publishes()

        self._emulate_state(10, 'CurrentLuminosity', 73)
        self._check_publish_state('Luminosity', 73)
        self._check_end_of_publishes()

    def test_luminosity_handling_of_occupied(self):
        self._write_value_to_profile('TuningType', 'Combo')

        self._emulate_state(6, 'On', True)
        self._emulate_state(6, 'CurrentPresence', True)
        self._check_publish_state('Presence', True)
        self._check_level_calling(100, params={"method": "RecallMaxLevel"})
        self._check_end_of_publishes()

        self._emulate_state(10, 'CurrentLuminosity', 73)
        self._check_publish_state('Luminosity', 73)
        self._check_tuning_switching(-0.46)
        self._check_end_of_publishes()

    def test_luminosity_handling(self):
        self._write_value_to_profile('TuningType', 'Luminosity')

        self._emulate_state(10, 'On', True)
        self._check_end_of_publishes()

        self._emulate_state(10, 'CurrentLuminosity', 73)
        self._check_publish_state('Luminosity', 73)
        self._check_tuning_switching(-0.46)
        self._check_end_of_publishes()

        self._emulate_state(10, 'CurrentLuminosity', 60)
        self._check_publish_state('Luminosity', 60)
        self._check_end_of_publishes()

        self._emulate_state(11, 'CurrentLuminosity', 63)
        self._check_publish_state('Luminosity', 62)
        self._check_end_of_publishes()

        self._emulate_state(12, 'CurrentLuminosity', 2)
        self._check_publish_state('Luminosity', 42)
        self._check_tuning_switching(0.16)
        self._check_end_of_publishes()

        self._emulate_state(10, 'On', False)
        self._check_publish_state('Luminosity', 32)
        self._check_end_of_publishes()

        self._emulate_state(12, 'On', False)
        self._check_publish_state('Luminosity', 63)
        self._check_tuning_switching(-0.26)
        self._check_end_of_publishes()

        self._emulate_state(11, 'On', False)
        self._check_tuning_switching(0)
        self._check_publish_state('Luminosity', None)
        self._check_end_of_publishes()


class TestButtonHandling(_TestLightingArea):
    def setUp(self) -> None:
        super(TestButtonHandling, self).setUp()

        self.adapter.button_presets = {
            0: ButtonPreset(0).update({
                "isToggle": False,
                'handlers': [
                    [],
                    [{"event": "ShortPress", "command": "Off"} for action_index in range(2)],
                ]
            }),
            1: ButtonPreset(1).update({
                "isToggle": False,
                'handlers': [
                    [{"event": "ShortPress", "command": "Off"} for action_index in range(2)],
                    [{"event": "ShortPress", "command": "Off"} for action_index in range(2)],
                ]
            }),
            2: ButtonPreset(1).update({
                "isToggle": True,
                'handlers': [
                    [{"event": "ShortPress", "command": "Off"} for action_index in range(2)],
                    []
                ]
            }),
            3: ButtonPreset(2).update({
                "isToggle": True,
                'handlers': [
                    [{"event": "ShortPress", "command": "Off"} for action_index in range(2)],
                    [{"event": "ShortPress", "command": "Off"} for action_index in range(2)],
                ]
            })
        }

        self.values['Buttons'] = [
            {'index': 0, 'preset': 0},  # 15
            {'index': 1, 'preset': 1},  # 16
            {'index': 2, 'preset': 2},  # 17
            {'index': 3, 'preset': 3},  # 18
            {'index': 4, 'preset': 99}  # 19
        ]

    def _select_preset(self, preset_id, handler_index, as_alt=False):
        action = self.adapter.button_presets[preset_id].handlers[handler_index][1]

        action.event = "SomeEvent"
        action.command = "AltCommand" if as_alt else "MainCommand"

    def emulate_event(self, provider_id: int, event_value: str):
        self.la.on_button_event(provider_id, Variable(event_value))

    def _test_event(self, providert_id, alt_expected=False):
        self.handled = False

        def play_action_handler(action):
            self.assertEqual(action.event, "SomeEvent")
            self.assertEqual(action.command, "AltCommand" if alt_expected else "MainCommand")
            self.handled = True

        self.la.play_button_action = play_action_handler

        self.emulate_event(providert_id, 'SomeEvent')
        self._check_end_of_publishes()
        if not self.handled:
            self.fail('Event not handled!')

    def test_unexistent_preset(self):
        with self.assertRaises(AssertionError):
            self.emulate_event(19, 'ShortPress')

        self._check_end_of_publishes()

    def test_unusable_preset(self):
        self.values['Buttons'].pop()
        with self.assertRaises(AssertionError) as ex:
            self._test_event(19)

        self.assertEqual(ex.exception.args[0], 'Event not handled!')

    def test_non_toggle_no_handler(self):
        with self.assertRaises(AssertionError) as ex:
            self._test_event(15)
        self.assertEqual(ex.exception.args[0], 'Event not handled!')

    def test_non_toggle(self):
        self._select_preset(1, 0)
        self._test_event(16)

    def test_toggle_no_alt_handler(self):
        self._select_preset(2, 0)
        self._test_event(17)

    def test_toggle(self):
        self._select_preset(3, 0)
        self._select_preset(3, 1, as_alt=True)

        self._test_event(18)
        self._test_event(18, alt_expected=True)


class TestButtonActionHandling(_TestLightingArea):
    @classmethod
    def setUpClass(cls) -> None:
        def create_action(command, scene=None, level=None, set_la=None, stop_tuning=None):
            data = {
                "event": "ShortPress",
                "command": command,
            }

            if scene is not None:
                data['scene'] = scene

            if level is not None:
                data['level'] = level

            if set_la is not None:
                data['setLightingArea'] = set_la

            if stop_tuning is not None:
                data['stopTuning'] = stop_tuning

            return ButtonAction(data)

        global ACTION
        ACTION = create_action

    def _check_level_calling(self, action, level, params=None, on_value: Any = 'not used', tuning_value=None):
        self.apply_action(action)

        if on_value != 'not used':
            self.assertEqual(self.values['On'], on_value)
            self._check_publish_state('On', on_value)

        if tuning_value is not None:
            self.assertEqual(self.la.tuning, 0)

            for provider_id in self.actuators:
                self._check_publish_command(provider_id, 'Tuning', 0)

        for provider_id in self.actuators:
            self._check_publish_command(provider_id, 'BrightnessLevel', level, params)

        self._check_end_of_publishes()

    def _check_presence_emulating(self, command_name, presence_value):
        def compare_emulating(value):
            self.assertEqual(value, presence_value)

        self.la.presence_action = compare_emulating
        self.apply_action(ACTION(command_name))
        self._check_publish_state('Presence', presence_value)

    def apply_action(self, action):
        self.la.play_button_action(action)

    def test_scene(self):
        self._check_level_calling(ACTION("Scene", scene=11), 11, params={'method': 'Scene'})

    def test_level(self):
        self._check_level_calling(ACTION("Level", level=88), 88)

    def test_level_min(self):
        self._check_level_calling(ACTION("MinLevel"), 1, params={'method': 'RecallMinLevel'})

    def test_level_last(self):
        self._check_level_calling(ACTION("LastLevel"), 50, params={'method': 'GoToLastActiveLevel'})

    def test_level_off(self):
        self._check_level_calling(ACTION("Off"), 0, params={'method': 'Off'})

    def test_level_step(self):
        for command in 'Up', 'Down', 'StepUp', 'StepDown', 'OnAndStepUp', 'StepDownAndOff':
            self._check_level_calling(ACTION(command), -1, params={'method': command})

    def test_emulate_occupancy(self):
        self._check_presence_emulating("EmulateOccupancy", True)

    def test_emulate_vacancy(self):
        self._check_presence_emulating("EmulateVacancy", False)

    def test_la(self):
        self._check_level_calling(ACTION("Level", level=88, set_la='On'), 88, on_value=True)
        self._check_level_calling(ACTION("Level", level=88, set_la='On'), 88)
        self._check_level_calling(ACTION("Level", level=88, set_la='Off'), 88, on_value=False)
        self._check_level_calling(ACTION("Level", level=88, set_la='Off'), 88)
        self._check_level_calling(ACTION("Level", level=88, set_la='On'), 88, on_value=True)
        self._check_level_calling(ACTION("Level", level=88, set_la='Off'), 88, on_value=False)

    def test_tuning(self):
        self.la.tuning = 1
        self._check_level_calling(ACTION("Level", level=88, stop_tuning=True), 88, tuning_value=0)
        self._check_level_calling(ACTION("Level", level=88, stop_tuning=True), 88)


class TestPause(_TestLightingArea):
    def setUp(self):
        super(TestPause, self).setUp()
        Pause.pause_timeout = 1

    def emulate_state(self, la, value):
        Pause(la, value).set()

    def test_pause_timer_work(self):
        self.assertIsNone(self.la.pause_timer)
        self.emulate_state(self.la, True)
        self._check_publish_state('Pause', True)
        self._check_end_of_publishes()
        self.assertTrue(self.la.values['Pause'])
        self.assertIsNotNone(self.la.pause_timer)
        self.assertTrue(self.la.pause_timer.is_alive())
        timer = self.la.pause_timer

        sleep(Pause.pause_timeout+1)

        self._check_publish_state('Pause', False)
        self._check_end_of_publishes()
        self.assertFalse(self.la.values['Pause'])
        self.assertIsNone(self.la.pause_timer)
        self.assertFalse(timer.is_alive())

    def test_pause_timer_break(self):
        self.assertIsNone(self.la.pause_timer)
        self.emulate_state(self.la, True)
        self._check_publish_state('Pause', True)
        self._check_end_of_publishes()
        self.assertTrue(self.la.values['Pause'])
        self.assertIsNotNone(self.la.pause_timer)
        self.assertTrue(self.la.pause_timer.is_alive())
        timer = self.la.pause_timer

        self.emulate_state(self.la, False)
        self._check_publish_state('Pause', False)
        self._check_end_of_publishes()
        self.assertFalse(self.la.values['Pause'])
        self.assertIsNone(self.la.pause_timer)
        sleep(0.1)
        self.assertFalse(timer.is_alive())

        sleep(Pause.pause_timeout+1)

        self.assertFalse(self.la.values['Pause'])
        self._check_end_of_publishes()

    def test_pause_timer_reset(self):
        self.assertIsNone(self.la.pause_timer)
        self.emulate_state(self.la, True)
        self._check_publish_state('Pause', True)
        self._check_end_of_publishes()
        self.assertTrue(self.la.values['Pause'])
        self.assertIsNotNone(self.la.pause_timer)
        self.assertTrue(self.la.pause_timer.is_alive())
        timer1 = self.la.pause_timer

        self.emulate_state(self.la, True)
        self._check_end_of_publishes()
        self.assertTrue(self.la.values['Pause'])
        self.assertFalse(timer1.is_alive())
        self.assertIsNotNone(self.la.pause_timer)
        self.assertTrue(self.la.pause_timer.is_alive())

    def test_pause_timer_several_la(self):
        la2 = LA.LightingArea(2, name="Test LA2")
        la2.get_info()
        la2.adapter = self.adapter
        self.assertIsNone(la2.pause_timer)
        self.assertFalse(la2.values['Pause'])

        self.assertIsNone(self.la.pause_timer)
        self.emulate_state(self.la, True)
        self._check_publish_state('Pause', True)
        self._check_end_of_publishes()
        self.assertTrue(self.la.values['Pause'])
        self.assertIsNotNone(self.la.pause_timer)
        self.assertTrue(self.la.pause_timer.is_alive())
        timer = self.la.pause_timer

        self.assertIsNone(la2.pause_timer)
        self.assertFalse(la2.values['Pause'])
        self.emulate_state(la2, True)
        self._check_publish_state_la2('Pause', True)
        self._check_end_of_publishes()
        self.assertTrue(la2.values['Pause'])
        self.assertIsNotNone(la2.pause_timer)
        self.assertTrue(la2.pause_timer.is_alive())
        timer2 = la2.pause_timer
        self.assertEqual(self.la.pause_timer, timer)
        self.assertNotEqual(self.la.pause_timer, la2.pause_timer)

        self.emulate_state(self.la, False)
        self._check_publish_state('Pause', False)
        self._check_end_of_publishes()
        self.assertFalse(self.la.values['Pause'])
        self.assertIsNone(self.la.pause_timer)
        sleep(0.1)
        self.assertFalse(timer.is_alive())
        self.assertTrue(la2.values['Pause'])
        self.assertTrue(la2.pause_timer.is_alive())
        self.assertEqual(la2.pause_timer, timer2)


class AbstractLightingAreaFunits(_TestLightingArea):
    funit = None
    valid_values = None
    invalid_values = []

    @classmethod
    def setUpClass(cls) -> None:
        if cls.funit is None:
            raise SkipTest(cls.__name__)

    def _check_addition_publishes(self):
        pass

    def test_invalid_values(self):
        for v in self.invalid_values:
            with self.subTest(unit=self.funit.funit_type, value=v):
                with self.assertRaises(ValueError):
                    self.funit(self.la, Variable(v)).set()

    def test_by_funit(self):
        for value in self.valid_values:
            self.la.set_value(self.funit, None)
            self._write_value_to_profile(self.funit, 'SomeValue')

            funit_instance = self.funit(self.la, Variable(value))
            funit_instance.set()

            self._compare_value_from_profile('Profile1', self.funit, value)
            self._compare_value_from_current_profile(self.funit, value)
            self._check_publish_state(self.funit, value)
            self._check_publish_state('Profiles', profiles_with_value('Profile1', self.funit, value))
            self._check_addition_publishes()
            self._check_end_of_publishes()

    def test_by_current_profile(self):
        for value in self.valid_values:
            self.la.set_value(self.funit, None)
            self._write_value_to_profile(self.funit, 'SomeValue')

            funit_instance = LA.Profiles(self.la, Variable({"Profile1": {self.funit: value}}))
            funit_instance.set()

            self._compare_value_from_profile('Profile1', self.funit, value)
            self._compare_value_from_current_profile(self.funit, value)
            self._check_publish_state(self.funit, value)
            self._check_publish_state('Profiles', profiles_with_value("Profile1", self.funit, value))
            self._check_addition_publishes()
            self._check_end_of_publishes()

    def test_by_another_profile(self):
        for value in self.valid_values:
            self._write_value_to_profile(self.funit, 'SomeValue', 'Profile3')

            funit_instance = LA.Profiles(self.la, Variable({"Profile3": {self.funit: value}}))
            funit_instance.set()

            self._compare_value_from_profile('Profile3', self.funit, value)
            self._check_publish_state('Profiles', profiles_with_value('Profile3', self.funit, value))
            self._check_end_of_publishes()


class TestOccupancyLevel(AbstractLightingAreaFunits):
    funit = LA.OccupancyLevel
    valid_values = [99]
    invalid_values = [None, -1, 256, 'qwer']


class TestVacancyLevel(AbstractLightingAreaFunits):
    funit = LA.VacancyLevel
    valid_values = [99]
    invalid_values = [None, -1, 256, 'qwer']


class TestTuningType(AbstractLightingAreaFunits):
    funit = LA.TuningType
    valid_values = funit.variants
    invalid_values = [None, 0, 'qwer']


class TestTargetLuminosity(AbstractLightingAreaFunits):
    funit = LA.TargetLuminosity
    valid_values = [99]


class TestHysteresis(AbstractLightingAreaFunits):
    funit = LA.Hysteresis
    valid_values = [99]


class TestTuningSpeed(AbstractLightingAreaFunits):
    funit = LA.TuningSpeed
    valid_values = [99]


class TestOccupancyAction(AbstractLightingAreaFunits):
    funit = LA.OccupancyAction
    valid_values = funit.variants


class TestVacancyAction(AbstractLightingAreaFunits):
    funit = LA.VacancyAction
    valid_values = funit.variants


class TestOccupancyScene(AbstractLightingAreaFunits):
    funit = LA.OccupancyScene
    valid_values = funit.variants
    invalid_values = ['qwe', None, -1, 16]


class TestVacancyScene(AbstractLightingAreaFunits):
    funit = LA.VacancyScene
    valid_values = funit.variants
    invalid_values = ['qwe', None, -1, 16]


class TestHoldTime(AbstractLightingAreaFunits):
    funit = LA.HoldTime
    valid_values = [123]

    def _check_addition_publishes(self):
        for provider_id in self.presences:
            self._check_publish_command(provider_id, 'EngineryHoldTime', 123)
