import unittest

import spread_dali as dali
from spread_core.bam.dali.providers.lights import DaliRgbwaf, Color, ChannelsCount
from spread_core.mqtt.spread.variable import Variable
from spread_core.tests.hardware.dali.test_dali_provider import TestDaliProvider
from spread_core.bam import const as C


COLOR_11 = 433
COLOR_22 = 866
COLOR_33 = 1299
COLOR_44 = 1732
COLOR_55 = 2165
COLOR_66 = 2598
COLOR_128 = 5039
COLOR_254 = 10000
COLOR_255 = None


class TestDaliRgb(TestDaliProvider):
    Provider = DaliRgbwaf


class TestRgb(TestDaliRgb):
    @unittest.skip
    def test_response_handler(self):
        self._emulate_request_response('ColorValue', 123, 10)
        self._check_state('Color', {C.RED: 255, C.GREEN: 123, C.BLUE: 255, C.WHITE: 255})
        self.manager.check_end_of_publishes(self)

        self._emulate_request_response('ColorValue', 123, 10)
        self.manager.check_end_of_publishes(self)

    def test_external_set_rgb(self):
        self._set_dtr(0, 11)
        self._set_dtr(1, 22)
        self._set_dtr(2, 33)
        self._emulate_external(dali.gear.colour_control.SetRgbColour(dali.address.Short(63)))
        self._check_state('Color', {
            C.RED: COLOR_11,
            C.GREEN: COLOR_22,
            C.BLUE: COLOR_33,
            C.WHITE: COLOR_255,
            C.AMBER: COLOR_255,
            C.FREE_COLOR: COLOR_255
        })
        self.manager.check_end_of_publishes(self)

    def test_external_set_waf(self):
        self._set_dtr(0, 44)
        self._set_dtr(1, 255)
        self._set_dtr(2, 255)
        self._emulate_external(dali.gear.colour_control.SetWafColour(dali.address.Short(63)))
        self._check_state('Color', {
            C.RED: COLOR_255,
            C.GREEN: COLOR_255,
            C.BLUE: COLOR_255,
            C.WHITE: COLOR_44,
            C.AMBER: COLOR_255,
            C.FREE_COLOR: COLOR_255
        })
        self.manager.check_end_of_publishes(self)

    def test_level_to_raw(self):
        for raw, level in [
            (0, 0),
            (10, 413), # 10.4902
            (11, 414), # 10.5156
            (11, COLOR_11),
            (44, COLOR_44),
            (254, COLOR_254),
            (255, COLOR_255),
        ]:
            self.assertEqual(raw, Color.to_raw(level))

    def test_raw_to_level(self):
        for level, raw in [
            (0, 0),
            (COLOR_11, 11),
            (COLOR_44, 44),
            (COLOR_254, 254),
            (COLOR_255, 255),
        ]:
            self.assertEqual(level, Color.from_raw(raw))


class TestFunitColor(TestDaliRgb):
    def setUp(self) -> None:
        super(TestFunitColor, self).setUp()
        self.provider.values[ChannelsCount] = 5
        self.provider._fadeTime = 0

    def _check_rgb_commands(self, r, g, b):
        self.bus.check_bus_message(self, dali.gear.general.DTR0(r))
        self.bus.check_bus_message(self, dali.gear.general.DTR1(g))
        self.bus.check_bus_message(self, dali.gear.general.DTR2(b))
        self.bus.check_bus_message(self, dali.gear.colour_control.SetRgbColour(dali.address.Short(63)))

    def _check_waf_commands(self, w, a, f):
        self.bus.check_bus_message(self, dali.gear.general.DTR0(w))
        self.bus.check_bus_message(self, dali.gear.general.DTR1(a,))
        self.bus.check_bus_message(self, dali.gear.general.DTR2(f))
        self.bus.check_bus_message(self, dali.gear.colour_control.SetWafColour(dali.address.Short(63)))

    def _check_activate_command(self):
        self.bus.check_bus_message(self, dali.gear.colour_control.Activate(dali.address.Short(63)))

    def _check_query(self):
        for i in range(5):
            self.bus.check_bus_message(self, dali.gear.general.DTR0(9+i))
            self.bus.check_bus_message(self, dali.gear.colour_control.QueryColourValue(dali.address.Short(63)))

    def _check_preliminary_state(self, *args):
        value = dict(zip((C.RED, C.GREEN, C.BLUE, C.WHITE, C.AMBER), args))
        super(TestFunitColor, self)._check_state('Color', value)

    def _check_finish(self):
        self.bus.check_end_bus_messages(self)
        self._check_end_of_publishes()

    def test_set_invalid_component(self):
        funit = Color(self.provider, Variable({C.RED: 10_001}))

        with self.assertRaises(ValueError) as ex:
            funit.set()

        self.assertEqual(ex.exception.args[0], f'Components level expected as [0..10000]: 10001')

    def test_set_incomplete(self):
        funit = Color(self.provider, Variable({C.BLUE: COLOR_22, C.FREE_COLOR: COLOR_66}))
        self.assertEqual(([255, 255, 22], [255, 255, 66]), funit.set_value)

    def test_set_rgb_only(self):
        funit = Color(self.provider, Variable({C.GREEN: COLOR_22}))
        self.assertEqual(([255, 22, 255], None), funit.set_value)

    def test_set_waf_only(self):
        funit = Color(self.provider, Variable({C.AMBER: COLOR_55}))
        self.assertEqual((None, [255, 55, 255]), funit.set_value)

    def test_set_empty(self):
        funit = Color(self.provider, Variable({}))

        with self.assertRaises(ValueError) as ex:
            _ = funit.set_value

        self.assertEqual(ex.exception.args[0], 'RGB and WAF is absent')

    def test_set(self):
        funit = Color(self.provider, Variable({
            C.RED: COLOR_11,
            C.GREEN: COLOR_22,
            C.BLUE: COLOR_33,
            C.WHITE: COLOR_44,
            C.AMBER: COLOR_55
        }))

        self.assertEqual(([11, 22, 33], [44, 55, 255]), funit.set_value)

        funit.set()

        self._check_rgb_commands(11, 22, 33)
        self._check_waf_commands(44, 55, 255)
        self._check_activate_command()
        self._check_query()

        self._check_state(
            'Color',
            {
                C.RED: COLOR_11,
                C.GREEN: COLOR_22,
                C.BLUE: COLOR_33,
                C.WHITE: None,
                C.AMBER: None
            },
            check_value=False,
            preliminary=True
        )

        self._check_state(
            'Color',
            {
                C.RED: COLOR_128,
                C.GREEN: COLOR_128,
                C.BLUE: COLOR_128,
                C.WHITE: COLOR_128,
                C.AMBER: COLOR_128
            },
            check_value=False
        )

        self._check_state(
            'Color',
            {
                C.RED: COLOR_128,
                C.GREEN: COLOR_128,
                C.BLUE: COLOR_128,
                C.WHITE: COLOR_44,
                C.AMBER: COLOR_55
            },
            check_value=False,
            preliminary=True
        )

        self._check_state(
            'Color',
            {
                C.RED: COLOR_128,
                C.GREEN: COLOR_128,
                C.BLUE: COLOR_128,
                C.WHITE: COLOR_128,
                C.AMBER: COLOR_128
            }
        )

        self._check_query()

        self._check_finish()

    def test_get(self):
        funit = Color(self.provider, Variable(None))
        funit.get()

        self._check_query()
        self._check_state(
            'Color',
            {
                C.RED: COLOR_128,
                C.GREEN: COLOR_128,
                C.BLUE: COLOR_128,
                C.WHITE: COLOR_128,
                C.AMBER: COLOR_128
            }
        )
        self._check_finish()
