import importlib
from unittest import TestCase

import spread_core.tools.settings_const as C
import spread_core.bam.generator as generator
from spread_core.errors.project_errors import ProjectError

generate_provider = generator.generate_provider


class _TestDaliProviders(TestCase):
    def _check_provider_generation(self, provider_type, attributes=None):
        provider = generate_provider(999, provider_type, {C.ATTRIBUTES: attributes or {}})
        self.assertEqual(type(provider).__name__, provider_type)


class TestDaliProviderGeneration(_TestDaliProviders):
    @classmethod
    def setUpClass(cls) -> None:
        super(TestDaliProviderGeneration, cls).setUpClass()
        importlib.import_module('spread_core.bam.dali.providers')

    def test_unknown_type(self):
        with self.assertRaises(ProjectError):
            generate_provider(999, 'SomeNonEexistType', {})

    def test_missing_attributes(self):
        with self.assertRaises(ProjectError) as ex:
            generate_provider(999, 'DaliProvider', {})

        self.assertEqual(
            ex.exception.args[0], f"DaliProvider(999): parse_args() missing 1 required positional argument: '{C.ATTRIBUTES}'"
        )

    def test_dimmer(self):
        self._check_provider_generation("DaliDimmer")

    def test_relay(self):
        self._check_provider_generation("DaliRelay")

    def test_tunable_white(self):
        self._check_provider_generation("DaliTunableWhite")

    def test_presence_sensor(self):
        self._check_provider_generation("DaliPresenceSensor")

    def test_light_sensor(self):
        self._check_provider_generation("DaliLightSensor")

    def test_push_button(self):
        self._check_provider_generation("DaliPushButton", {"instanceIndex": 31})

    def _check_provider_generation(self, provider_type, attributes: dict = None):
        if attributes is None:
            attributes = {}

        attributes = {
            "binding": "Device",
            "device": 63,
            **attributes
        }

        super(TestDaliProviderGeneration, self)._check_provider_generation(provider_type, attributes)


class TestMercuryProviderGeneration(_TestDaliProviders):
    @classmethod
    def setUpClass(cls) -> None:
        super(TestMercuryProviderGeneration, cls).setUpClass()
        importlib.import_module('spread_core.bam.mercury.providers')

    def test_meter(self):
        attributes = {
            "address": 254,
            "password1": 'p1',
            "password2": 'p2'
        }
        self._check_provider_generation("MercuryElectricMeter", attributes)


class TestKnxProviderGeneration(_TestDaliProviders):
    @classmethod
    def setUpClass(cls) -> None:
        super(TestKnxProviderGeneration, cls).setUpClass()
        importlib.import_module('spread_core.bam.knx.providers')

    def setUp(self) -> None:
        self.attributes = {
            "state": "0/0/1",
            "control": "0/1/0"
        }

    def test_missing_state(self):
        with self.assertRaises(ProjectError) as ex:
            self._check_provider_generation("KnxDimmer", {"control": '1/2/3'})

        self.assertEqual(ex.exception.args[0], "KnxDimmer(999): __init__() missing 1 required positional argument: 'state'")

    def test_missing_control(self):
        with self.assertRaises(ProjectError) as ex:
            self._check_provider_generation("KnxDimmer", {"control": '4/5/6'})

        self.assertEqual(ex.exception.args[0], "KnxDimmer(999): __init__() missing 1 required positional argument: 'state'")

    def test_dimmer(self):
        self._check_provider_generation("KnxDimmer", self.attributes)

    def test_relay(self):
        self._check_provider_generation("KnxRelay", self.attributes)

    def test_motor(self):
        self._check_provider_generation("KnxMotor", self.attributes)


class TestLomProviderGeneration(_TestDaliProviders):
    @classmethod
    def setUpClass(cls) -> None:
        super(TestLomProviderGeneration, cls).setUpClass()
        importlib.import_module('spread_core.bam.lom.providers')

    def setUp(self) -> None:
        self.attributes = {
            "devEui": "ABCDEF123456789",
        }

    def test_missing_address(self):
        with self.assertRaises(ProjectError) as ex:
            self._check_provider_generation("LomDimmer")

        self.assertEqual(ex.exception.args[0], "LomDimmer(999): 'devEui'")

    def test_dimmer(self):
        self._check_provider_generation("LomDimmer", self.attributes)
