from unittest import TestCase
import asyncio

from spread_core.bam.engineries.lighting_area import ButtonPreset
from spread_lom.profiles import DimmingProfile
from spread_core.projector.base import FileSystemStorage, file_sys_storage_modify
from spread_core import mqtt
import os
import json
import shutil


class TestFileSystemStorage(TestCase):
    root = '/var/project/test'
    project_id = 123
    storage = FileSystemStorage(root, project_id)

    def setUp(self) -> None:
        # Clear test dir
        if os.path.exists(self.root):
            shutil.rmtree(self.root)

    def create_file(self, file_path, data):
        # Create test file
        file_dir = os.path.dirname(file_path)
        if not os.path.exists(file_dir):
            os.makedirs(file_dir, exist_ok=True)
        with open(file_path, mode='w') as file:
            json.dump(data, file, indent=2)

    def save_topic(self, topic: mqtt.project.Local, data):
        self.storage.save(topic, data)
        if data:
            if self.assertTrue(os.path.exists(topic.path(self.root))):
                with open(topic.path(self.root), mode='rb') as file:
                    self.assertEqual(data, json.load(file))
        else:
            self.assertFalse(os.path.exists(topic.path(self.root)))

    def load_topic(self, topic: mqtt.project.Local, data):
        if data:
            self.assertTrue(os.path.exists(topic.path(self.root)))
        else:
            self.assertFalse(os.path.exists(topic.path(self.root)))
        self.assertEqual(data, self.storage.load(topic))

    def test_save(self):
        self.save_topic(mqtt.project.Preset(self.project_id, ButtonPreset.__name__, 1), 'test_data')
        self.save_topic(mqtt.project.Preset(self.project_id, ButtonPreset.__name__, 1), None)
        self.save_topic(mqtt.project.Cycle(self.project_id, 1), 'test_data')
        self.save_topic(mqtt.project.Cycle(self.project_id, 1), None)

    def test_load(self):
        topic = mqtt.project.Preset(self.project_id, ButtonPreset.__name__, 1)
        self.load_topic(topic, None)
        self.create_file(topic.path(self.root), 'test_data')
        self.load_topic(topic, 'test_data')

        topic = mqtt.project.Cycle(self.project_id, 1)
        self.load_topic(topic, None)
        self.create_file(topic.path(self.root), 'test_data')
        self.load_topic(topic, 'test_data')

    def test_topics(self):
        topics = {
            mqtt.project.Preset(self.project_id, ButtonPreset.__name__, 1),
            mqtt.project.Cycle(self.project_id, 1)
        }
        for topic in topics:
            self.create_file(topic.path(self.root), 'test_data')
            topic = str(topic)
        storage_topics = {topic for topic in self.storage.topics}
        self.assertSetEqual(topics, storage_topics)


class TestFileSysStorageModify(TestCase):
    class Client:
        def __init__(self):
            self.calls = []

        async def publish(self, topic, payload, retain):
            self.calls.append((topic, payload, retain))

    def setUp(self) -> None:
        self.root = '/var/project/test'
        self.project_id = 123
        self.storage = FileSystemStorage(self.root, self.project_id)
        self.data = 'qwe'
        self.change_list = {
            'Local/Equipment/Lighting/LightingArea/ButtonPresets':
                mqtt.project.Preset(self.project_id, ButtonPreset.__name__, 1).path(self.root),
            'Local/Hardware/Lom/Profiles':
                mqtt.project.Preset(self.project_id, DimmingProfile.__name__, 1).path(self.root),
            'Local/Incubator/Cycles':
                mqtt.project.Cycle(self.project_id, 1).path(self.root)
        }

        # Clear test dir
        if os.path.exists(self.root):
            shutil.rmtree(self.root)

        # Create test files
        for name in self.change_list:
            file_dir = os.path.join(self.root, str(self.project_id), name)
            if not os.path.exists(file_dir):
                os.makedirs(file_dir, exist_ok=True)

            # В spread_dashboard файлы Cycle создаются без расширения!!!
            if name == 'Local/Incubator/Cycles':
                file_name = '1'
            else:
                file_name = '1.json'

            with open(os.path.join(file_dir, file_name), mode='w') as file:
                json.dump(self.data, file, indent=2)

            if os.listdir(file_dir) != [file_name]:
                self.skipTest(f'Directory "{file_dir}" contains extra files. For clear test, remove this files.')

    def test_positive(self):
        client = self.Client()
        asyncio.run(file_sys_storage_modify(client, self.root, self.project_id))
        calls = []
        for src in self.change_list:
            calls.append((f'Project/{self.project_id}/{src}/1', bytes(), True))

        self.assertListEqual(calls, client.calls)

        for name in self.change_list:
            with self.subTest(file_path=self.change_list[name]):
                self.assertFalse(os.path.exists(os.path.join(self.root, str(self.project_id), name, '1.json')))
                if self.assertTrue(os.path.exists(self.change_list[name])):
                    with open(self.change_list[name], "rb") as file:
                        self.assertEqual(self.data, json.load(file))

        # with mock.AsyncMock(Client, 'publish') as mocked_client_publish:
        #     asyncio.run(file_sys_storage_modify(Client(hostname='host', port=1883), self.root, self.project_id))
        #
        #     calls = []
        #     for src in self.change_list:
        #         calls.append(mock.call(f'Project/{self.project_id}/{src}/1', payload=bytes(), retain=True))
        #
        #     self.assertTrue(mocked_client_publish.called)
        #     mocked_client_publish.assert_has_calls(calls)
