import logging
import os.path
import sqlite3

from spread_core.tools.path import DUMP_PATH

UNKNOWN = 0
NORMAL = 1
FUNCTION_TEST = 2
DURATION_TEST = 3


class DbClient:
    _primary_key = dict(field='id', descr='INTEGER PRIMARY KEY AUTOINCREMENT')

    def __init__(self, name: str, path=''):
        self._name = name
        self._path = path
        self._connection = None

    @property
    def primary_key(self):
        return self._primary_key['field']

    @property
    def db_name(self) -> str:
        if self._path and len(self._path) > 0:
            return '/'.join([self._path, self._name])
        return self._name

    @property
    def database(self):
        try:
            if self._connection is None:
                self._connection = sqlite3.connect(self.db_name)
            return self._connection
        except sqlite3.Error:
            raise NotImplementedError()

    @database.setter
    def database(self, value):
        if self._connection:
            self._connection.close()
            self._connection = None

    def _execute_command(self, query: str, need_commit=True) -> bool:
        if query and len(query) > 0:
            try:
                _cursor = self.database.cursor()
                _cursor.execute(query)
                if need_commit:
                    self.database.commit()
                self.database = None
                return True
            except sqlite3.Error:
                self.database = None
                return False

    def _execute_query(self, query: str, count=0) -> list:
        if query and len(query) > 0:
            try:
                _cursor = self.database.cursor()
                _cursor.execute(query)
                if count:
                    _result = _cursor.fetchmany(count)
                else:
                    _result = _cursor.fetchall()
                _cursor.close()
                self.database = None
                return _result
            except sqlite3.Error:
                self.database = None
                return list()

    def get_table_struct(self, table_name: str) -> list:
        if table_name is None or table_name == '':
            raise NotImplementedError()

        try:
            _cursor = self.database.cursor()
            _cursor.execute(f'PRAGMA table_info({table_name})')
            _result = list()
            for _field in _cursor.fetchall():
                if _field[1] != self.primary_key:
                    _d = _field[2]
                    if _field[5] == 1:
                        _d += ' PRIMARY KEY AUTOINCREMENT'
                    if _field[3] == 1:
                        _d += ' NOT NULL'

                    _field_descr = dict(field=_field[1], descr=_d)
                    _result.append(_field_descr)
                else:
                    continue
            _cursor.close()
            self.database = None
            return _result
        except sqlite3.Error:
            self.database = None
            return list()

    def _to_add(self, original_struct: list, expected_struct: list) -> list:
        _to_add = list()
        for _expected in expected_struct:
            if _expected not in original_struct:
                _to_add.append(_expected)
        return _to_add

    def _to_del(self, original_struct: list, expected_struct: list):
        _to_del = list()
        for _orig in original_struct:
            if _orig['field'] != self.primary_key:
                if _orig not in expected_struct:
                    _to_del.append(_orig)
        return _to_del

    def _create_table(self, table_name: str, table_struct: list):
        if table_name is None or table_name == '':
            raise NotImplementedError()

        if table_struct is None or len(table_struct) == 0:
            raise NotImplementedError()

        _fields = f"{self._primary_key['field']} {self._primary_key['descr']}"
        for _f in table_struct:
            if len(_fields) > 0:
                _fields += ', '
            _fields += f"{_f['field']} {_f['descr']}"

        return self._execute_command(f'CREATE TABLE {table_name} ({_fields})')

    def _update_table(self, table_name: str, to_add: list, to_del: list):
        if table_name is None or table_name == '':
            raise NotImplementedError()

        _changed = False
        if len(to_add) > 0:
            for _f in to_add:
                _changed = self._execute_command(f"ALTER TABLE {table_name} ADD COLUMN {_f['field']} {_f['descr']}")
        if len(to_del) > 0:
            for _f in to_del:
                _changed = self._execute_command(f"ALTER TABLE {table_name} DROP COLUMN {_f['field']}")
        return _changed

    def create_table(self, table_name: str, table_struct: list):
        if table_name is None or table_name == '':
            raise NotImplementedError()

        if table_struct is None or len(table_struct) == 0:
            raise NotImplementedError()

        _original_struct = self.get_table_struct(table_name)
        if len(_original_struct) == 0:
            self._create_table(table_name, table_struct)
        else:
            _to_add = self._to_add(_original_struct, table_struct)
            _to_del = self._to_del(_original_struct, table_struct)
            self._update_table(table_name, _to_add, _to_del)

    def drop_table(self, table_name: str):
        if table_name is None or table_name == '':
            raise NotImplementedError()

        return self._execute_command(f'DROP TABLE {table_name}')

    def insert(self, table_name: str, values: list):
        if table_name is None or table_name == '':
            raise NotImplementedError()

        _fields = ''
        _values = ''
        for v in values:
            if len(_fields) > 0:
                _fields += ', '
                _values += ', '
            _fields += v['field']
            _values += '"%s"' % v['value']

        return self._execute_command(f'INSERT INTO {table_name} ({_fields}) VALUES ({_values})')

    def update(self, table_name: str, values: list, condition: str):
        if table_name is None or table_name == '':
            raise NotImplementedError()
        if condition is None or condition == '':
            raise NotImplementedError()

        _data = ''
        for v in values:
            if len(_data) > 0:
                _data += ', '
            _data += v['field'] + "=" + "'" + v['value'] + "'"

        return self._execute_command(f'UPDATE {table_name} SET {_data} WHERE {condition}')

    def select(self, table_name: str, condition: str) -> list:
        if table_name is None or table_name == '':
            raise NotImplementedError()

        if condition and len(condition) > 0:
            condition = f'WHERE {condition}'
        else:
            condition = ''

        return self._execute_query(f'SELECT * FROM {table_name} {condition}')

    def count(self, table_name: str, condition: str) -> int:
        if table_name is None or table_name == '':
            raise NotImplementedError()

        _query = f'SELECT COUNT({self.primary_key}) FROM {table_name}'
        if condition and len(condition) > 0:
            _query += f'WHERE {condition}'

        _result = self._execute_query(_query, 1)
        for _count in _result:
            return _count[0]

    def delete(self, table_name: str, condition: str):
        try:
            if condition and len(condition) > 0:
                self.database.cursor().execute(f'DELETE FROM {table_name} WHERE {condition}')
            else:
                self.database.cursor().execute(f'DELETE FROM {table_name}')
            self.database.commit()
            self.database = None
        except sqlite3.Error:
            self.database = None
            return

    def pack(self):
        try:
            self.database.execute('VACUUM')
            self.database = None
        except sqlite3.Error:
            return


class EmergencyLightDb(DbClient):
    _database = os.path.join(DUMP_PATH, 'emergency.sqlite')
    _table_name = 'info'
    _struct = [dict(field='address', descr='TEXT'),
               dict(field='min_level', descr='INTEGER'),
               dict(field='max_level', descr='INTEGER'),
               dict(field='emergency_level', descr='INTEGER'),
               dict(field='battery_charge', descr='INTEGER'),
               dict(field='status', descr='INTEGER'),
               dict(field='function_test_date', descr='TEXT'),
               dict(field='function_test_result', descr='TEXT'),
               dict(field='duration_test_date', descr='TEXT'),
               dict(field='duration_test_result', descr='TEXT')
               ]

    def __init__(self, dev_id: int):
        super().__init__(self._database)
        self.create_table(self._table_name, self._struct)
        self._id = dev_id

    @property
    def dev_id(self) -> int:
        return self._id

    @property
    def table(self):
        return self._table_name

    def add_device(self, dev_id: int):
        self.delete(self.table, f'address="{self.dev_id}"')
        self.insert(self.table, [dict(field='address', value=str(dev_id))])

    def add_min_level(self, min_level: int):
        self._save_info('min_level', str(min_level))

    def add_max_level(self, max_level: int):
        self._save_info('max_level', str(max_level))

    def add_emergency_level(self, emergency_lvl: int):
        self._save_info('emergency_level', str(emergency_lvl))

    def add_battery_charge(self, battery_lvl: int):
        self._save_info('battery_charge', str(battery_lvl))

    def add_status(self, status: str):
        self._save_info('status', str(EmergencyLightDb.covert_status(status)))

    def add_function_test_date(self, date: str):
        self._save_info('function_test_date', date)

    def add_duration_test_date(self, date: str):
        self._save_info('duration_test_date', date)

    def add_function_test_result(self, result: str):
        self._save_info('function_test_result', result)

    def add_duration_test_result(self, result: str):
        self._save_info('duration_test_result', result)

    def add_info(self, data: dict):
        try:
            _info = list()
            _info.append(dict(field='address', value=data.pop('address', None)))
            _info.append(dict(field='min_level', value=data.pop('min_level', 0)))
            _info.append(dict(field='max_level', value=data.pop('max_level', 0)))
            _info.append(dict(field='emergency_level', value=data.pop('emergency_level', 0)))
            _info.append(dict(field='battery_charge', value=data.pop('battery_charge', 0)))
            _info.append(dict(field='status', value=EmergencyLightDb.covert_status(data.pop('status', None))))
            _info.append(dict(field='function_test_date', value=data.pop('function_test_date', None)))
            _info.append(dict(field='function_test_result', value=data.pop('function_test_result', None)))
            _info.append(dict(field='duration_test_date', value=data.pop('duration_test_date', None)))
            _info.append(dict(field='duration_test_result', value=data.pop('duration_test_result', None)))
            return self.insert(self.table, _info)
        except Exception as ex:
            logging.error('DB INSERT ERROR:' + str(ex))

    def update_info(self, index: str, kwargs: dict):
        try:
            _condition = f'address = {index}'

            _info = list()
            if 'min_level' in kwargs.keys():
                _info.append(dict(field='min_level', value=kwargs.pop('min_level', 0)))
            if 'max_level' in kwargs.keys():
                _info.append(dict(field='max_level', value=kwargs.pop('max_level', 0)))
            if 'emergency_level' in kwargs.keys():
                _info.append(dict(field='emergency_level', value=kwargs.pop('emergency_level', 0)))
            if 'battery_charge' in kwargs.keys():
                _info.append(dict(field='battery_charge', value=kwargs.pop('battery_charge', 0)))
            if 'status' in kwargs.keys():
                _info.append(dict(field='status', value=kwargs.pop('status', 0)))
            if 'function_test_date' in kwargs.keys():
                _info.append(dict(field='function_test_date', value=kwargs.pop('function_test_date', 0)))
            if 'function_test_result' in kwargs.keys():
                _info.append(dict(field='function_test_result', value=kwargs.pop('function_test_result', 0)))
            if 'duration_test_date' in kwargs.keys():
                _info.append(dict(field='duration_test_date', value=kwargs.pop('duration_test_date', 0)))
            if 'duration_test_result' in kwargs.keys():
                _info.append(dict(field='duration_test_result', value=kwargs.pop('duration_test_result', 0)))

            return self.update(self.table, _info, _condition)
        except Exception as ex:
            logging.error('DB UPDATE ERROR:' + str(ex))

    def _save_info(self, field: str, value: str) -> bool:
        return self.update(self.table, [dict(field=field, value=value)], f'address="{self.dev_id}"')

    @staticmethod
    def covert_status(status_str: str) -> int:
        status_str = status_str.lower()
        if status_str == 'normal':
            return NORMAL
        elif status_str == 'function test':
            return FUNCTION_TEST
        elif status_str == 'duration test':
            return DURATION_TEST
        else:
            return UNKNOWN
