import json
import os

from django.db import models as django_models
from jsonfield import JSONField
from spread_core import mqtt
from spread_core.tools.path import DUMP


class Entity(django_models.Model):
    class Meta:
        abstract = True

    id = django_models.PositiveIntegerField(primary_key=True)
    type = django_models.CharField(max_length=255)
    name = django_models.CharField(max_length=255, blank=True)
    attributes = JSONField()
    key = django_models.CharField(max_length=64)

    @property
    def v_name(self): return self.name if self.name else 'id:%s' % self.id

    @property
    def name_not_star(self):
        if self.name.endswith('*'):
            return self.name[:-1]
        return self.v_name

    def serialize(self) -> dict:
        return {field.attname: getattr(self, field.attname) for field in self._meta.fields if not field.auto_created}

    @staticmethod
    def separator(entity: dict) -> dict: return entity

    def mqtt_address(self, funit):
        raise NotImplementedError()

    def state(self, funit):
        return mqtt.spread.topic.State(
            self.mqtt_address(funit)
        )

    def renew(self, funit):
        return mqtt.spread.topic.Renew(
            self.mqtt_address(funit)
        )

    def set(self, funit):
        return mqtt.spread.topic.Set(
            self.mqtt_address(funit)
        )


class Header(django_models.Model):
    version = django_models.CharField(max_length=4)
    firmware = JSONField()
    project = JSONField()

    @property
    def project_id(self) -> int:
        return self.project['id']

    @property
    def root_location(self) -> int:
        return self.project['rootLocationID']

    @property
    def title(self) -> int:
        return self.project['title']

    @property
    def name(self) -> int:
        return self.project['name']


class Manager(Entity):
    serverID = django_models.PositiveIntegerField()


class Enginery(Entity):
    recipe = JSONField(blank=True)
    label = django_models.CharField(max_length=16)

    @staticmethod
    def get_scheduled(loc_id):
        return Enginery.objects.filter(
            name__endswith='*',
            id__in=Subginery.objects.get(
                locationID=loc_id,
                type='Lighting').engineries
        )

    def mqtt_address(self, funit):
        return mqtt.spread.address.EngineryAddress(
            Header.objects.get().project_id,
            self.type,
            self.id,
            funit
        )


class Provider(Entity):
    manager = django_models.ForeignKey(Manager, on_delete=django_models.deletion.CASCADE, db_column='managerID')

    def mqtt_address(self, funit):
        return mqtt.spread.address.ProviderAddress(
            Header.objects.get().project_id,
            self.manager.type,
            self.manager.id,
            self.type,
            self.id,
            funit
        )


class Cycle(django_models.Model):
    id = django_models.AutoField(primary_key=True)
    name = django_models.CharField(max_length=255, blank=True)
    period = django_models.PositiveIntegerField()
    duration = django_models.FloatField()
    startLevel = django_models.PositiveIntegerField()
    triggers = JSONField(default={})

    @staticmethod
    def fill():
        """ Метод для заполнения таблиц из файлов старой версии """
        project_dir = str(Header.objects.get().project_id)
        cycle_dir = DUMP(project_dir, 'Local', 'Cycle')
        with os.scandir(cycle_dir) as cycles:
            for cycle in cycles:
                with open(cycle.path, 'r', encoding='utf-8') as obj:
                    data = json.load(obj)
                cycle_id = cycle.name.split('.')[0]  # избавились от возможного расширения
                Cycle.objects.update_or_create(id=int(cycle_id), defaults=data)
        # теперь припишем циклы к локациям
        loc_dir = DUMP(project_dir, 'Location')
        with os.scandir(loc_dir) as locations:
            for loc in locations:
                path = os.path.join(loc.path, 'Incubator')
                if os.path.exists(path):
                    data = {}
                    with os.scandir(path) as params:
                        for param in params:
                            with open(param.path, 'r', encoding='utf-8') as obj:
                                # 1-я буква должна быть маленькой
                                name = param.name[0].lower() + param.name[1:]
                                name = 'cycle_id' if name == 'cycle' else name
                                value = obj.readline().strip('\n"')
                                value = int(value) if value.isdigit() else value
                                if value == 'true':
                                    value = True
                                elif value == 'false':
                                    value = False
                                data[name] = value
                        Location.objects.update_or_create(id=int(loc.name), defaults=data)


class Location(Entity):
    scheduled = django_models.BooleanField()
    enabled = django_models.BooleanField(null=True)  # поле enabled из schlocation
    navigations = JSONField(default=[], blank=True)
    models = JSONField(default=[], blank=True)
    type = None
    # поля, относящиеся к циклу инкубатора
    cycle = django_models.ForeignKey(Cycle, null=True, on_delete=django_models.deletion.SET_NULL)
    active = django_models.BooleanField(default=False)
    recursive = django_models.BooleanField(default=False)
    start_time = django_models.DateField(null=True, blank=True)

    @staticmethod
    def separator(entity: dict) -> dict:
        scheduled = False
        if entity.get('name', '').endswith('*'):
            entity['name'] = entity.get('name', '')[:-1].strip()
            scheduled = True

        return {**{k: entity[k] for k in entity if k in ['id', 'name', 'key', 'navigations', 'models']}, 'scheduled': scheduled}

    def engineries(self, available_types):
        engineries = []
        for model in self.models:
            for ctrl in model.get('controls', []):
                if ctrl.get('clickable', False) and str(ctrl.get('engineryID', None)).isdigit():
                    engineries.append(int(ctrl['engineryID']))

        yield from Enginery.objects.select_related().filter(id__in=engineries, type__in=available_types)

    @property
    def sub_locations(self):
        for sub_location in Location.objects.filter(navigations__contains=self.id):
            if sub_location.navigations[:-1] == self.navigations:
                yield sub_location

    def __str__(self): return '%s(%s)' % (self.name, self.id)

    @staticmethod
    def as_tree() -> list:
        def make_item(location: Location) -> dict:
            return {
                'id': location.id,
                'name': location.name,
                'locations': [make_item(sub) for sub in location.sub_locations]
            }

        header = Header.objects.first()
        result = [make_item(Location.objects.get(id=header.root_location))]

        return result


class Subginery(Entity):
    locationID = django_models.PositiveIntegerField()
    engineries = JSONField(blank=True)


class RecipeIngredient(django_models.Model):
    id = django_models.AutoField(primary_key=True)
    enginery = django_models.ForeignKey(Enginery, on_delete=django_models.deletion.CASCADE)
    provider = django_models.ForeignKey(Provider, on_delete=django_models.deletion.CASCADE)
    index = django_models.PositiveIntegerField()
    ingredient_type = django_models.CharField(max_length=255, blank=True)
    recipe_type = django_models.CharField(max_length=255, blank=True)

    @staticmethod
    def fill_recipes():
        """ Заполняет таблицу при обновлении проекта """
        engineries = Enginery.objects.exclude(recipe="")
        for enginery in engineries:
            recipe = enginery.recipe
            if not recipe:
                continue

            ingredients = recipe.get('ingredients', [])
            for ingr in ingredients:
                try:
                    provider = Provider.objects.get(id=int(ingr.get('id')))
                    RecipeIngredient.objects.update_or_create(
                        enginery=enginery,
                        provider=provider,
                        ingredient_type=ingr.get('type', ''),
                        index=ingr.get('index'),
                        recipe_type=recipe.get('type', '')
                    )
                except Provider.DoesNotExist:
                    print(f"\n Can`n set RecipeIngredient for {enginery} as Provider({ingr.get('id')})")
                except Exception as exc:
                    print(f'\n {enginery}; Ingredient: {ingr}; {provider}', str(exc))
