import json
from collections import Counter
from unittest import TestCase

import spread_core.bam.entities
from spread_core import mqtt
from spread_core.mqtt.spread.address import EngineryAddress, SubgineryAddress, ProviderAddress
from spread_core.mqtt.spread.topic import Set, State

from spread_core.mqtt.spread.variable import Variable
from spread_core.tools import settings

from spread_core.bam.subgineries.subginery import Subginery

from spread_core.bam.engineries import Enginery

from spread_core.bam.recipe import Recipe

from spread_core.bam.equipment import Equipment, LightingEquipment, AlarmEquipment, ShadingEquipment, \
                                      HandlingEquipment, ClimateEquipment
from spread_core.tests.utils import TestMqttClient


class TestEquipmentAdapter(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        setattr(settings.config, 'args', True)

    def setUp(self) -> None:
        self.broker = TestMqttClient()
        adapter = Equipment(1, self.broker)
        adapter.subginery_type = Subginery
        adapter.add_manager(2, 'Manager')
        adapter.add_provider(Recipe(1, "Manager", 2, "Provider", 21))
        adapter.add_provider(Recipe(1, "Manager", 2, "Provider", 22))
        adapter.add_enginery(self.enginery(211), 21, "Simple")
        adapter.add_enginery(self.enginery(212), 21, "Simple")
        adapter.add_enginery(self.enginery(221), 22, "Simple")
        adapter.add_enginery(self.enginery(222), 22, "Simple")
        adapter.add_subginery(Subginery(3, 4, [211, 212, 221, 222]))

        self.entities = adapter._entities

        self.adapter = adapter

    @staticmethod
    def enginery(id: int) -> Enginery:
        eng = Enginery(id)
        eng.funit_class = spread_core.bam.entities.Funit
        return eng

    def check_publish(self, _topic, _value, _retain=True, params=None, **kwargs):
        if len(self.broker.messages) == 0:
            self.fail('Expected publish is absent: %s' % _topic)
        else:
            topic, data, retain, delay, log_level, kwargs = self.broker.messages.popleft()
            variable = Variable.of(json.loads(data))
            self.assertEqual(_topic, topic)
            self.assertEqual(_value, variable.value)
            self.assertEqual(_retain, retain)

            if params:
                for param_name in params:
                    try:
                        self.assertEqual(params[param_name], variable.params[param_name])
                    except KeyError:
                        self.fail('Expected variable parameter "%s" is absent' % param_name)

    def check_subscribes(self, _topic):
        try:
            topic = self.broker.topics.popleft()
        except IndexError:
            self.fail('Expected subscribe is absent: %s' % _topic)

        self.assertEqual(_topic, topic)

    def check_end_of_publishes(self):
        if self.broker.messages:
            unexpected_publishes = f'Unexpected publishes({len(self.broker.messages)} pcs):\n'
            unexpected_publishes += "\n".join(str(i) for i in self.broker.messages)
            self.fail(unexpected_publishes)

    def check_end_of_subscribes(self):
        if self.broker.topics:
            unexpected = f'Unexpected subscribe({len(self.broker.topics)} pcs):\n'
            unexpected += "\n".join(str(i) for i in self.broker.topics)
            self.fail(unexpected)

    def check_exploit_topics(self, adapter, topics):
        adapter.exploit_topics(self.broker.subscribe)

        for topic in topics:
            self.check_subscribes(topic)

        self.check_end_of_subscribes()

    def check_command_handling(self, entity, topic):
        data = []
        entity.on_command = lambda *args, **kwargs: data.extend(args)

        variable = Variable(123)

        self.adapter.on_command(topic, variable)

        self.assertIs(topic, data[0])
        self.assertIs(variable, data[1])

    def emulate_project_message(self, file: str, data: dict, next_file=None):
        self.broker.topics.append(f'Project/1/File/{file}.json')
        self.adapter.on_project(mqtt.project.ProjectFile(1, f'{file}.json'), json.dumps({file: [data]}))

        if next_file:
            self.check_subscribes(f'Project/1/File/{next_file}.json')
            self.check_end_of_subscribes()

    def test_init(self):
        self.assertEqual(1, self.adapter.id)
        self.assertEqual(False, self.adapter.loaded)

    def test_start(self):
        self.adapter.loaded = False
        self.adapter.start()
        self.check_subscribes('Project/1/File/managers.json')
        self.check_end_of_publishes()

    def test_hot_start(self):
        self.adapter.loaded = True

        for entity in self.entities.values():
            entity.values['Funit'] = 123

        self.adapter.start()
        self.check_publish('Spread/State/1/Equipment/Enginery/211/Funit', 123)
        self.check_publish('Spread/State/1/Equipment/Enginery/212/Funit', 123)
        self.check_publish('Spread/State/1/Equipment/Enginery/221/Funit', 123)
        self.check_publish('Spread/State/1/Equipment/Enginery/222/Funit', 123)
        self.check_publish('Spread/State/1/Location/4/Subginery/Funit', 123)

        self.check_end_of_publishes()
        self.check_end_of_subscribes()

    def test_project_managers_receiver(self):
        manager_data = {
            'id': 999,
            'type': "Manager"
        }

        self.emulate_project_message('managers', manager_data, next_file='providers')
        self.assertIn(999, self.adapter._managers)

    def test_project_providers_receiver(self):
        provider_data = {
            'id': 999,
            'type': "Provider",
            "managerID": 2
        }

        self.emulate_project_message('providers', provider_data, next_file='engineries')
        self.assertIn(999, self.adapter._managers[2]['providers'])

    def test_project_engineries_receiver(self):
        self.adapter.enginery_types = (Enginery, )

        enginery_data = {
            'id': 999,
            'type': "Enginery",
            'recipe': {
                'type': 'Simple',
                'ingredients': [
                    {
                        "index": 0,
                        "type": "Solid",
                        "id": 21
                    }
                ]
            }
        }
        self.emulate_project_message('engineries', enginery_data, next_file='subgineries')
        self.assertIn(999, self.entities)
        del self.entities[999]

        enginery_data['recipe']['ingredients'][0]['id'] = 99
        self.emulate_project_message('engineries', enginery_data)
        self.assertNotIn(999, self.entities)

        enginery_data['recipe']['ingredients'].pop(0)
        self.emulate_project_message('engineries', enginery_data)
        self.assertNotIn(999, self.entities)

        del enginery_data['recipe']['ingredients']
        self.emulate_project_message('engineries', enginery_data)
        self.assertNotIn(999, self.entities)

        del enginery_data['recipe']
        self.emulate_project_message('engineries', enginery_data)
        self.assertNotIn(999, self.entities)

        del enginery_data['id']
        self.emulate_project_message('engineries', enginery_data)
        self.assertNotIn(999, self.entities)

    def test_project_subgineries_receiver(self):
        counter = Counter()

        self.adapter.on_ready = lambda: counter.update(['on_ready'])

        self.emulate_project_message('subgineries', {
            'id': 999,
            'type': "Subginery",
            'locationID': 4,
            'engineries': []
        })

        self.assertIn(4, self.entities)
        self.assertEqual(1, counter['on_ready'], '"on_ready" not handled')
        self.check_end_of_subscribes()

    def test_ready_handler(self):
        counter = Counter()
        self.adapter.exploit_topics = lambda action: counter.update([action.__name__])
        for entity in self.entities.values():
            entity.get_info = lambda: counter.update(['get_info'])

        self.adapter.on_ready()

        self.assertTrue(self.adapter.loaded)
        self.assertEqual(1, counter['subscribe'], '"exploit_topics(subscribe)" not called')
        self.assertEqual(5, counter['get_info'], '"get_info" handled %s pcs, but expected 5' % counter['get_info'])

    def test_exploit_light(self):
        self.check_exploit_topics(LightingEquipment(1, self.broker), [
            'Spread/State/1/Hardware/+/+/+/+/On',
            'Spread/State/1/Hardware/+/+/+/+/BrightnessLevel',
            'Spread/State/1/Hardware/+/+/+/+/Temperature',
            'Spread/State/1/Hardware/+/+/+/+/EngineryHoldTime',
            'Spread/State/1/Hardware/+/+/+/+/CurrentLuminosity',
            'Spread/State/1/Hardware/+/+/+/+/CurrentPresence',
            'Spread/State/1/Hardware/+/+/+/+/LastEvent',
            'Spread/State/1/Hardware/+/+/+/+/Color',
            'Spread/State/1/Hardware/+/+/+/+/ChargeLevel',
            'Spread/State/1/Hardware/+/+/+/+/LastFunctionTestInfo',
            'Spread/State/1/Hardware/+/+/+/+/LastDurationTestInfo',
            'Spread/State/1/Hardware/+/+/+/+/Mode',
            'Spread/Set/1/Location/+/Lighting/On',
            'Spread/Set/1/Location/+/Lighting/Off',
            'Spread/Set/1/Location/+/Lighting/BrightnessLevel',
            'Spread/Set/1/Location/+/Lighting/LightSensorsOn',
            'Spread/Set/1/Location/+/Lighting/LightSensorsOff',
            'Spread/Set/1/Location/+/Lighting/PresenceSensorsOn',
            'Spread/Set/1/Location/+/Lighting/PresenceSensorsOff',
            'Spread/Set/1/Location/+/Lighting/SaveScene',
            'Spread/Set/1/Location/+/Lighting/SceneOn',
            'Spread/Set/1/Equipment/DimmingLight/+/On',
            'Spread/Set/1/Equipment/DimmingLight/+/BrightnessLevel',
            'Spread/Set/1/Equipment/DimmingLight/+/GroupLevel',
            'Spread/Set/1/Equipment/DimmingLight/+/GroupOn',
            'Spread/Set/1/Equipment/SwitchingLight/+/On',
            'Spread/Set/1/Equipment/SwitchingLight/+/GroupLevel',
            'Spread/Set/1/Equipment/SwitchingLight/+/GroupOn',
            'Spread/Set/1/Equipment/TunableWhiteLight/+/On',
            'Spread/Set/1/Equipment/TunableWhiteLight/+/BrightnessLevel',
            'Spread/Set/1/Equipment/TunableWhiteLight/+/GroupLevel',
            'Spread/Set/1/Equipment/TunableWhiteLight/+/GroupOn',
            'Spread/Set/1/Equipment/TunableWhiteLight/+/Temperature',
            'Spread/Set/1/Equipment/PresenceSensor/+/On',
            'Spread/Set/1/Equipment/PresenceSensor/+/HoldTime',
            'Spread/Set/1/Equipment/LightSensor/+/On',
            'Spread/Set/1/Equipment/LightingArea/+/On',
            'Spread/Set/1/Equipment/LightingArea/+/TuningType',
            'Spread/Set/1/Equipment/LightingArea/+/Discovery',
            'Spread/Set/1/Equipment/LightingArea/+/OccupancyLevel',
            'Spread/Set/1/Equipment/LightingArea/+/VacancyLevel',
            'Spread/Set/1/Equipment/LightingArea/+/TargetLuminosity',
            'Spread/Set/1/Equipment/LightingArea/+/Hysteresis',
            'Spread/Set/1/Equipment/LightingArea/+/TuningSpeed',
            'Spread/Set/1/Equipment/LightingArea/+/OccupancyAction',
            'Spread/Set/1/Equipment/LightingArea/+/OccupancyScene',
            'Spread/Set/1/Equipment/LightingArea/+/VacancyAction',
            'Spread/Set/1/Equipment/LightingArea/+/VacancyScene',
            'Spread/Set/1/Equipment/LightingArea/+/OccupancySceneSave',
            'Spread/Set/1/Equipment/LightingArea/+/OccupancySceneLoad',
            'Spread/Set/1/Equipment/LightingArea/+/VacancySceneSave',
            'Spread/Set/1/Equipment/LightingArea/+/VacancySceneLoad',
            'Spread/Set/1/Equipment/LightingArea/+/Pause',
            'Spread/Set/1/Equipment/LightingArea/+/CurrentProfile',
            'Spread/Set/1/Equipment/LightingArea/+/Profiles',
            'Spread/Set/1/Equipment/LightingArea/+/HoldTime',
            'Spread/Set/1/Equipment/LightingArea/+/Buttons',
            'Spread/Set/1/Equipment/LightingArea/+/VirtualPresence',
            'Spread/Set/1/Equipment/LightingArea/+/VirtualLuminosity',
            'Spread/Renew/1/Equipment/LightingArea/+/On',
            'Spread/Renew/1/Equipment/LightingArea/+/TuningType',
            'Spread/Renew/1/Equipment/LightingArea/+/Discovery',
            'Spread/Renew/1/Equipment/LightingArea/+/OccupancyLevel',
            'Spread/Renew/1/Equipment/LightingArea/+/VacancyLevel',
            'Spread/Renew/1/Equipment/LightingArea/+/TargetLuminosity',
            'Spread/Renew/1/Equipment/LightingArea/+/Hysteresis',
            'Spread/Renew/1/Equipment/LightingArea/+/TuningSpeed',
            'Spread/Renew/1/Equipment/LightingArea/+/OccupancyAction',
            'Spread/Renew/1/Equipment/LightingArea/+/OccupancyScene',
            'Spread/Renew/1/Equipment/LightingArea/+/VacancyAction',
            'Spread/Renew/1/Equipment/LightingArea/+/VacancyScene',
            'Spread/Renew/1/Equipment/LightingArea/+/OccupancySceneSave',
            'Spread/Renew/1/Equipment/LightingArea/+/OccupancySceneLoad',
            'Spread/Renew/1/Equipment/LightingArea/+/VacancySceneSave',
            'Spread/Renew/1/Equipment/LightingArea/+/VacancySceneLoad',
            'Spread/Renew/1/Equipment/LightingArea/+/Presence',
            'Spread/Renew/1/Equipment/LightingArea/+/Luminosity',
            'Spread/Renew/1/Equipment/LightingArea/+/Pause',
            'Spread/Renew/1/Equipment/LightingArea/+/CurrentProfile',
            'Spread/Renew/1/Equipment/LightingArea/+/Profiles',
            'Spread/Renew/1/Equipment/LightingArea/+/HoldTime',
            'Spread/Renew/1/Equipment/LightingArea/+/Buttons',
            'Spread/Set/1/Equipment/RgbLight/+/On',
            'Spread/Set/1/Equipment/RgbLight/+/BrightnessLevel',
            'Spread/Set/1/Equipment/RgbLight/+/GroupLevel',
            'Spread/Set/1/Equipment/RgbLight/+/GroupOn',
            'Spread/Set/1/Equipment/RgbLight/+/Color',
            'Spread/Set/1/Equipment/RgbwLight/+/On',
            'Spread/Set/1/Equipment/RgbwLight/+/BrightnessLevel',
            'Spread/Set/1/Equipment/RgbwLight/+/GroupLevel',
            'Spread/Set/1/Equipment/RgbwLight/+/GroupOn',
            'Spread/Set/1/Equipment/RgbwLight/+/Color',
            'Spread/Set/1/Equipment/EmergencyUnit/+/EmergencyMode',
            'Spread/Set/1/Equipment/EmergencyUnit/+/FunctionTest',
            'Spread/Set/1/Equipment/EmergencyUnit/+/DurationTest',
            'Project/1/Local/Preset/ButtonPreset/#',
        ])

    def test_exploit_alarm(self):
        self.check_exploit_topics(AlarmEquipment(1, self.broker), [
            'Spread/State/1/Hardware/+/+/+/+/Guard',
            'Spread/State/1/Hardware/+/+/+/+/Triggered',
            'Spread/Set/1/Location/+/Alarm/IntruderSensorsGuard',
            'Spread/Set/1/Location/+/Alarm/FireSensorsGuard',
            'Spread/Set/1/Location/+/Alarm/LeakageSensorsGuard',
            'Spread/Set/1/Equipment/FireSensor/+/Guard',
            'Spread/Set/1/Equipment/IntruderSensor/+/Guard',
            'Spread/Set/1/Equipment/LeakageSensor/+/Guard'
        ])

    def test_exploit_shading(self):
        self.check_exploit_topics(ShadingEquipment(1, self.broker), [
            'Spread/State/1/Hardware/+/+/+/+/PositionLevel',
            'Spread/State/1/Hardware/+/+/+/+/Motion',
            'Spread/Set/1/Location/+/Shading/Motion',
            'Spread/Set/1/Location/+/Shading/PositionLevel',
            'Spread/Set/1/Location/+/Shading/Rotation',
            'Spread/Set/1/Location/+/Shading/PositionAngle',
            'Spread/Set/1/Equipment/Shutter/+/Motion',
            'Spread/Set/1/Equipment/Shutter/+/PositionLevel',
            'Spread/Set/1/Equipment/Curtain/+/Motion',
            'Spread/Set/1/Equipment/Curtain/+/PositionLevel',
            'Spread/Set/1/Equipment/Blind/+/Motion',
            'Spread/Set/1/Equipment/Blind/+/PositionLevel',
            'Spread/Set/1/Equipment/Blind/+/Rotation',
            'Spread/Set/1/Equipment/Blind/+/PositionAngle'
        ])

    def test_exploit_handling(self):
        self.check_exploit_topics(HandlingEquipment(1, self.broker), [
            'Spread/Set/1/Equipment/Scenario/+/On'
        ])

    def test_exploit_climate(self):
        self.check_exploit_topics(ClimateEquipment(1, self.broker), [
            'Spread/State/1/Hardware/+/+/+/+/Temperature',
            'Spread/State/1/Hardware/+/+/+/+/TargetTemperature',
            'Spread/State/1/Hardware/+/+/+/+/On',
            'Spread/State/1/Hardware/+/+/+/+/Preset',
            'Spread/State/1/Hardware/+/+/+/+/Mode',
            'Spread/State/1/Hardware/+/+/+/+/FanSpeedMode',
            'Spread/State/1/Hardware/+/+/+/+/LouverMode',
            'Spread/Set/1/Location/+/Climate/TargetTemperature',
            'Spread/Set/1/Location/+/Climate/ThermoregulatorsOn',
            'Spread/Set/1/Location/+/Climate/Preset',
            'Spread/Set/1/Location/+/Climate/Mode',
            'Spread/Set/1/Location/+/Climate/FanSpeedMode',
            'Spread/Set/1/Location/+/Climate/LouverMode',
            'Spread/Set/1/Location/+/Climate/FansOn',
            'Spread/Set/1/Location/+/Climate/HeatedFloorsOn',
            'Spread/Set/1/Equipment/Thermoregulator/+/On',
            'Spread/Set/1/Equipment/Thermoregulator/+/TargetTemperature',
            'Spread/Set/1/Equipment/Thermoregulator/+/Preset',
            'Spread/Set/1/Equipment/Thermoregulator/+/Mode',
            'Spread/Set/1/Equipment/Thermoregulator/+/FanSpeedMode',
            'Spread/Set/1/Equipment/Thermoregulator/+/LouverMode',
            'Spread/Set/1/Equipment/Fan/+/On',
            'Spread/Set/1/Equipment/HeatedFloor/+/On'
        ])

    def test_command_handling(self):
        entity = Enginery(99)
        topic = Set(EngineryAddress(1, 'Enginery', 99, 'Funit'))
        self.adapter.add_enginery(entity, 21, 'Simple')

        self.check_command_handling(entity, topic)

    def test_command_handling_2(self):
        entity = Subginery(98, 99, [])
        topic = Set(SubgineryAddress(1, 99, "Subginery", 'Funit'))
        self.adapter.add_subginery(entity)

        self.check_command_handling(entity, topic)

    def test_state_handling(self):
        topic = State(ProviderAddress(1, 'Manager', 2, 'Provider', 21, 'Funit'))
        variable = Variable(123)

        data = []
        self.adapter.get_recipe(2, 21).dispatch_state = lambda *args, **kwargs: data.extend(args)

        self.adapter.on_state(topic, variable)

        self.assertIs(topic, data[0])
        self.assertIs(variable, data[1])

    def test_exit_handling(self):
        counter = Counter()
        for entity in self.entities.values():

            entity.set_invalid = lambda *args, **kwargs: counter.update(['set_invalid'])

        self.adapter.on_exit()

        self.assertIs(5, counter['set_invalid'])

    def test_local_handling(self):
        adapter = LightingEquipment(1, self.broker)
        topic = mqtt.project.Preset(1, 'ButtonPreset', 99)

        adapter.on_local(topic, json.dumps({'handlers': []}))
        self.assertIn(99, adapter.button_presets)

        adapter.on_local(topic, b'')
        self.assertNotIn(99, adapter.button_presets)
